﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows;

namespace FliProCameraLib
{
    /// <summary>
    /// This file contains overridden methods specific 
    /// to second generation FLI Cameras using libflipro.dll
    /// </summary>
    public class FliProCameraGen2 : FliProCamera
    {
        //**************************************************************
        #region Public Constants

        public const uint CAPABILITIES_LENGTH_VERSION_1 = 58;
        public const uint CAPABILITIES_LENGTH_VERSION_2 = 74;
        public const uint CAPABILITIES_VERSION_1 = 1;
        public const uint CAPABILITIES_VERSION_2 = 2;

        public const int INITIAL_TARGET_TEMPERATURE = -15;

        public const int MAX_EXPOSURE_TIME_MSECS = 39600000;       // 11 hours
        public const int MAX_EXPOSURE_TIME_SECS = 39600;           // 11 hours
        public const int MAX_EXPOSURE_TIME_MINS = 660;             // 11 hours
        public const int MAX_EXPOSURE_TIME_HRS = 11;               // 11 hours

        // Factory Config constants
        public const int FACTORY_CONFIG_DEVICE_BACK_SIDE_ILLUMINATION = 0x31;
        public const int FACTORY_CONFIG_DEVICE_COOLING_RATE_5_DEGREES = 0x62;
        public const int FACTORY_CONFIG_DEVICE_SHUTTER_DELAY = 0x31;
        public const int FACTORY_CONFIG_REGISTER_BACK_SIDE_ILLUMINATION = 0x0A;
        public const int FACTORY_CONFIG_REGISTER_BACK_SIDE_ILLUMINATION_MASK = 0x20000000;
        public const int FACTORY_CONFIG_REGISTER_COOLING_RATE_5_DEGREES = 0x4C;
        public const uint FACTORY_CONFIG_REGISTER_COOLING_RATE_5_DEGREES_MASK = 0xFFFFFFFF;
        public const int FACTORY_CONFIG_REGISTER_COOLING_RATE_5_DEGREES_VALUE = 180;
        public const int FACTORY_CONFIG_REGISTER_COOLING_RATE_10_DEGREES_VALUE = 90;
        public const int FACTORY_CONFIG_REGISTER_SHUTTER_DELAY = 0x38;
        public const int FACTORY_CONFIG_REGISTER_SHUTTER_DELAY_OPEN_MASK = 0xFFFF;
        public const uint FACTORY_CONFIG_REGISTER_SHUTTER_DELAY_MASK = 0xFFFFFFFF;
        public const int FACTORY_CONFIG_SERIAL_NUM_MAX_CHARS = 24;
        public const int FACTORY_CONFIG_REGISTER_COUNT = 3;           // The number of extra registers that we are setting

        // Register constants
        //public const int REGISTER_C_VALUE_BIAS_FRAME_MASK = 0x40000000;
        //public const int REGISTER_C_VALUE_DARK_FRAME_MASK = 0x20000000;
        //public const int REGISTER_C_VALUE_NORMAL_FRAME_MASK = 0;
        //public const int REGISTER_C_VALUE_FRAME_TYPE_CLEAR_MASK = 0x1FFFFFFF;
        //public const uint REGISTER_9_VALUE_LED_FLOOD_CLEAR_MASK = 0xFFFFFF7F;
        //public const int REGISTER_9_VALUE_LED_FLOOD_ON_MASK = 0x00000080;
        //public const int REGISTER_9_VALUE_FORCE_TRAINING_MASK = 0x00000010;
        //public const uint REGISTER_60_VALUE_LED_TOGGLE_ON = 0xFFFFFFFF;
        //public const uint REGISTER_60_VALUE_LED_TOGGLE_OFF = 0;

        //public const uint REGISTER_9_INDEX = 0x9;           // Register 9
        //public const uint REGISTER_C_INDEX = 0xC;           // Register C
        //public const uint REGISTER_60_INDEX = 0x3C;         // Register 3C

        #endregion Public Constants

        //**************************************************************
        #region Constructor and Initialization

        public FliProCameraGen2(FPRODEVICEINFO DevInfo = new FPRODEVICEINFO()) : base(DevInfo)
        {
            fliProCameraManager = FliProCameraManager.Instance;
            Type = CameraType.GEN2;
            BitsPerPixel = FliFrame.BITS_PER_PIXEL_12;
            exposureTimer.Elapsed += OnExposureTimedEvent;
        }

        protected override void Initialize()
        {
            FliFrameCountList = new System.Collections.ObjectModel.ObservableCollection<FliFrameCountNameValueClass>();
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Single", FliFrameCount.SINGLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Multiple", FliFrameCount.MULTIPLE));
            FliFrameCountList.Add(new FliFrameCountNameValueClass("Continuous", FliFrameCount.CONTINUOUS));

            FliFrameTypeList = new System.Collections.ObjectModel.ObservableCollection<FliFrameTypeNameValueClass>();
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Normal", FliFrameType.NORMAL));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Dark", FliFrameType.DARK));
            FliFrameTypeList.Add(new FliFrameTypeNameValueClass("Bias", FliFrameType.BIAS));
        }

        #endregion Constructor and Initialization

        //**************************************************************
        #region Public Types

        public enum DeviceID_Enum
        {
            I2C_DEVICE_ID_8051 = 0x62,
            I2C_DEVICE_ID_FPGA = 0x31,
            I2C_DEVICE_ID_FX3EEPROM = 0x50,
            I2C_DEVICE_ID_FPGAEEPROM = 0x32,
            I2C_DEVICE_ID_QSFPHOSTIF = 0x33,
            I2C_DEVICE_ID_DAC_1 = 0x4C,
            I2C_DEVICE_ID_DAC_2 = 0x4D
        }

        #endregion Public Types

        //**************************************************************
        #region Protected Properties

        private static byte[] ReturnArray = new byte[20];
        private System.Timers.Timer exposureTimer = new System.Timers.Timer();
        private AutoResetEvent autoEvent = new AutoResetEvent(false);
        private string gs_errorText = String.Empty;
        private bool gb_errorEncountered = false;
        private bool gb_PopulatingUI = false;
        private FliProCameraManager fliProCameraManager;

        #endregion Protected Properties

        //**************************************************************
        #region Observable Properties
        #endregion Observable Properties

        //**************************************************************
        #region Public Properties

        //-------------------------------------------
        /// <summary>
        /// LSB position of active data within each frame pixel value
        /// </summary>
        public override uint PixelLSB
        {
            get { return _PixelLSB; }

            set
            {
                if ((value + PixelDepth) > MAX_PIXEL_DEPTH)
                {
                    value = MAX_PIXEL_DEPTH - PixelDepth;
                }

                if (SetValue<uint>(ref _PixelLSB, value))
                {
                    SetPixelLSB(value);
                }
            }
        }
        protected uint _PixelLSB;

        public override bool HDRMode
        {
            get { return _HDRMode; }
            protected set { SetValue<bool>(ref _HDRMode, value); }
        }
        protected bool _HDRMode;



        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Cancel the active exposure, if any
        /// </summary>
        public override void CancelExposure()
        {
            IsAborted = true;
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            exposureTimer.Enabled = false;
            autoEvent.Set();
            NativeMethods.FPROFrame_CaptureAbort(Handle);
        }

        public override bool CheckGPS()
        {
            bool changed = false;
            FliGPSState newGPSState = FliGPSState.FPRO_GPS_NOT_DETECTED;

            if (NativeMethods.FPROCtrl_GetGPSState(Handle, out newGPSState) >= 0)
            {
                if ((newGPSState == FliGPSState.FPRO_GPS_NOT_DETECTED) && (CurrentGPSState != FliGPSState.FPRO_GPS_NOT_DETECTED))
                {
                    // If we have detected that the GPS is present, don't allow the screen icon to disappear
                    newGPSState = FliGPSState.FPRO_GPS_DETECTED_NO_SAT_LOCK;
                }
                changed = (CurrentGPSState != newGPSState);
                CurrentGPSState = newGPSState;
            }

            return changed;
        }

        //-------------------------------------------
        /// <summary>
        /// Cancel the active exposure, if any
        /// </summary>
        public override void ClearTriggerWatcher()
        {
            if (Handle < 0)
            {
                IsAborted = true;
                return;
            }
            NativeMethods.FPROFrame_CaptureAbort(Handle);
        }

        //-------------------------------------------
        /// <summary>
        /// Stop exposing a frame (Deprecated)
        /// </summary>
        public override void EndExposure()
        {
            if (IsCapturing)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                exposureTimer.Enabled = false;
                autoEvent.Set();
                NativeMethods.FPROFrame_CaptureEnd(Handle);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Start exposing a frame (Deprecated?)
        /// </summary>
        public override void ExposeFrame()
        {
            //-=-=ThrowOnError(NativeMethods.FLIExposeFrame(Handle));
        }

        public override void ForceTraining(FliProCamera camera)
        {
            fliProCameraManager.PublishStatus("Force Training ... ", false);

            fliProCameraManager.ActiveCamera.SetTrainingEnable(true);
            //Thread.Sleep(100);
            fliProCameraManager.ActiveCamera.SetTrainingEnable(false);
            fliProCameraManager.PublishStatus("Done", true);
        }

        public override long GetAbsoluteExposureTime(long value, ExposureUnits_Enum unit)
        {
            long factor = 1;

            switch (unit)
            {
                case ExposureUnits_Enum.units_uSecs:
                    if ((value < 20) && (value > 0))
                    {
                        value = 20;
                    }
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1000000;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000000000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = (long)1000000000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)1000000000 * 3600;
                    break;
            }
            value *= factor;

            return value;
        }

        //-------------------------------------------
        /// <summary>
        /// Read camera base temperature 
        /// </summary>
        /// <returns></returns>
        public override double GetBaseTemperature()
        {
            int value = 0;
            if (Handle < 0)
            {
                IsAborted = true;
                return value;
            }
            NativeMethods.FPROCtrl_GetSensorTemperature(Handle, out value);
            BaseTemperature = value;
            return BaseTemperature;
        }

        public override int GetCoolerDutyCycle(ref int DutyCycle)
        {
            int status = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return status;
            }
            status = NativeMethods.FPROCtrl_GetCoolerDutyCycle(Handle, ref DutyCycle);

            return status;
        }

        public override int GetHDREnable(ref bool bIsHDR)
        {
            int status = 0;
            bIsHDR = true;   // default to true

            if (Handle < 0)
            {
                IsAborted = true;
                return status;
            }
            status = NativeMethods.FPROSensor_GetHDREnable(Handle, out bIsHDR);

            return status;
        }
        
        public override uint GetFWRevision()
        {
            uint fwVersion = 0;
            var versions = new FPRODEVICEVERSIONS();

            try
            {
                NativeMethods.FPROCam_GetDeviceVersion(Handle, out versions);

                fwVersion = Convert.ToUInt32(versions.FirmwareVersion);
            }
            catch { }

            return fwVersion;
        }

        public override void GetDeviceVersion(ref string firmwareVersion, ref string fpgaVersion, ref string controllerVersion)
        {
            try
            {
                // Debug
                //var mode = new SENSEMODE();
                //NativeMethods.FPROSensor_GetMode(Handle, 0, out mode);
                //End debug

                var versions = new FPRODEVICEVERSIONS();
                NativeMethods.FPROCam_GetDeviceVersion(Handle, out versions);

                firmwareVersion = versions.FirmwareVersion;
                fpgaVersion = versions.FPGAVersion;
                controllerVersion = versions.ControllerVersion;
            }
            catch
            {
                firmwareVersion = "";
                fpgaVersion = "";
                controllerVersion = "";
            }
        }

        //-------------------------------------------
        //-=-= TODO: Convert to Property
        public override int GetRemainingExposureTime()
        {
            int time = 0;
            //-=-=ThrowOnError(NativeMethods....);
            return time;
        }

        //-------------------------------------------
        /// <summary>
        /// Read camera sensor temperature 
        /// </summary>
        /// <returns></returns>
        public override double GetSensorTemperature()
        {
            double pAmbientTemp = 0;
            double pBaseTemp = 0;
            double pCoolerTemp = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return pBaseTemp;
            }
            int status = NativeMethods.FPROCtrl_GetTemperatures(Handle, ref pAmbientTemp, ref pBaseTemp, ref pCoolerTemp);
            //ThrowOnError(NativeMethods.FPROCtrl_GetBaseTemperature(Handle, out value));
            SensorTemperature = pCoolerTemp;
            return SensorTemperature;
        }

        public override bool GetShutterOpen()
        {
            bool isOpen = ShutterOpen;
            int status = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return isOpen;
            }
            status = NativeMethods.FPROCtrl_GetShutterOpen(Handle, out isOpen);

            return isOpen;
        }

        public override int GetTemperatures(ref double AmbientTemp, ref double BaseTemp, ref double CoolerTemp)
        {
            int status = 0;

            if (Handle < 0)
            {
                IsAborted = true;
                return status;
            }
            status = NativeMethods.FPROCtrl_GetTemperatures(Handle, ref AmbientTemp, ref BaseTemp, ref CoolerTemp);
            SensorTemperature = CoolerTemp;
            BaseTemperature = BaseTemp;
            AmbientTemperature = AmbientTemp;

            return status;
        }

        public override FliRectangle GetVisibleArea()
        {
            FliRectangle visible = new FliRectangle(0, 0, DEFAULT_IMAGE_WIDTH - 1, DEFAULT_IMAGE_HEIGHT - 1);

            // If we have connected then use the values from the device
            if (MaxImageHeight > 0)
            {
                visible = new FliRectangle(0, 0, MaxImageWidth - 1, MaxImageHeight - 1);
            }

            return visible;
        }

        //-------------------------------------------
        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold the specified number of pixels at the current
        /// camera settings
        /// </summary>
        /// <param name="pixelcount">number of pixels</param>
        /// <returns></returns>
        public override int PixelsToBytes(int pixelcount)
        {
            //-=-= FIXME: This must be generalized to support all possible pixel depths,
            //-=-=        line sizes, and binning settings
            int bytes = MetaDataSize;
            double binningFactor = 1;

            // For now the camera only does 2 x 2 binning
            if ((DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode()) &&
                ((HBin == 2) && (VBin == 2)))
            {
                binningFactor = 0.25;   // 2 x 2 binning takes 4 pixels and makes them 1 resulting in 1/4 the data
            }

            bytes += (int)((((pixelcount * (int)PixelDepth) * binningFactor) + 7) / 8);
            return bytes;
        }

        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold an image of the specified size at the current
        /// camera settings
        /// </summary>
        /// <param name="width">image width in pixels</param>
        /// <param name="height">image height in pixels</param>
        /// <returns></returns>
        public override int PixelsToBytes(int width, int height)
        {
            if (HDRMode) height *= 2;

            if ((DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()))
            {
                width += COBALT_PRESCAN_PIXELS * 2;  // Prescan pixels are at the start of each sector
            }
            int pixels = width * height;

            return PixelsToBytes(pixels);
        }

        public override void SetDefaultSupportedPixelDepths()
        {
            SupportedPixelDepths.Clear();
            SupportedPixelDepths.Add(12);
            PixelDepth = SupportedPixelDepths[0];
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        private long GetExposureTimeInMS(long value, ExposureUnits_Enum units)
        {
            double factor = 0;
            double realValue = value;
            switch (units)
            {
                case ExposureUnits_Enum.units_uSecs:
                    factor = 0.001;
                    break;
                case ExposureUnits_Enum.units_mSecs:
                    factor = 1;
                    break;
                case ExposureUnits_Enum.units_secs:
                    factor = 1000;
                    break;
                case ExposureUnits_Enum.units_mins:
                    factor = (long)1000 * 60;
                    break;
                case ExposureUnits_Enum.units_hours:
                    factor = (long)1000 * 3600;
                    break;
            }
            realValue *= factor;
            value = (int)realValue;

            return value;
        }

        /// <summary>
        /// Evaluates the firmware on the camera and raises a message if it will not work with this version of pilot
        /// </summary>
        /// <returns></returns>
        private bool CameraFirmwareIncompatable()
        {
            bool incompatable = false;
            string firmwareVersion = String.Empty;
            string fpgaVersion = String.Empty;
            string controllerVersion = String.Empty;

            try
            {
                GetDeviceVersion(ref firmwareVersion, ref fpgaVersion, ref controllerVersion);
                int index = firmwareVersion.IndexOf(".");
                int major = Convert.ToInt32(firmwareVersion.Substring(0, index));
            }
            catch
            { }

            return incompatable;
        }

        private void OnExposureTimedEvent(Object source, System.Timers.ElapsedEventArgs e)
        {
            autoEvent.Set();
        }

        //-------------------------------------------
        /// <summary>
        /// Set both binning values
        /// </summary>
        /// <param name="vValue"></param>
        /// <param name="hValue"></param>
        protected override void SetBinningIndex(int value)
        {
            // This is how the new interface will set binning
            // ThrowOnError(NativeMethods.FPROSensor_SetBinningIndex(Handle, value));
        }

        /// <summary>
        /// Sets the register to capture image data
        /// </summary>
        /// <param name="value"></param>
        protected override void SetEnableImageData(bool value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROFrame_SetImageDataEnable(Handle, value);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set the target base temperature
        /// </summary>
        /// <param name="value"></param>
        protected override void SetTargetTemperature(double value)
        {
            //int ivalue= (int)(value * TEMPERATURE_SCALE_FACTOR); 
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetTemperatureSetPoint(Handle, value);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set the pixel bit depth
        /// </summary>
        /// <param name="value"></param>
        protected override void SetPixelDepth(uint value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROFrame_SetPixelConfig(Handle, value, PixelLSB);
            }
            // Clamp the LSB if overflow
            if (value + PixelLSB > MAX_PIXEL_DEPTH)
            {
                PixelLSB = MAX_PIXEL_DEPTH - value;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set the Pixel LSB position
        /// </summary>
        /// <param name="value"></param>
        protected override void SetPixelLSB(uint value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROFrame_SetPixelConfig(Handle, PixelDepth, value);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set the Time Delay Integration Rate
        /// </summary>
        /// <param name="value"></param>
        protected override void SetTDIRate(int value)
        {
            //-=-=ThrowOnError(NativeMethods....);
        }

        //-------------------------------------------
        /// <summary>
        /// Set the Camera Mode
        /// </summary>
        /// <param name="value"></param>
        protected override void SetMode(int value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }

                NativeMethods.FPROSensor_SetMode(Handle, (uint)value);

                // Populate the screen from the camera settings now that we have changed the mode
                GetCameraControls(ref gs_errorText, ref gb_errorEncountered);
            }
            //-=-= TODO: is there a better way to do this?
            HDRMode = (value >= 2) || ((DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode()) &&
                (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) &&
                (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()));
        }

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable heater
        /// </summary>
        /// <param name="value"></param>
        protected override void SetHeaterEnable(bool value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    if (value)
                    {
                        NativeMethods.FPROCtrl_SetHeaterPower(Handle, (int)HeaterPower);
                    }
                    else
                    {
                        NativeMethods.FPROCtrl_SetHeaterPower(Handle, 0);
                    }
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set heater power (0-100%)
        /// </summary>
        /// <param name="value"></param>
        protected override void SetHeaterPower(uint value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    if (HeaterEnabledOn)
                    {
                        NativeMethods.FPROCtrl_SetHeaterPower(Handle, (byte)value);
                    }
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Sets the Horizontal and Vertical Binning in the camera
        /// </summary>
        /// <param name="value"></param>
        protected override void SetBinning()
        {
            // For now the cameras can only do 2 x 2 binning on the later products.
            // All other binning will be done in software from the maximum pixels captured.
            if (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode())
            {
                bool found = false;
                for (int i = 0; i < SupportedBinRatios.Count; i++)
                {
                    if ((HBin == SupportedBinRatios[i].HBin) && (VBin == SupportedBinRatios[i].VBin))
                    {
                        found = true;
                        BinRatioValue = SupportedBinRatios[i];
                        break;
                    }
                }

                // For now always go with 1:1 binning
                found = false;

                if (found)
                {
                    NativeMethods.FPROSensor_SetBinning(Handle, VBin, HBin);
                }
                else
                {
                    NativeMethods.FPROSensor_SetBinning(Handle, 1, 1);
                }
                if ((BinRatioValue == null) || (BinRatioValue.HBin != HBin) || (BinRatioValue.VBin != VBin))
                {
                    // If the pair of HBin & VBin do not match a SupportedBinRatio, we send 1 x 1 binning to the camera
                    // and we will do the binning in software and display the target ratio.
                    BinRatioValue = new BinRatio(0, HBin, VBin);
                }
            }
            else
            {
                NativeMethods.FPROSensor_SetBinning(Handle, 1, 1);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// </summary>
        /// <param name="value"></param>
        protected override void SetFrameCountType(FliFrameCount value)
        {
            if (value == FliFrameCount.CONTINUOUS)
            {
                FrameCount = 0;
            }
            else if (value == FliFrameCount.MULTIPLE)
            {
                FrameCount = MultipleFrameCount;
            }
            else
            {
                FrameCount = 1;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// </summary>
        /// <param name="value"></param>
        protected override void SetFrameType(FliFrameType value)
        {
            if (value == FliFrameType.BIAS)
            {
                // For Bias we want to set the Exposure Time to 0
                NativeMethods.FPROCtrl_SetExposure(Handle, 0, 0, Immediate);
                NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_BIAS);  // Set the Frame Type to Bias
                //regValue |= REGISTER_C_VALUE_BIAS_FRAME_MASK;     // Set the Frame Type to Bias
            }
            else
            {
                // For Non Bias we want to restore the Exposure Time
                SetExposure(ExposureTime, ExposureDelayTime, ExposureUnits, ExposureDelayUnits, Immediate);
                if (FrameCountType == FliFrameCount.CONTINUOUS)
                {
                    FrameCount = 0;
                }
                else if (FrameCountType == FliFrameCount.MULTIPLE)
                {
                    FrameCount = MultipleFrameCount;
                }
                else
                {
                    FrameCount = 1;
                }

                if (value == FliFrameType.DARK)
                {
                    NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_DARK); // Set the Frame Type to Dark
                    //regValue |= REGISTER_C_VALUE_DARK_FRAME_MASK;      // Set the Frame Type to Dark
                }
                else
                {
                    NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_NORMAL); // Set the Frame Type to Normal
                    //regValue |= REGISTER_C_VALUE_NORMAL_FRAME_MASK;     // Set the Frame Type to Normal
                    if (value == FliFrameType.FLOOD)
                    {
                        NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_LIGHTFLASH);
                        //reg9Value |= REGISTER_9_VALUE_LED_FLOOD_ON_MASK;
                    }
                    else if (value == FliFrameType.RBI_FLUSH)
                    {
                        NativeMethods.FPROFrame_SetFrameType(Handle, FPRO_FRAME_TYPE.FPRO_FRAMETYPE_DARK); // Set the Frame Type to Dark
                    }
                }
            }

        }

        //-------------------------------------------
        /// <summary>
        /// </summary>
        /// <param name="value"></param>
        protected override void SetNFlushes(int value)
        {
            /// -=-= Is this supported?
        }

        //-------------------------------------------
        /// <summary>
        /// Set Exposure time (in microseconds)
        /// </summary>
        /// <param name="value"></param>
        protected override void SetExposure(long value, long delay, ExposureUnits_Enum unit, ExposureUnits_Enum unitDelay, bool Immediate)
        {
            value = GetAbsoluteExposureTime(value, unit);
            delay = GetAbsoluteExposureTime(delay, unitDelay);
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetExposure(Handle, (long)value, (long)delay, Immediate);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// </summary>
        /// <returns></returns>
        public override FliRectangleWH GetReadoutDimensions()
        {
            /// -=-= Is this supported/needed
            LastReadoutAreaWH = new FliRectangleWH(0, 0, 0, 0);
            return LastReadoutAreaWH;
        }

        public override string GetDeviceType()
        {
            string deviceTypeStr = String.Empty;

            switch ((FPRODEVICETYPE)DeviceType)
            {
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400:
                    deviceTypeStr += " (400)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040:
                    deviceTypeStr += " (4040)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE2020:
                    deviceTypeStr += " (2020)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_KODAK29050:
                    deviceTypeStr += " (K29050)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_KODAK47051:
                    deviceTypeStr += " (K47051)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42:
                    deviceTypeStr += " (DC230-42)";
                    break;
                case FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84:
                    deviceTypeStr += " (DC230-84)";
                    break;
                default:
                    deviceTypeStr += " (Unknown)";
                    break;
            }

            return deviceTypeStr;
        }

        protected override void SetDac(uint value)
        {
            //-=-=ThrowOnError(NativeMethods...);
        }

        protected override void SetPortConfig(int value)
        {
            //-=-=ThrowOnError(NativeMethods....);
        }

        protected override void SetBlackLevel(uint value, bool isHigh)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROSensor_SetBlackLevelAdjustEx(Handle, (isHigh ? FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_HDR : FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_LDR), value);
                }
            }
        }
        protected override void SetBlackSun(uint value, bool isHigh)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    uint echan = (uint)(isHigh ? FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_HDR.GetHashCode() :
                        FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_LDR.GetHashCode());
                    NativeMethods.FPROSensor_SetBlackSunAdjustEx(Handle, echan, value);
                }
            }
        }

        protected override void SetFanEnable(bool value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetFanEnable(Handle, value);
                }
            }
        }

        protected override int OpenDevice()
        {
            int status = -1;

            try
            {
                status = NativeMethods.FPROCam_Open(ref _DeviceInfo, out _Handle);
            }
            catch
            {
                status = -1;
            }
            return status;
        }

        protected override void CloseDevice()
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROCam_Close(Handle);
                EnableTemperatureTimer(false);   // Turn off the temperature timer.
            }
        }

        protected void GetCameraControls(ref string errorText, ref bool errorEncountered)
        {
            int result = 0;

            gb_PopulatingUI = true;

            uint adjustValue = 0;
            result = NativeMethods.FPROSensor_GetBlackLevelAdjustEx(Handle, (uint)FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_LDR.GetHashCode(), out adjustValue);
            if (result < 0)
            {
                errorText += "\nGet Black Level Adjust failed";
                errorEncountered = true;
            }
            else
            {
                BlackLevelLow = adjustValue;
            }

            result = NativeMethods.FPROSensor_GetBlackSunAdjustEx(Handle, (uint)FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_LDR.GetHashCode(), out adjustValue);
            if (result < 0)
            {
                errorText += "\nGet Black Sun Adjust failed";
                errorEncountered = true;
            }
            else
            {
                BlackSunLow = adjustValue;
            }

            if (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode())
            {
                result = NativeMethods.FPROSensor_GetBlackLevelAdjustEx(Handle, (uint)FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_HDR.GetHashCode(), out adjustValue);
                if (result < 0)
                {
                    errorText += "\nGet Black Level Adjust failed";
                    errorEncountered = true;
                }
                else
                {
                    BlackLevelHigh = adjustValue;
                }

                result = NativeMethods.FPROSensor_GetBlackSunAdjustEx(Handle, (uint)FPROBLACKADJUSTCHAN.FPRO_BLACK_ADJUST_CHAN_HDR.GetHashCode(), out adjustValue);
                if (result < 0)
                {
                    errorText += "\nGet Black Sun Adjust failed";
                    errorEncountered = true;
                }
                else
                {
                    BlackSunHigh = adjustValue;
                }
            }

            //
            // NOTE: We always want to keep the temperature that the user set regardless of the factory settings for this mode
            //
            //bool fanIsEnabled = false;
            //result = NativeMethods.FPROCtrl_GetFanEnable(Handle, out fanIsEnabled);
            //if (result < 0)
            //{
            //    errorText += "\nGet Fan Enable failed";
            //    errorEncountered = true;
            //}
            //else
            //{
            //    FanIsEnabledOn = fanIsEnabled;
            //}

            uint heaterPower = 0;
            result = NativeMethods.FPROCtrl_GetHeaterPower(Handle, out heaterPower);
            if (result < 0)
            {
                errorText += "\nGet Heater Power failed";
                errorEncountered = true;
            }
            else
            {
                HeaterPower = heaterPower;
                HeaterEnabledOn = (heaterPower > 0);
            }

            bool bOn = false;
            result = NativeMethods.FPROCtrl_GetIlluminationOn(Handle, out bOn);
            if (result < 0)
            {
                errorText += "\nGet Illumination On failed";
                errorEncountered = true;
            }
            else
            {
                IlluminationOn = bOn;
            }

            bool bEnable = false;
            result = NativeMethods.FPROFrame_GetImageDataEnable(Handle, out bEnable);
            if (result < 0)
            {
                errorText += "\nGet Image Data Enable failed";
                ImageDataEnabled = true;
            }
            else
            {
                ImageDataEnabled = bEnable;
            }

            uint uiOnDelay = 0;
            uint uiOffDelay = 0;
            result = NativeMethods.FPROCtrl_GetIlluminationDelay(Handle, out uiOnDelay, out uiOffDelay);
            if (result < 0)
            {
                errorText += "\nGet Illumination Delay failed";
                errorEncountered = true;
            }
            else
            {
                IlluminationOnDelay = (int)uiOnDelay;
                IlluminationOffDelay = (int)uiOffDelay;
            }

            bEnable = false;
            result = NativeMethods.FPROCtrl_GetSensorTemperatureReadEnable(Handle, out bEnable);
            if (result < 0)
            {
                errorText += "\nGet Sensor Temperature Read Enable failed";
                errorEncountered = true;
            }
            else
            {
                ReadDuringExposureOn = bEnable;
            }

            bool vInverted = false;
            bool hInverted = false;
            result = NativeMethods.FPROSensor_GetScanDirection(Handle, out hInverted, out vInverted);
            if (result < 0)
            {
                errorText += "\nGet Scan Direction failed";
                errorEncountered = true;
            }
            else
            {
                VerticalScanDirInverted = vInverted;
                HorizontalScanDirInverted = hInverted;
            }

            uint gainIndex = 0;
            if (LowGainCount > 0)
            {
                result = NativeMethods.FPROSensor_GetGainIndex(Handle, GainTable.LOW, out gainIndex);
                if (result < 0)
                {
                    errorText += "\nGet Gain Index for GainTable.LDR failed";
                    errorEncountered = true;
                }
                else
                {
                    if (IsGainIndexValid(LowGainTable, (int)gainIndex))
                    {
                        LowGainIndex = (int)gainIndex;
                    }
                    else
                    {
                        // This means that the factory config has not been set for this mode so just use the 0 entry
                        if ((LowGainTable != null) && (LowGainTable.Count > 0))
                        {
                            LowGainIndex = LowGainTable[0].Index;
                        }
                        else
                        {
                            LowGainIndex = 0;
                        }
                    }
                }
            }

            if (HighGainCount > 0)
            {
                result = NativeMethods.FPROSensor_GetGainIndex(Handle, GainTable.HIGH, out gainIndex);
                if (result < 0)
                {
                    errorText += "\nGet Gain Index for GainTable.HDR failed";
                    errorEncountered = true;
                }
                else
                {
                    HighGainIndex = (int)gainIndex;
                    if (IsGainIndexValid(HighGainTable, (int)gainIndex))
                    {
                        HighGainIndex = (int)gainIndex;
                    }
                    else
                    {
                        // This means that the factory config has not been set for this mode so just use the 0 entry
                        if ((HighGainTable != null) && (HighGainTable.Count > 0))
                        {
                            HighGainIndex = HighGainTable[0].Index;
                        }
                        else
                        {
                            HighGainIndex = 0;
                        }
                    }
                }
            }

            long ExposureTimeInMicros = 0;
            long ExposureDelayTimeInMicros = 0;
            bool immediate = false;
            result = NativeMethods.FPROCtrl_GetExposure(Handle, out ExposureTimeInMicros, out ExposureDelayTimeInMicros, out immediate);
            if (result < 0)
            {
                errorText += "\nGet Exposure Time failed ";
                errorEncountered = true;
            }
            else
            {
                ProcessExposureTime(ExposureTimeInMicros);
                ProcessExposureDelayTime(ExposureDelayTimeInMicros);
                Immediate = immediate;
            }

            gb_PopulatingUI = false;
        }

        public bool IsGainIndexValid(ObservableCollection<SensorGain> gainTable, int gainIndex)
        {
            bool isValid = false;
            if (gainTable != null)
            {
                for (int i = 0; i < gainTable.Count; i++)
                {
                    if (gainTable[i].Index == gainIndex)
                    {
                        isValid = true;
                        break;
                    }
                }
            }

            return isValid;
        }

        public override int GetTemperatureSetPoint(ref double SetPoint)
        {
            int result = -1;

            SetPoint = 25;

            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return result;
                }
                result = NativeMethods.FPROCtrl_GetTemperatureSetPoint(Handle, ref SetPoint);
            }
            return 0;
        }

        protected override void SetImageArea(FliRectangle value)
        {
            SetImageArea(new FliRectangleWH(value));
        }

        protected override void SetImageArea(FliRectangleWH value)
        {
            if ((value.UL.X + value.W > MaxImageWidth) && (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) &&
                 (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()))
            {
                ImageUpperLeftX = 0;
                ImageWidth = MaxImageWidth;
            }
            if (value.UL.Y + value.H > MaxImageHeight)
            {
                value.UL.Y = 0;
                value.H = MaxImageHeight;
                ImageUpperLeftY = 0;
                ImageHeight = MaxImageHeight;
            }
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }

                if (VBin > 1)
                {
                    // Need to start on the first bin row

                }

                // The sensor can only give the total width so always set the area to the total sensor width and crop it later.
                if ((DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                    (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()))
                {
                    // The Cobalt camera has to have the same number of rows above the centerline as it does below.
                    // Therefore we need to adjust the image in the Y direction to be correct.
                    if ((MaxImageHeight - (value.UL.Y + value.H)) < value.UL.Y)
                    {
                        // Then the bottom is closer to the Total Bottom than the top is to 0 so use the bottom distance
                        int newY = MaxImageHeight - (value.UL.Y + value.H);
                        RequestedImageAreaWH = new FliRectangleWH(0, newY, MaxImageWidth + (COBALT_PRESCAN_PIXELS * 2), MaxImageHeight - (2 * newY));
                    }
                    else
                    {
                        // Then the image is closer to the top
                        RequestedImageAreaWH = new FliRectangleWH(0, value.UL.Y, MaxImageWidth + (COBALT_PRESCAN_PIXELS * 2), MaxImageHeight - (2 * value.UL.Y));
                    }
                }
                else
                {
                    RequestedImageAreaWH = new FliRectangleWH(0, value.UL.Y, MaxImageWidth, value.H);
                }
                int result = NativeMethods.FPROFrame_SetImageArea(Handle,
                   (uint)RequestedImageAreaWH.UL.X, (uint)RequestedImageAreaWH.UL.Y, (uint)RequestedImageAreaWH.W, (uint)RequestedImageAreaWH.H);
            }
        }

        protected override ObservableCollection<BinRatio> GetBinningTable(int pTableSizeBytes)
        {
            ObservableCollection<BinRatio> binRatioList = new ObservableCollection<BinRatio>();
            int status = 0;

            try
            {
                ushort[] tablebuf = new ushort[pTableSizeBytes / 2];
                GCHandle bufGch = GCHandle.Alloc(tablebuf, GCHandleType.Pinned);
                IntPtr bufPtr = bufGch.AddrOfPinnedObject();
                BinRatio binRatio;

                status = NativeMethods.FPROSensor_GetBinningTable(Handle, bufPtr, out pTableSizeBytes);

                for (int i = 0; i < tablebuf.Length; i += 2) //We take two bytes per loop to create the binRatio
                {
                    binRatio = new BinRatio(i / 2, tablebuf[i], tablebuf[i + 1]);
                    binRatioList.Add(binRatio);
                }
            }
            catch { }

            return binRatioList;
        }

        protected override uint GetRawImageFrame(byte[] framebuf, int framesize, bool useDelay)
        {
            GCHandle bufGch = GCHandle.Alloc(framebuf, GCHandleType.Pinned);
            IntPtr bufPtr = bufGch.AddrOfPinnedObject();
            uint size = (uint)framesize;

            try
            {
                if (IsConnected)
                {
                    if (Handle < 0)
                    {
                        IsAborted = true;
                        return size;
                    }

                    // The Exposure Time plus the Delay Time constitutes the entire wait time for FPROFrame_GetVideoFrame
                    long ExposureTimeInMS = GetExposureTimeInMS(ExposureTime, ExposureUnits) +
                        (useDelay ? GetExposureTimeInMS(ExposureDelayTime, ExposureDelayUnits) : (ManualShutterControlEnabled ? 0 : ShutterOpenDelay));
                    if (!IsAborted)
                    {
                        int status = NativeMethods.FPROFrame_GetVideoFrame(Handle, bufPtr, ref size, (uint)ExposureTimeInMS);
                        if (status < 0 && !IsAborted)
                        {
                            MessageBox.Show("GetVideoFrame failed with a Semiphore error.", "Warning", MessageBoxButton.OK,
                                MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                            IsAborted = true;
                        }
                    }
                }
            }
            finally
            {
                bufGch.Free();
            }
            return size;
        }

        //-------------------------------------------
        /// <summary>
        /// Update dynamically changing values
        /// </summary>
        protected override void UpdateStatus()
        {
            GetBaseTemperature();
            GetSensorTemperature();
            // Add other camera status updates here
        }

        //-------------------------------------------
        /// <summary>
        /// Device Specific actions when connection is established
        /// </summary>
        protected override unsafe void OnConnect()
        {
            string errorText = String.Empty;
            bool errorEncountered = false;
            gs_errorText = String.Empty;

            // Make sure any defaults get down to the camera
            ForcePropertyChange = true;

            var cap = new CAMERACAP();
            uint caplength = (uint)sizeof(CAMERACAP);
            uint targetCaplength = caplength;
            int result = NativeMethods.GetCapabilities(Handle, out cap, ref caplength);

            if (result < 0)
            {
                MessageBox.Show("Failed to get the Sensor Capabilities from the Camera.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }
            // Need to add a check here for Version and Size here
            else if (((cap.CapVersion == CAPABILITIES_VERSION_1) && (caplength != CAPABILITIES_LENGTH_VERSION_1)) ||
                ((cap.CapVersion == CAPABILITIES_VERSION_2) && (caplength != CAPABILITIES_LENGTH_VERSION_2)))
            {
                MessageBox.Show("Camera did not return the correct Sensor Capabilities size:\n TargetSize = " +
                    (cap.CapVersion == CAPABILITIES_VERSION_1 ? CAPABILITIES_LENGTH_VERSION_1.ToString() : CAPABILITIES_LENGTH_VERSION_2.ToString()) + 
                    ", returnedSize = " + caplength.ToString(), "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            if (CameraFirmwareIncompatable())
            {
                return;
            }

            // Parse the capabilities struct
            ParseCapabilities(cap, caplength);

            // Camera Controls
            if (SupportedBinRatios.Count > 0)
            {
                BinRatioValue = SupportedBinRatios[0];
            }
            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            ShutterOpen = false;
            LEDOn = false;

            // Get the Gain Tables
            LowGainTable = GetSensorGainTable(GainTable.LOW, LowGainCount);
            HighGainTable = GetSensorGainTable(GainTable.HIGH, HighGainCount);

            // Get the list of available Modes
            uint modecount = 0;
            uint currentmode = 0;
            result = NativeMethods.FPROSensor_GetModeCount(Handle, out modecount, out currentmode);
            ModeCount = modecount;
            SupportedModes.Clear();
            for (uint ui = 0; ui < modecount; ui++)
            {
                var mode = new SENSEMODE();
                result = NativeMethods.FPROSensor_GetMode(Handle, ui, out mode);
                if (result < 0)
                {
                    errorText += "\nGet Mode failed getting mode = " + ui.ToString();
                    errorEncountered = true;
                    break;
                }
                else
                {
                    //// This code is added temporarily for the 4040 to filter out all but one gain value
                    //if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040.GetHashCode())
                    //{
                    //    if (mode.Name.Equals("Rolling HDR"))
                    //    {
                    //        SupportedModes.Add(new NameValuePair(mode.Name, (int)mode.Index));
                    //    }
                    //}
                    //else
                    //{
                        SupportedModes.Add(new NameValuePair(mode.Name, (int)mode.Index));
                    //}
                }
            }
            Mode = (int)currentmode;

            // SetMode calls GetCameraControls which will put any errors in the globals
            errorText += gs_errorText;
            errorEncountered = errorEncountered || gb_errorEncountered;
            gs_errorText = String.Empty;
            gb_errorEncountered = false;

            var array = new FliRectangle();
            //-=-=NativeMethods.FLIGetArrayArea(Handle, out array.UL.X, out array.UL.Y, out array.LR.X, out array.LR.Y);
            ArrayArea = array;

            VisibleArea = GetVisibleArea();

            // Put the camera in a known state
            PixelLSB = 0;
            PixelDepth = MaxPixelDepth;
            PixelSize = new FliPixelSize(DEFAULT_PIXEL_WIDTH, DEFAULT_PIXEL_HEIGHT);

            TDIRate = 0;

            NFlushes = 1;
            ImageArea = VisibleArea;
            VideoModeEnabled = false;
            ColorPattern = FliBayerPattern.NONE;

            // Now only send changes down to the camera
            ForcePropertyChange = false;

            // Train on connect just to insure camera is trained
            ForceTraining(this);

            if (errorEncountered)
            {
                MessageBox.Show("Errors encountered communicating with the Camera:" + errorText, "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Device Specific actions when connection is established
        /// </summary>
        protected override unsafe void OnConnectSimple()
        {
            var cap = new CAMERACAP();
            uint caplength = (uint)sizeof(CAMERACAP);
            int result = NativeMethods.GetCapabilities(Handle, out cap, ref caplength);

            if (result < 0)
            {
                MessageBox.Show("Failed to get the Sensor Capabilities from the Camera.", "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }
            // Need to add a check here for Version and Size here
            else if (((cap.CapVersion == CAPABILITIES_VERSION_1) && (caplength != CAPABILITIES_LENGTH_VERSION_1)) ||
                ((cap.CapVersion == CAPABILITIES_VERSION_2) && (caplength != CAPABILITIES_LENGTH_VERSION_2)))
            {
                MessageBox.Show("Camera did not return the correct Sensor Capabilities size:\n TargetSize = " +
                    (cap.CapVersion == CAPABILITIES_VERSION_1 ? CAPABILITIES_LENGTH_VERSION_1.ToString() : CAPABILITIES_LENGTH_VERSION_2.ToString()) +
                    ", returnedSize = " + caplength.ToString(), "Warning", MessageBoxButton.OK,
                    MessageBoxImage.Warning, MessageBoxResult.OK, MessageBoxOptions.DefaultDesktopOnly);
                return;
            }

            if (CameraFirmwareIncompatable())
            {
                return;
            }

            // Parse the capabilities struct
            ParseCapabilities(cap, caplength);

            FrameCountType = FliFrameCount.SINGLE;
            FrameType = FliFrameType.NORMAL;
            ShutterOpen = false;
            LEDOn = false;

            VisibleArea = GetVisibleArea();
            ImageArea = VisibleArea;
        }

        protected ObservableCollection<SensorGain> GetSensorGainTable(GainTable tabletype, uint count)
        {
            var gainarray = new FliGainValue[count];
            var gainlist = new List<SensorGain>();
            if (IsConnected)
            {
                NativeMethods.FPROSensor_GetGainTableEx(Handle, tabletype, gainarray, ref count);

                for (int i = 0; i < gainarray.Length; i++)
                {
                    // This code is added temporarily for the 4040 to filter out all but one gain value
                    if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040.GetHashCode())
                    {
                        if (IsValid4040GainTableValue(tabletype, gainarray[i]))
                        {
                            gainlist.Add(new SensorGain(gainarray[i].uiDeviceIndex, gainarray[i].uiValue, i));
                        }
                    }
                    else
                    {
                        gainlist.Add(new SensorGain(gainarray[i].uiDeviceIndex, gainarray[i].uiValue, i));
                    }
                }
                gainlist = gainlist.OrderBy(g => g.Gain).ToList();
            }
            return new ObservableCollection<SensorGain>(gainlist);
        }

        protected bool IsValid4040GainTableValue(GainTable tabletype, FliGainValue gain)
        {
            return
                ((tabletype == GainTable.LOW) && (gain.uiValue == 2800)) ||
                ((tabletype == GainTable.LOW) && (gain.uiValue == 5200)) ||
                ((tabletype == GainTable.LOW) && (gain.uiValue == 6600)) ||
                ((tabletype == GainTable.LOW) && (gain.uiValue == 14000)) ||
                ((tabletype == GainTable.HIGH) && (gain.uiValue == 13500)) ||
                ((tabletype == GainTable.HIGH) && (gain.uiValue == 16500));
        }

        protected unsafe void ParseCapabilities(CAMERACAP cap, uint length)
        {
            //-=-= TODO: Check struct length and version

            // Supported Binning ratios
            BinRatioCount = cap.BinRatioTableSize / 2; // There are two values per BinRatio
            SupportedBinRatios.Clear();
            if (BinRatioCount > 0)
            {
                SupportedBinRatios = GetBinningTable((int)cap.BinRatioTableSize);
                if (SupportedBinRatios.Count == 0)
                {
                    // 1:1 is always supported
                    BinRatioCount = 1;
                    SupportedBinRatios.Add(new BinRatio(0, 1, 1));
                }
            }
            else
            {
                // 1:1 is always supported
                BinRatioCount = 1;
                SupportedBinRatios.Add(new BinRatio(0, 1, 1));
            }

            // Supported Pixel Depths
            uint bitmask = 1;
            MaxPixelDepth = 1;
            SupportedPixelDepths.Clear();
            for (uint i = 1; i <= MAX_PIXEL_DEPTH; i++)
            {
                if ((cap.AvailablePixelDepths & bitmask) != 0)
                {
                    SupportedPixelDepths.Add(i);
                    if (i > MaxPixelDepth) MaxPixelDepth = i;
                }
                bitmask <<= 1;
            }

            DeviceType = (int)cap.DeviceType;

            if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE400.GetHashCode())
            {
                SetDefaultGainTablesKepler0400(true);
            }
            else if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040.GetHashCode())
            {
                SetDefaultGainTablesKepler4040(true);
            }

            // For the 4040 we now have binning so enable it here
            if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040.GetHashCode())
            {
                MaxHBin = 1;
                MaxVBin = 1;
                for (int i = 0; i < SupportedBinRatios.Count; i++)
                {
                    MaxHBin = Math.Max(MaxHBin, SupportedBinRatios[i].HBin);
                    MaxVBin = Math.Max(MaxVBin, SupportedBinRatios[i].VBin);
                }
            }
            else
            {
                // For all others only allow 1:1 binning
                MaxHBin = 1;
                MaxVBin = 1;
            }

            //Set to 1:1 binning
            HBin = 1;
            VBin = 1;
            BinningSquared = true;

            MaxImageHeight = (int)cap.MaxPixelImageHeight;
            MaxImageWidth = (int)cap.MaxPixelImageWidth;
            MaxBlackLevel = cap.BlackLevelMax;
            MaxBlackSun = cap.BlackSunMax;
            LowGainCount = cap.LowGain;
            HighGainCount = cap.HighGain;
            RowScanTime = cap.RowScanTime;
            HInverseScanSupported = cap.HScanInvertable;
            VInverseScanSupported = cap.VScanInvertable;
            MetaDataSize = (int)cap.MetaDataSize;

            if ((DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()))
            {
                MetaDataSize = 384;
                MaxPixelDepth = 16;
                SupportedPixelDepths.Clear();
                SupportedPixelDepths.Add(16);
            }

            if (MetaDataSize == 0)
            {
                MetaDataSize = FliProFrameMetadata.META_DATA_LENGTH_VERSION_1;
            }
        }

        protected override void SetImmediate(bool value)
        {
            SetExposure(ExposureTime, ExposureDelayTime, ExposureUnits, ExposureDelayUnits, value);
        }

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="frameCount">number of frames to capture. 0=no limit</param>
        /// <returns>false on any error</returns>
        protected override bool SetStartCapture(int frameCount)
        {
            bool ok = false;
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return ok;
                }

                // We only want to call FPROFrame_CaptureStart for Internal trigger
                int result = NativeMethods.FPROFrame_CaptureStart(Handle, (uint)frameCount);
                ok = (result >= 0);
            }
            return ok;
        }

        //-------------------------------------------
        /// <summary>
        /// Stop current frame capture sequence, if any)
        /// </summary>
        protected override void SetStopCapture()
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROFrame_CaptureStop(Handle);
            }
        }

        protected override void SetHighGainIndex(int value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROSensor_SetGainIndex(Handle, GainTable.HIGH, (uint)value);
                }
            }
        }

        protected override void SetLowGainIndex(int value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROSensor_SetGainIndex(Handle, GainTable.LOW, (uint)value);
                }
            }
        }

        protected override void SetLEDOn(bool on)
        {
            //ulong oldRegisterValue = 0;

            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROCtrl_SetLED(Handle, on);

                if (on)
                {
                    SetLEDOnDuration(uint.MaxValue);
                }
                else
                {
                    SetLEDOnDuration(0);
                }

                OnLEDAction(new EventArgs());
            }
        }

        protected override void SetLEDOnDuration(uint uiDuration)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROCtrl_SetLEDDuration(Handle, uiDuration);
            }
        }

        protected override void SetIlluminationOn(bool on)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetIlluminationOn(Handle, on);
                }
            }
        }

        protected override void SetIlluminationDelay(int ondelay, int offdelay)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetIlluminationDelay(Handle, (ushort)ondelay, (ushort)offdelay);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Set the Vertical and Horizontal Scan directions
        /// </summary>
        /// <param name="hInverted">Flag to set the horizontal scan direction</param>
        /// <param name="vInverted">Flag to set the vertical scan direction</param>
        protected override void SetScanDirection(bool hInverted, bool vInverted)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROSensor_SetScanDirection(Handle, hInverted, vInverted);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Open/Close Shutter
        /// </summary>
        /// <param name="value"></param>
        public override void SetShutterOpen(bool open)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                // First override the normal shutter control
                NativeMethods.FPROCtrl_SetShutterOverride(Handle, true);
                NativeMethods.FPROCtrl_SetShutterOpen(Handle, open);
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Override the normal Shutter control
        /// </summary>
        /// <param name="value"></param>
        public override void SetShutterOverride(bool value)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                NativeMethods.FPROCtrl_SetShutterOverride(Handle, value);
            }
        }

        protected override void SetReadDuringExposure(bool enable)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROCtrl_SetSensorTemperatureReadEnable(Handle, enable);
                }
            }
        }

        public override void SetTrainingEnable(bool enable)
        {
            if (IsConnected)
            {
                if (Handle < 0)
                {
                    IsAborted = true;
                    return;
                }
                if (!gb_PopulatingUI)
                {
                    NativeMethods.FPROSensor_SetTrainingEnable(Handle, enable);
                }
            }
        }

        #endregion Protected Methods

    }

}
