﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Runtime;
using System.Text;
using System.Threading;
using System.Windows;

namespace FliProCameraLib
{
    /// <summary>
    /// This file contains Frame related properties and methods
    /// for the FliProCamera class
    /// 2016/09/17
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        public enum PurgeActionEnum
        {
            WaitPurge,
            ForcePurge,
            FinalPurge
        }

        public const ushort MAX_12_BIT_VALUE = 4095;
        public const ushort MAX_12_BIT_VALUE_PLUS_ONE = 4096;
        public const ushort COBALT_PRESCAN_PIXELS = 50;

        #endregion Constants and Enums

        //**************************************************************
        #region Public Classes

        public class FrameListEventArgs : EventArgs
        {
            public FrameListEventArgs() { IsInsert = false; }

            public FrameListEventArgs(bool isInsert)
            {
                IsInsert = isInsert;
            }

            public bool IsInsert { get; set; }
        }

        #endregion Public Classes

        //**************************************************************
        #region Protected Properties

        private List<FliFrame> WriteFrameList = new List<FliFrame>();

        #endregion Protected Properties

        //**************************************************************
        #region Public Properties
        public int FramePoolCount { get { return FramePool.Count; } }

        public int FramePoolBufSize { get; protected set; }

        public int TrackingFramePoolCount { get { return TrackingFramePool.Count; } }

        public int TrackingFramePoolBufSize { get; protected set; }

        public int FrameDataBytes
        {
            get { return PixelsToBytes(MaxImageWidth, ImageAreaWH.H); }
        }

        #endregion Public Properties

        //**************************************************************
        #region Observable Collections
        #endregion Observable Collections

        //**************************************************************
        #region Observable Properties

        public bool AccumulateTrackingFrameData
        {
            get { return _AccumulateTrackingFrameData; }
            set { SetValue<bool>(ref _AccumulateTrackingFrameData, value); }
        }
        protected bool _AccumulateTrackingFrameData;

        //-------------------------------------------
        /// <summary>
        /// The electrons per ADU ratio
        /// </summary>
        public double ADUValue
        {
            get { return _ADUValue; }
            set { SetValue(ref _ADUValue, value); }
        }
        private double _ADUValue = 1.1;

        //-------------------------------------------
        /// <summary>
        /// The electrons per ADU ratio
        /// </summary>
        public double ADUsPerElectron
        {
            get { return _ADUsPerElectron; }
            set
            {
                if (SetValue(ref _ADUsPerElectron, value))
                {
                    OnADUsPerElectronChangeAction(new EventArgs());
                }
            }
        }
        private double _ADUsPerElectron = 1.1;

        /// <summary>
        /// Used to calculate the Binning pixels when binning values are > 1
        /// </summary>
        public FliFrame CurrentBiasFrame
        {
            get { return _CurrentBiasFrame; }
            set { SetValue<FliFrame>(ref _CurrentBiasFrame, value); }
        }
        protected FliFrame _CurrentBiasFrame = null;

        public FliFrame CurrentFrame
        {
            get { return _CurrentFrame; }
            set { SetValue<FliFrame>(ref _CurrentFrame, value); }
        }
        protected FliFrame _CurrentFrame = null;

        public int CurrentFrameIndex
        {
            get { return _CurrentFrameIndex; }
            set
            {
                bool startDisplayProcess = false;
                lock (FrameListLock)
                {
                    if ((value > 0) && (value <= FrameList.Count))
                    {
                        SetValue<int>(ref _CurrentFrameIndex, value);
                        if (CurrentFrame != FrameList[value - 1])
                        {
                            CurrentFrame = FrameList[value - 1];
                            if (CurrentFrame.IsCapturedAndProcessed)
                            {
                                // Only start the display process if the image data is complete
                                startDisplayProcess = true;
                            }
                        }
                    }
                }
                if (startDisplayProcess)
                {
                    // Only start the display process if the image data is complete
                    if (fliProCameraManager == null)
                    {
                        fliProCameraManager = FliProCameraManager.Instance;
                    }
                    fliProCameraManager.OnActiveCameraAction(new FliProCameraManager.ActiveCameraEventArgs(this));
                }
            }
        }
        protected int _CurrentFrameIndex = 0;

        public bool FastThroughput
        {
            get { return _FastThroughput; }
            set { SetValue<bool>(ref _FastThroughput, value); }
        }
        protected bool _FastThroughput = false;

        public int FrameCount
        {
            get { return _FrameCount; }
            set { SetValue<int>(ref _FrameCount, value); }
        }
        protected int _FrameCount = 1;

        //-------------------------------------------
        /// <summary>
        ///  List of Frames
        /// </summary>
        public List<FliFrame> FrameList
        {
            get { return _FrameList; }
            set { SetValue<List<FliFrame>>(ref _FrameList, value); }
        }
        protected List<FliFrame> _FrameList = new List<FliFrame>();

        //-------------------------------------------
        /// <summary>
        /// Number of frames captured during current sequence
        /// </summary>
        public uint FramesCaptured
        {
            get { return _FramesCaptured; }
            protected set { SetValue<uint>(ref _FramesCaptured, value); }
        }
        protected uint _FramesCaptured;

        public uint FrameTimeout
        {
            get { return _FrameTimeout; }
            set { SetValue<uint>(ref _FrameTimeout, value); }
        }
        protected uint _FrameTimeout;

        //-------------------------------------------
        /// <summary>
        /// Frame Type
        /// </summary>
        public FliFrameCount FrameCountType
        {
            get { return _FrameCountType; }
            set
            {
                if (SetValue<FliFrameCount>(ref _FrameCountType, value))
                {
                    SetFrameCountType(value);
                }
            }
        }
        FliFrameCount _FrameCountType = FliFrameCount.SINGLE;

        //-------------------------------------------
        /// <summary>
        /// Frame Type
        /// </summary>
        public FliFrameType FrameType
        {
            get { return _FrameType; }
            set
            {
                if (SetValue<FliFrameType>(ref _FrameType, value))
                {
                    SetFrameType(value);
                }
            }
        }
        FliFrameType _FrameType = FliFrameType.NORMAL;

        //-------------------------------------------
        /// <summary>
        /// The Gain Ratio
        /// </summary>
        public double GainRatio
        {
            get { return _GainRatio; }
            set
            {
                if (SetValue(ref _GainRatio, value))
                {
                    OnGainRatioChangeAction(new EventArgs());
                }
            }
        }
        private double _GainRatio = 44.94;

        //-------------------------------------------
        /// <summary>
        /// The parameter TH (DN) is usually set to 3800 DN which is the maximum value of 
        /// the linear region of the response curve for high gain image.
        /// </summary>
        public int HighLowThreshold
        {
            get { return _HighLowThreshold; }
            set { SetValue(ref _HighLowThreshold, value); }
        }
        private int _HighLowThreshold = 3800;

        //-------------------------------------------
        /// <summary>
        /// Active Histogram area
        /// </summary>
        public FliRectangle HistogramArea
        {
            get { return _HistogramArea; }
            set
            {
                if (SetValue<FliRectangle>(ref _HistogramArea, value))
                {
                    OnPropertyChanged("HistogramUpperLeftX");
                    OnPropertyChanged("HistogramUpperLeftY");
                    OnPropertyChanged("HistogramWidth");
                    OnPropertyChanged("HistogramHeight");
                }
            }
        }
        protected FliRectangle _HistogramArea;

        //-------------------------------------------
        /// <summary>
        /// Active Histogram area
        /// </summary>
        public FliRectangleWH HistogramAreaWH
        {
            get { return new FliRectangleWH(_HistogramArea); }
            set { HistogramArea = new FliRectangle(value); }
        }

        /// <summary>
        /// Active Histogram Area upper left X offset
        /// </summary>
        public int HistogramUpperLeftX
        {
            get { return HistogramArea.UL.X; }
            set
            {
                int width = HistogramAreaWH.W;
                int newValue = value;
                if (value > MaxImageWidth)
                {
                    newValue = 0;
                    width = MaxImageWidth;
                }

                // Keep the width the same if we can. If not simply make the lrx the right column
                int lrx = (newValue + width > MaxImageWidth ? MaxImageWidth - 1 : newValue + width - 1);
                HistogramAreaWH = new FliRectangleWH(newValue, HistogramAreaWH.UL.Y, lrx - newValue + 1, HistogramAreaWH.H);
                OnPropertyChanged("HistogramWidth");
            }
        }

        /// <summary>
        /// Active Histogram Area upper left Y offset
        /// </summary>
        public int HistogramUpperLeftY
        {
            get { return HistogramArea.UL.Y; }
            set
            {
                int height = HistogramAreaWH.H;
                int newValue = value;
                if (value > MaxImageHeight)
                {
                    newValue = 0;
                    height = MaxImageHeight;
                }

                // Keep the height the same if we can. If not simply make the lry the bottom row
                int lry = (newValue + height > MaxImageHeight ? MaxImageHeight - 1 : newValue + height - 1);
                HistogramAreaWH = new FliRectangleWH(HistogramAreaWH.UL.X, newValue, HistogramAreaWH.W, lry - newValue + 1);
                OnPropertyChanged("HistogramHeight");
            }
        }

        public int HistogramWidth
        {
            get { return HistogramAreaWH.W; }
            set
            {
                int ulx = HistogramAreaWH.UL.X;
                if (value + HistogramAreaWH.UL.X > MaxImageWidth)
                {
                    if (value > MaxImageWidth)
                    {
                        ulx = 0;
                        value = MaxImageWidth;
                    }
                    else
                    {
                        ulx = MaxImageWidth - value;
                    }
                }
                HistogramAreaWH = new FliRectangleWH(ulx, HistogramAreaWH.UL.Y, value, HistogramAreaWH.H);
                OnPropertyChanged("HistogramWidth");
                OnPropertyChanged("HistogramUpperLeftX");
            }
        }

        public int HistogramHeight
        {
            get { return HistogramAreaWH.H; }
            set
            {
                int uly = HistogramAreaWH.UL.Y;
                if (value + HistogramAreaWH.UL.Y > MaxImageHeight)
                {
                    if (value > MaxImageHeight)
                    {
                        uly = 0;
                        value = MaxImageHeight;
                    }
                    else
                    {
                        uly = MaxImageHeight - value;
                    }
                }
                HistogramAreaWH = new FliRectangleWH(HistogramAreaWH.UL.X, uly, HistogramAreaWH.W, value);
                OnPropertyChanged("HistogramHeight");
                OnPropertyChanged("HistogramUpperLeftY");
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Active image area
        /// </summary>
        public FliRectangle ImageArea
        {
            get { return _ImageArea; }
            set
            {
                if( SetValue<FliRectangle>(ref _ImageArea, value))
                {
                    SetImageArea(value);
                    OnPropertyChanged("ImageAreaWH");
                    OnPropertyChanged("ImageUpperLeftX");
                    OnPropertyChanged("ImageUpperLeftY");
                    OnPropertyChanged("ImageWidth");
                    OnPropertyChanged("ImageHeight");
                    OnPropertyChanged("IsEnabledArea");
                }
            }
        }
        protected FliRectangle _ImageArea;

        //-------------------------------------------
        /// <summary>
        /// Active Image area
        /// </summary>
        public FliRectangleWH ImageAreaWH
        {
            get { return new FliRectangleWH(_ImageArea); }
            set { ImageArea = new FliRectangle(value); }
        }

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable actual image data in frame
        /// </summary>
        public bool ImageDataEnabled
        {
            get { return _ImageDataEnabled; }
            set
            {
                if (SetValue<bool>(ref _ImageDataEnabled, value) && IsConnected)
                {
                    SetEnableImageData(value);
                }
            }
        }
        protected bool _ImageDataEnabled = true;

        /// <summary>
        /// Active Image Area upper left X offset
        /// </summary>
        public int ImageUpperLeftX
        {
            get { return ImageArea.UL.X; }
            set
            {
                int width = ImageAreaWH.W;
                int newValue = value;
                if (value > MaxImageWidth)
                {
                    newValue = 0;
                    width = MaxImageWidth;
                }

                // Keep the width the same if we can. If not simply make the lrx the right column
                int lrx = (newValue + width > MaxImageWidth ? MaxImageWidth - 1 : newValue + width - 1);
                ImageAreaWH = new FliRectangleWH(newValue, ImageAreaWH.UL.Y, lrx - newValue + 1, ImageAreaWH.H);
                OnPropertyChanged("ImageWidth");
            }
        }

        /// <summary>
        /// Active Image Area upper left Y offset
        /// </summary>
        public int ImageUpperLeftY
        {
            get { return ImageArea.UL.Y; }
            set
            {
                int height = ImageAreaWH.H;
                int newValue = value;
                if (value > MaxImageHeight)
                {
                    newValue = 0;
                    height = MaxImageHeight;
                }

                // Keep the height the same if we can. If not simply make the lry the bottom row
                int lry = (newValue + height > MaxImageHeight ? MaxImageHeight - 1 : newValue + height - 1);  
                ImageAreaWH = new FliRectangleWH(ImageAreaWH.UL.X, newValue, ImageAreaWH.W, lry - newValue + 1);
                OnPropertyChanged("ImageHeight");
            }
        }

        /// <summary>
        /// Active Image Area lower right X
        /// </summary>
        public int ImageLowerRightX
        {
            get { return ImageArea.LR.X; }
            set
            {
                int ulx = (value < ImageArea.UL.X ? value : ImageArea.UL.X);
                ImageArea = new FliRectangle(ulx, ImageArea.UL.Y, value, ImageArea.LR.Y);
            }
        }

        /// <summary>
        /// Active Image Area lower right Y
        /// </summary>
        public int ImageLowerRightY
        {
            get { return ImageArea.LR.Y; }
            set
            {
                int uly = (value < ImageArea.UL.Y ? value : ImageArea.UL.Y);
                ImageArea = new FliRectangle(ImageArea.UL.X, uly, ImageArea.LR.X, value);
            }
        }

        public int ImageWidth
        {
            get { return ImageAreaWH.W; }
            set
            {
                int ulx = ImageAreaWH.UL.X;
                if ((value + ImageAreaWH.UL.X > MaxImageWidth) && (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) &&
                    (DeviceType != FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()))
                {
                    if (value > MaxImageWidth)
                    {
                        ulx = 0;
                        value = MaxImageWidth;
                    }
                    else
                    {
                        ulx = MaxImageWidth - value;
                    }
                }
                ImageAreaWH = new FliRectangleWH(ulx, ImageAreaWH.UL.Y, value, ImageAreaWH.H);
                OnPropertyChanged("ImageWidth");
                OnPropertyChanged("ImageUpperLeftX");
                OnPropertyChanged("IsEnabledArea");
            }
        }

        public int ImageHeight
        {
            get { return ImageAreaWH.H; }
            set
            {
                int uly = ImageAreaWH.UL.Y;
                if (value + ImageAreaWH.UL.Y > MaxImageHeight)
                {
                    if (value > MaxImageHeight)
                    {
                        uly = 0;
                        value = MaxImageHeight;
                    }
                    else
                    {
                        uly = MaxImageHeight - value;
                    }
                }
                ImageAreaWH = new FliRectangleWH(ImageAreaWH.UL.X, uly, ImageAreaWH.W, value);
                OnPropertyChanged("ImageHeight");
                OnPropertyChanged("ImageUpperLeftY");
            }
        }

        public FliFrame LastFrameQueued
        {
            get { return _LastFrameQueued; }
            set { SetValue<FliFrame>(ref _LastFrameQueued, value); }
        }
        protected FliFrame _LastFrameQueued = null;
        
        //-------------------------------------------
        /// <summary>
        /// Actual dimensions of frame data as top left and width/height
        /// Deprecated
        public FliRectangleWH LastReadoutAreaWH
        {
            get { return _LastReadoutAreaWH; }
            protected set
            {
                if (SetValue<FliRectangleWH>(ref _LastReadoutAreaWH, value))
                {
                    OnPropertyChanged("LastReadoutArea");
                }
            }
        }
        protected FliRectangleWH _LastReadoutAreaWH;

        /// <summary>
        /// Actual dimensions of frame data as top left, lower right
        /// Deprecated
        public FliRectangle LastReadoutArea
        {
            get { return new FliRectangle(_LastReadoutAreaWH); }
        }

         public int MergeRange1Count
        {
            get { return _MergeRange1Count; }
            set { SetValue<int>(ref _MergeRange1Count, value); }
        }
        protected int _MergeRange1Count;

        public int MergeRange2Count
        {
            get { return _MergeRange2Count; }
            set { SetValue<int>(ref _MergeRange2Count, value); }
        }
        protected int _MergeRange2Count;

        public int MergeRange3Count
        {
            get { return _MergeRange3Count; }
            set { SetValue<int>(ref _MergeRange3Count, value); }
        }
        protected int _MergeRange3Count;

        public int MergeRange1Lower
        {
            get { return _MergeRange1Lower; }
            set { SetValue<int>(ref _MergeRange1Lower, value); }
        }
        protected int _MergeRange1Lower;

        public int MergeRange2Lower
        {
            get { return _MergeRange2Lower; }
            set { SetValue<int>(ref _MergeRange2Lower, value); }
        }
        protected int _MergeRange2Lower;

        public int MergeRange3Lower
        {
            get { return _MergeRange3Lower; }
            set { SetValue<int>(ref _MergeRange3Lower, value); }
        }
        protected int _MergeRange3Lower;

        public int MergeRange1Upper
        {
            get { return _MergeRange1Upper; }
            set { SetValue<int>(ref _MergeRange1Upper, value); }
        }
        protected int _MergeRange1Upper;

        public int MergeRange2Upper
        {
            get { return _MergeRange2Upper; }
            set { SetValue<int>(ref _MergeRange2Upper, value); }
        }
        protected int _MergeRange2Upper;

        public int MergeRange3Upper
        {
            get { return _MergeRange3Upper; }
            set { SetValue<int>(ref _MergeRange3Upper, value); }
        }
        protected int _MergeRange3Upper;

        /// <summary>
        /// The number of frames to capture when FrameType is MULTIPLE
        /// </summary>
        public int MultipleFrameCount
        {
            get { return _MultipleFrameCount; }
            set
            {
                SetValue<int>(ref _MultipleFrameCount, value);
            }
        }
        int _MultipleFrameCount = 1;

        public int ProcessedFrameThroughputCount
        {
            get { return _ProcessedFrameThroughputCount; }
            set { SetValue<int>(ref _ProcessedFrameThroughputCount, value); }
        }
        protected int _ProcessedFrameThroughputCount = 0;

        public PurgeActionEnum ProcessingForcePurge { get; set; }
        public bool ProcessingSkippedFiles { get; set; }

        /// <summary>
        /// The offset between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset
        {
            get { return _LineOffset; }
            set
            {
                if (SetValue(ref _LineOffset, value))
                {
                    OnLineOffsetChangeAction(new EventArgs());
                }
            }
        }
        private int _LineOffset = 8000;

        /// <summary>
        /// The offset at the first temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset1
        {
            get { return _LineOffset1; }
            set { SetValue(ref _LineOffset1, value); }
        }
        private int _LineOffset1 = 1700;

        /// <summary>
        /// The offset at the second temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset2
        {
            get { return _LineOffset2; }
            set { SetValue(ref _LineOffset2, value); }
        }
        private int _LineOffset2 = 1750;

        /// <summary>
        /// The offset at the third temperature Merge settings between the HDR slope and the LDR slope
        /// </summary>
        public int LineOffset3
        {
            get { return _LineOffset3; }
            set { SetValue(ref _LineOffset3, value); }
        }
        private int _LineOffset3 = 1800;

        //-------------------------------------------
        /// <summary>
        /// The reference columns are only setable on the legacy cameras and the Cobalt cameras
        /// </summary>
        public bool ReferenceColumnsVisible
        {
            get
            {
                return IsLegacyCamera || (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                                         (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode());
            }
        }

        //-------------------------------------------
        /// <summary>
        /// The First Temperature to set default Merge settings
        /// </summary>
        public double Temperature1
        {
            get { return _Temperature1; }
            set { SetValue(ref _Temperature1, value); }
        }
        private double _Temperature1 = -10;

        //-------------------------------------------
        /// <summary>
        /// The Second Temperature to set default Merge settings
        /// </summary>
        public double Temperature2
        {
            get { return _Temperature2; }
            set { SetValue(ref _Temperature2, value); }
        }
        private double _Temperature2 = 5;

        //-------------------------------------------
        /// <summary>
        /// The Third Temperature to set default Merge settings
        /// </summary>
        public double Temperature3
        {
            get { return _Temperature3; }
            set { SetValue(ref _Temperature3, value); }
        }
        private double _Temperature3 = 20;

        //-------------------------------------------
        /// <summary>
        /// The Active Electrons Per Count High to set default Merge settings
        /// </summary>
        public double ActiveElectronsPerCountHigh
        {
            get { return _ActiveElectronsPerCountHigh; }
            set { SetValue(ref _ActiveElectronsPerCountHigh, value); }
        }
        private double _ActiveElectronsPerCountHigh;

        //-------------------------------------------
        /// <summary>
        /// The Active Electrons Per Count Low to set default Merge settings
        /// </summary>
        public double ActiveElectronsPerCountLow
        {
            get { return _ActiveElectronsPerCountLow; }
            set { SetValue(ref _ActiveElectronsPerCountLow, value); }
        }
        private double _ActiveElectronsPerCountLow;

        //-------------------------------------------
        /// <summary>
        /// The Active Manufacturer's Electrons Per Count Low to set default Merge settings
        /// </summary>
        public double ActiveManufactureElectronsPerCountHigh
        {
            get { return _ActiveManufactureElectronsPerCountHigh; }
            set { SetValue(ref _ActiveManufactureElectronsPerCountHigh, value); }
        }
        private double _ActiveManufactureElectronsPerCountHigh;

        //-------------------------------------------
        /// <summary>
        /// The Active Manufacturer's Electrons Per Count Low to set default Merge settings
        /// </summary>
        public double ActiveManufactureElectronsPerCountLow
        {
            get { return _ActiveManufactureElectronsPerCountLow; }
            set { SetValue(ref _ActiveManufactureElectronsPerCountLow, value); }
        }
        private double _ActiveManufactureElectronsPerCountLow;

        public FliRectangleWH RequestedImageAreaWH { get; set; }

        public List<FliRectangleWH> TargetImageAreaWHList = new List<FliRectangleWH>();

        #endregion Observable Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Add a frame to the frame list
        /// </summary>
        /// <param name="frame"></param>
        public void AddFrame(FliFrame frame, PurgeActionEnum forcePurge)
        {
            // Always put the newest frame at the start of the list so it gets displayed
            if (frame.IsQueued == false)
            {
                lock (FrameListLock)
                {
                    FrameList.Insert(0, frame);
                }
                OnFrameListChanged(new FrameListEventArgs(true));
                frame.IsQueued = true;
            }

            StartBackgroundFrameParsing(forcePurge);
        }

        //-------------------------------------------
        /// <summary>
        /// Empty the frame list
        /// </summary>
        public void ClearFrameList()
        {
            lock (FrameListLock)
            {
                FrameList.Clear();
            }
            OnFrameListChanged(new FrameListEventArgs(false));
        }

        //-------------------------------------------
        /// <summary>
        /// Purges unprocessed frames so that they do not interfere with new task being started
        /// </summary>
        public void DisgardUnprocessedFrames()
        {
            lock (FrameListLock)
            {
                for (int i = FrameList.Count - 1; i >= 0; i--)
                {
                    if (!FrameList[i].IsProcessed)
                    {
                        FrameList.RemoveAt(i);
                    }
                }
            }
            OnFrameListChanged(new FrameListEventArgs(false));
        }

        //-------------------------------------------
        /// <summary>
        /// Get the current frame in the frame list. This will be the last frame displayed during the slide show.
        /// </summary>
        /// <returns></returns>
        public FliFrame GetCurrentFrame()
        {
            FliFrame frame = null;

            lock (FrameListLock)
            {
                if ((CurrentFrameIndex > 0) && (CurrentFrameIndex <= FrameList.Count) && (FrameList[CurrentFrameIndex - 1].IsProcessed))
                {
                    frame = FrameList[CurrentFrameIndex - 1];
                }
                else
                {
                    for (int i = 0; i < FrameList.Count; i++)
                    {
                        if (FrameList[i].IsProcessed)
                        {
                            frame = FrameList[i];
                            break;
                        }
                    }
                }
            }
            return frame;
        }

        //-------------------------------------------
        /// <summary>
        /// Get the most recent frame in the frame list
        /// </summary>
        /// <returns></returns>
        public FliFrame GetLatestFrame()
        {
            FliFrame frame = null;

            lock (FrameListLock)
            {
                for (int i = 0; i < FrameList.Count; i++)
                {
                    if (FrameList[i].IsProcessed)
                    {
                        frame = FrameList[i];
                        break;
                    }
                }
            }
            return frame;
        }
        
        //-------------------------------------------
        /// <summary>
        /// Get the frame immediately following the given frame
        /// </summary>
        /// <returns></returns>
        public FliFrame GetNextFrame(FliFrame oldFrame)
        {
            FliFrame frame = null;

            lock (FrameListLock)
            {
                // Frames are added to the start so look for oldFrame from the end backwards
                for (int i = FrameList.Count - 1; i >= 0; i--)
                {
                    if (oldFrame == null)
                    {
                        // just return the first frame
                        frame = FrameList[i];
                        break;
                    }
                    if (FrameList[i] == oldFrame)
                    {
                        if (i > 0)
                        {
                            frame = FrameList[i - 1];
                        }
                        break;
                    }
                }
            }
            return frame;
        }


        //-------------------------------------------
        /// <summary>
        /// Get the second most recent processed frame in the frame list
        /// </summary>
        /// <returns></returns>
        public FliFrame GetPreviousFrame()
        {
            FliFrame frame = null;

            try
            {
                lock (FrameListLock)
                {
                    for (int i = 1; i < FrameList.Count; i++)
                    {
                        if (FrameList[i].IsProcessed)
                        {
                            frame = FrameList[i];
                            break;
                        }
                    }
                }
            }
            catch (Exception e)
            {
            }  // One time we had a race condition and frames were freed while in this loop.
               // In that case just return null.
            return frame;
        }

        //-------------------------------------------
        /// <summary>
        /// Deprecated ?
        /// </summary>
        /// <param name="rect"></param>
        /// <param name="hbin"></param>
        /// <param name="vbin"></param>
        public void GetReadoutDimensions(out FliRectangleWH rect, out uint hbin, out uint vbin)
        {
            rect = GetReadoutDimensions();
            hbin = HBin;
            vbin = VBin;
        }

        //-------------------------------------------
        /// <summary>
        /// Deprecated?
        /// </summary>
        /// <param name="width"></param>
        /// <param name="hoffset"></param>
        /// <param name="hbin"></param>
        /// <param name="height"></param>
        /// <param name="voffset"></param>
        /// <param name="vbin"></param>
        public void GetReadoutDimensions(out int width, out int hoffset, out uint hbin,
                                         out int height, out int voffset, out uint vbin)
        {
            GetReadoutDimensions();
            width = LastReadoutAreaWH.WH.W;
            height = LastReadoutAreaWH.WH.H;
            hoffset = LastReadoutAreaWH.UL.X;
            voffset = LastReadoutAreaWH.UL.Y;
            hbin = HBin;
            vbin = VBin;
        }

        //-------------------------------------------
        /// <summary>
        /// Deprecated?
        /// </summary>
        /// <param name="rect"></param>
        /// <param name="hbin"></param>
        /// <param name="vbin"></param>
        public void GetReadoutDimensions(out FliRectangle rect, out uint hbin, out uint vbin)
        {
            GetReadoutDimensions();
            rect = LastReadoutArea;
            hbin = HBin;
            vbin = VBin;
        }

        public void MergeRangeValuesUpdated()
        {
            OnMergeRangeChangeAction(new EventArgs());
        }

        //-------------------------------------------
        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold the specified number of pixels at the current
        /// camera settings
        /// </summary>
        /// <param name="pixelcount">number of pixels</param>
        /// <returns></returns>
        public virtual int PixelsToBytes(int pixelcount)
        {
            return 0;
        }

        /// <summary>
        /// Calculates the raw frame buffer size in bytes required
        /// to hold an image of the specified size at the current
        /// camera settings
        /// </summary>
        /// <param name="width">image width in pixels</param>
        /// <param name="height">image height in pixels</param>
        /// <returns></returns>
        public virtual int PixelsToBytes(int width, int height)
        {
            return 0;
        }

        public virtual int ReadThumbnail(long length, byte[] rbuf)
        {
            return -7;
        }

        public virtual int WriteThumbnail(long length, byte[] wbuf)
        {
            return -7;
        }

        public void ResetImageArea()
        {
            ImageArea = GetVisibleArea();
        }

        //-------------------------------------------
        /// <summary>
        /// Ensure the FramePool contains at least the
        /// minimum acceptable number of frames compatible
        /// with the current frame capture parameters.
        /// </summary>
        /// <param name="init">Clear and recreate frame pool if true</param>
        public void ReplenishFramePool(bool startOfGrab = false, bool init = false)
        {
            //int bufsize = PixelsToBytes(ImageAreaWH.W, ImageAreaWH.H);
            int imageWidth = ((DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                              (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()) ? ImageAreaWH.W : MaxImageWidth);
            int bufsize = PixelsToBytes(MaxImageWidth, ImageAreaWH.H);
            if (init || (FramePoolBufSize < bufsize))
            {
                FramePool = new ConcurrentQueue<FliFrame>();
            }

            try
            {
                while (FramePool.Count < MinimumFramePoolCount)
                {
                    FramePool.Enqueue(new FliFrame(this, bufsize));
                }
            }
            catch (Exception ex)
            {
                if (startOfGrab)
                {
                    MessageBox.Show("The memory on your computer has limited the number of available memory buffers to " + FramePool.Count.ToString() +
                        ".\nWe will use this number and attempt to complete the Grab.\nYou can reduce the number of frames kept in memory on the Camera Tab.\nError: " + ex.Message);
                }
            }
            FramePoolBufSize = bufsize;
        }

        public void SetCaptureComplete(FliFrame frame)
        {
            if (fliProCameraManager == null)
            {
                fliProCameraManager = FliProCameraManager.Instance;
            }
            frame.IsInUse = false;
            frame.IsCapturedAndProcessed = frame.IsProcessed;
        }

        public virtual FliRectangle GetSensorArea()
        {
            return new FliRectangle();
        }
        
        public virtual FliRectangle GetVisibleArea()
        {
            return new FliRectangle(0, 0, DEFAULT_IMAGE_WIDTH - 1, DEFAULT_IMAGE_HEIGHT - 1);
        }

        /// <summary>
        /// Start a Grab for this camera using the current parameter set
        /// </summary>
        public virtual void GrabFrame()
        {
            IsCapturing = true;
        }

        /// <summary>
        /// Start a Grab for this camera using the current parameter set with the frame count passed in
        /// </summary>
        public virtual void GrabFrame(FliFrameCount frameCount)
        {
            IsCapturing = true;
        }

        /// <summary>
        /// Grab one 8bpp frame from the camera
        /// </summary>
        /// <returns>Actual number of pixels grabbed</returns>
        public virtual uint GrabFrame(byte[,] buff)
        {
            return 0;
        }

        /// <summary>
        /// Grab one 16bpp frame from the camera
        /// </summary>
        /// <returns>number of pixels grabbed</returns>
        public virtual uint GrabFrame(ushort[,] buff)
        {
            return 0;
        }

        
        /// <summary>
        /// Grab one video frame from the camera
        /// Caller is responsible for pinning and freeing buffer
        /// </summary>
        public virtual void GrabVideoFrame(ushort[] BuffPtr, int BuffLength)
        {
        }

        /// <summary>
        /// Grab one video frame from the camera
        /// </summary>
        public virtual void GrabVideoFrame(byte[] buff)
        {
        }

        /// <summary>
        /// Grab one 8bpp video frame from the camera
        /// </summary>
        public virtual void GrabVideoFrame(byte[,] buff)
        {
        }

        /// <summary>
        /// Grab one 16bpp frame from the camera
        /// </summary>
        public virtual void GrabVideoFrame(ushort[,] buff)
        {
        }

        public virtual void SetVerticalTableEntry(int index, int height, int bin, int mode)
        {
        }

        public virtual void GetVerticalTableEntry(int index, out int height, out int bin, out int mode)
        {
            height = 0; 
            bin = 0;   
            mode = 0;  
        }

        public virtual void EnableVerticalTable(int width, int offset, int flags)
        {
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        public FliFrame GetCompletedFrame()
        {
            int imageWidth = ((DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_42.GetHashCode()) ||
                              (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_DC230_84.GetHashCode()) ? ImageAreaWH.W : MaxImageWidth);
            int bufsize = PixelsToBytes(MaxImageWidth, ImageAreaWH.H);
            FliFrame frame = new FliFrame(this, bufsize);

            SetStartCapture(1);   // Simply calls FPROFrame_CaptureStart(Handle, 1);

            // Calculate the Exposure and frame dely time, then call FPROFrame_GetVideoFrame(Handle, bufPtr, ref size, (uint)ExposureTimeInMS);
            frame.RawDataLength = GetRawImageFrame(frame.RawData, bufsize, false);

            frame.IsValid = (frame.RawDataLength == FrameDataBytes);

            SetStopCapture();   // Simply calls FPROFrame_CaptureStop(Handle);

            int currentIndex = 0;
            frame.RawMetaData = new byte[FliProFrameMetadata.GetMetadataLengthFromArray(frame.RawData)];
            Buffer.BlockCopy(frame.RawData, 0, frame.RawMetaData, 0, frame.RawMetaData.Length);
            currentIndex = frame.RawMetaData.Length;


            // Parse the raw bytes into the metadata structure
            frame.FrameMetaData = new FliProFrameMetadata(frame.RawMetaData, frame.CaptureTime);
            if (frame.IsValid)
            {
                // Set the sizes so that the parsing function has all the needed details
                int totalHeight = (frame.FrameMetaData.PreDummyRows + (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1) + frame.FrameMetaData.PostDummyRows) /
                    frame.FrameMetaData.VerticalBinning;
                frame.FrameAreaWH = new FliPro.FliRectangleWH(0, 0,
                    (frame.FrameMetaData.ImageStopColumn - frame.FrameMetaData.ImageStartColumn + 1) / frame.FrameMetaData.HorizontalBinning,
                    (frame.FrameMetaData.ImageStopRow - frame.FrameMetaData.ImageStartRow + 1) / frame.FrameMetaData.VerticalBinning);
                frame.FrameAreaWHTotal = new FliPro.FliRectangleWH(0, 0,
                    frame.FrameAreaWH.W + (frame.FrameMetaData.DummyPixelsPerRow / frame.FrameMetaData.HorizontalBinning), totalHeight);
                frame.IsProcessed = true;  // Wait to set flag until metadata has been processed

                AddFrame(frame, FliProCamera.PurgeActionEnum.ForcePurge);
                CurrentFrame = frame;
                CurrentFrameIndex = 1;

            }

            return frame;
        }

        //-------------------------------------------
        /// <summary>
        /// Deprecated?
        /// </summary>
        /// <returns></returns>
        public virtual FliRectangleWH GetReadoutDimensions()
        {
            LastReadoutAreaWH = new FliRectangleWH();
            return LastReadoutAreaWH;
        }

        public virtual string GetDeviceType() { return String.Empty; }

        private FliRectangle GetRootedArea(int root)
        {
            FliRectangle visibleArea = GetVisibleArea();
            double power = 1 / Math.Pow(root, 0.5);
            int newWidth = (int)((visibleArea.LR.X - visibleArea.UL.X + 1) * power);
            int newHeight = (int)((visibleArea.LR.Y - visibleArea.UL.Y + 1) * power);
            int newX = ((visibleArea.LR.X - visibleArea.UL.X + 1) - newWidth) / 2;
            int newY = ((visibleArea.LR.Y - visibleArea.UL.Y + 1) - newHeight) / 2;
            FliRectangle newArea = new FliRectangle(newX, newY, newX + newWidth - 1, newY + newHeight - 1);
            return newArea;
        }

        private FliFrame GetUnprocessedFrame()
        {
            FliFrame frame = null;

            lock (FrameListLock)
            {
                for (int index = FrameList.Count - 1; index >= 0; index--)
                {
                    if (!FrameList[index].IsProcessed)
                    {
                        frame = FrameList[index];
                        break;
                    }
                }
            }

            return frame;
        }

        public void InsertWriteList(FliFrame frame)
        {
            lock (FrameListLock)
            {
                int index = 0;

                for (index = 0; index < WriteFrameList.Count; index++)
                {
                    if (FrameList[index].FrameMetaData.FrameNumber > frame.FrameMetaData.FrameNumber)
                    {
                        break;
                    }
                }
                FrameList.Insert(index, frame);
            }
        }

        public FliFrame PopWriteList()
        {
            FliFrame frame = null;

            lock (FrameListLock)
            {
                if (WriteFrameList.Count > 0)
                {
                    frame = WriteFrameList[0];
                    WriteFrameList.RemoveAt(0);
                }
            }
            return frame;
        }

        private void StartBackgroundFrameWriteThread()
        {

        }

        private void PurgeFrameList(PurgeActionEnum forcePurge)
        {
            bool changed = false;
            lock (FrameListLock)
            {
                // Want to work from the bottom of the list and free up any completed frames below the MaximumFrameStorage threshhold
                for (int i = FrameList.Count - 1; i >= MaximumFrameStorage; i--)
                {
                    // Don't purge the current frame or any frames during Test routines
                    if ((FrameList[i].IsCapturedAndProcessed || (forcePurge == PurgeActionEnum.ForcePurge)) &&
                        (FrameList[i] != CurrentFrame) && (!FrameList[i].IsInUse))
                    {
                        string frameNumber = (FrameList[i].FrameMetaData == null ? "??" :
                            FrameList[i].FrameMetaData.FrameNumber.ToString());
                        FliFrame frame = FrameList[i];
                        FrameList.RemoveAt(i);
                        if (FastThroughput)
                        {
                            // Put this buffer back on the FramePool
                            FramePool.Enqueue(new FliFrame(this, frame.RawData));
                        }
                        else
                        {
                            frame.DisposeFrame();
                            frame.Dispose();
                        }
                        changed = true;
                    }
                }
            }
            if (changed)
            {
                OnPropertyChanged("CancelBackgroundFileSaveText");
                OnFrameListChanged(new FrameListEventArgs(false));
            }
        }

        /// <summary>
        /// Allows the test routines to purge up to the latest frame if needed
        /// </summary>
        /// <param name="targetFrame"></param>
        public void PurgeFrameListTest(FliFrame targetFrame)
        {
            bool changed = false;
            lock (FrameListLock)
            {
                // Want to work from the bottom of the list and free up any completed frames below the MaximumFrameStorage threshhold
                for (int i = FrameList.Count - 1; i >= MaximumFrameStorage; i--)
                {
                    if (FrameList[i] == targetFrame)
                    {
                        // Don't purge any more frames
                        break;
                    }

                    // Don't purge the current frame
                    if ((FrameList[i].IsCapturedAndProcessed) && (FrameList[i] != CurrentFrame))
                    {
                        string frameNumber = (FrameList[i].FrameMetaData == null ? "??" :
                            FrameList[i].FrameMetaData.FrameNumber.ToString());
                        FliFrame frame = FrameList[i];
                        FrameList.RemoveAt(i);
                        frame.DisposeFrame();
                        frame.Dispose();
                        changed = true;
                    }
                }
            }
            if (changed)
            {
                OnFrameListChanged(new FrameListEventArgs(false));
            }
        }

        protected virtual void SetDummyPixelsEnable(bool value) { }
        protected virtual void SetEnableImageData(bool value) { }
        protected virtual void SetDummyRowPreFrameEnable(int preFrameDummyRowCount) { }
        protected virtual void SetDummyRowPostFrameEnable(int postFrameDummyRowCount) { }
        protected virtual void SetFrameCountType(FliFrameCount value) { }
        protected virtual void SetFrameType(FliFrameType value) { }
        public virtual void SetFrameTypeForced(FliFrameType value) { }

        public void StartBackgroundFrameParsing(PurgeActionEnum forcePurge)
        {
            // If we are processing the final purge, don't overwrite it
            if (!ProcessingSkippedFiles || (ProcessingForcePurge == PurgeActionEnum.WaitPurge))
            {
                ProcessingForcePurge = forcePurge;  // This will change on the last call
            }
            if (!ProcessingSkippedFiles)
            {
                ProcessingSkippedFiles = true;
                Thread newThread = new Thread(StartBackgroundFrameParsingThread);
                newThread.Start();
            }
        }

        private void StartBackgroundFrameParsingThread()
        {
            FliFrame frame = GetUnprocessedFrame();

            if (ProcessingForcePurge != PurgeActionEnum.ForcePurge)
            {
                while (frame != null)
                {
                    SetCaptureComplete(frame);
                    if (ProcessingForcePurge != PurgeActionEnum.WaitPurge)
                    {
                        // New images are no longer being generated so we can purge as we go.
                        // We need to do this to keep the system memory consumption at a minimum.
                        PurgeFrameList(ProcessingForcePurge);
                    }
                    frame = GetUnprocessedFrame();
                }
            }

            // Now the unprocessed images have been written to file so we can purge the excess
            PurgeFrameList(ProcessingForcePurge);

            if (ProcessingForcePurge != PurgeActionEnum.WaitPurge)
            {
                // All of the frames from this run have been processed and then purged so do the Garbage Collection 
                lock (FrameListLock)
                {
                    FrameList.TrimExcess();
                }
                OnFrameListChanged(new FrameListEventArgs(false));
                //GC.Collect();
            }
            ProcessingSkippedFiles = false;
        }

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="frameCount">number of frames to capture. 0=no limit</param>
        /// <returns>false on any error</returns>
        protected virtual bool SetStartCapture(int frameCount)
        {
            return false;
        }

        //-------------------------------------------
        /// <summary>
        /// Perform camera specific actions to stop current frame capture sequence, if any)
        /// </summary>
        protected virtual void SetStopCapture()
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Perform camera specific actions to set frame image area
        /// </summary>
        /// <param name="value"></param>
        protected virtual void SetImageArea(FliRectangle value)
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Perform camera specific actions to set frame image area
        /// </summary>
        /// <param name="value"></param>
        protected virtual void SetImageArea(FliRectangleWH value)
        {
        }

        //-------------------------------------------
        /// <summary>
        /// Perform camera specific actions to get one raw frame
        /// </summary>
        /// <param name="framebuf">buffer to receive frame data</param>
        /// <param name="framesize">length of frame in bytes</param>
        /// <returns>actual number of bytes received</returns>
        protected virtual uint GetRawImageFrame(byte[] framebuf, int framesize, bool useDelay)
        {
            throw (new Exception("Base GetRawImageFrame called!"));
        }

        #endregion Protected Methods

        //**************************************************************
        #region Public Events

        public event EventHandler<FrameListEventArgs> FrameListChanged;

        protected virtual void OnFrameListChanged(FrameListEventArgs e)
        {
            EventHandler<FrameListEventArgs> handler = FrameListChanged;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        #endregion Public Events

    }
}
