﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using FliUtils.Jobs;


namespace FliProCameraLib
{
    /// <summary>
    /// This file contains Sensor related properties and methods
    /// for the FliProCamera class
    /// 2016/09/17
    /// </summary>
    public partial class FliProCamera : FliPro
    {
        //**************************************************************
        #region Constants and Enums

        public const int MAX_LEGACY_SENSOR_BINNING_AVAILABLE = 255;

        #endregion Constants and Enums

        //**************************************************************
        #region Protected Properties
        #endregion Protected Properties

        //**************************************************************
        #region Public Properties
        public uint MaxPixelDepth { get; protected set; }

        public FliRectangle ArrayArea { get; protected set; }
        public FliRectangleWH ArrayAreaWH
        {
            get { return new FliRectangleWH(ArrayArea); }
            protected set { ArrayArea = new FliRectangle(value); }
        }

        public FliRectangle VisibleArea { get; protected set; }
        public FliRectangleWH VisibleAreaWH
        {
            get { return new FliRectangleWH(VisibleArea); }
            protected set { VisibleArea = new FliRectangle(value); }
        }

        public FliPixelSize PixelSize { get; protected set; }
        public FliBayerPattern ColorPattern { get; protected set; }

        public uint ModeCount { get; protected set; }
        public uint BinRatioCount { get; protected set; }
        public uint LowGainCount { get; protected set; }
        public uint HighGainCount { get; protected set; }
        public uint RowScanTime { get; protected set; }
        public bool HInverseScanSupported { get; protected set; }
        public bool VInverseScanSupported { get; protected set; }

        #endregion Public Properties

        //**************************************************************
        #region Observable Collections

        public ObservableCollection<uint> SupportedPixelDepths { get; protected set; }
        public ObservableCollection<BinRatio> SupportedBinRatios { get; protected set; }
        public ObservableCollection<NameValuePair> SupportedModes { get; protected set; }

        public ObservableCollection<SensorGain> LowGainTable
        {
            get { return _LowGainTable; }
            protected set { SetValue<ObservableCollection<SensorGain>>(ref _LowGainTable, value); }
        }
        protected ObservableCollection<SensorGain> _LowGainTable;

        public ObservableCollection<SensorGain> HighGainTable
        {
            get { return _HighGainTable; }
            protected set { SetValue<ObservableCollection<SensorGain>>(ref _HighGainTable, value); }
        }
        protected ObservableCollection<SensorGain> _HighGainTable;

        public ObservableCollection<FliFrameCountNameValueClass> FliFrameCountList { get; protected set; }
        public ObservableCollection<FliFrameTypeNameValueClass> FliFrameTypeList { get; protected set; }
        public ObservableCollection<SensorGain> DNETableLG { get; set; }     // Counts per electron Low Gain
        public ObservableCollection<SensorGain> DNETableHG { get; set; }     // Counts per electron High Gain
        public ObservableCollection<SensorGain> DNETableLGFactory { get; set; }     // Counts per electron Low Gain designated by the Manufacturer
        public ObservableCollection<SensorGain> DNETableHGFactory { get; set; }     // Counts per electron High Gain designated by the Manufacturer

        #endregion Observable Collections

        //**************************************************************
        #region Observable Properties

        public bool BinningTypeAverage
        {
            get { return _BinningTypeAverage; }
            set { SetValue<bool>(ref _BinningTypeAverage, value); }
        }
        protected bool _BinningTypeAverage = true;
        
        public bool BinningSquared
        {
            get { return _BinningSquared; }
            set
            {
                if (SetValue<bool>(ref _BinningSquared, value))
                {
                    if (value)
                    {
                        VBin = HBin;
                    }
                }
            }
        }
        protected bool _BinningSquared;

        public bool BinningTypeSum
        {
            get { return _BinningTypeSum; }
            set { SetValue<bool>(ref _BinningTypeSum, value); }
        }
        protected bool _BinningTypeSum;

        public BinRatio BinRatioValue
        {
            get { return _BinRatioValue; }
            set
            {
                if (SetValue(ref _BinRatioValue, value))
                {
                    if (value != null)
                    {
                        SetBinningIndex(value.Index);
                    }
                }
            }
        }
        private BinRatio _BinRatioValue;

        public string CameraMode
        {
            get { return _CameraMode; }
            set { SetValue<string>(ref _CameraMode, value); }
        }
        protected string _CameraMode = String.Empty;

        /// <summary>
        /// The Horizontal Dir value
        /// </summary>
        public bool HorizontalScanDirInverted
        {
            get { return _HorizontalScanDirInverted; }
            set
            {
                if (SetValue(ref _HorizontalScanDirInverted, value))
                {
                    SetScanDirection(value, VerticalScanDirInverted);
                }
            }
        }
        private bool _HorizontalScanDirInverted = false;

        public SensorGain HighGainValue
        {
            get { return _HighGainValue; }
            set
            {
                if (SetValue(ref _HighGainValue, value))
                {
                    HighGainIndex = value.Index;
                }
            }
        }
        private SensorGain _HighGainValue;

        public SensorGain LowGainValue
        {
            get { return _LowGainValue; }
            set
            {
                if (SetValue(ref _LowGainValue, value))
                {
                    LowGainIndex = value.Index;
                }
            }
        }
        private SensorGain _LowGainValue;

        public uint MaxHBin
        {
            get { return _MaxHBin; }
            set { SetValue<uint>(ref _MaxHBin, value); }
        }
        protected uint _MaxHBin = 1;

        public uint MaxVBin
        {
            get { return _MaxVBin; }
            set { SetValue<uint>(ref _MaxVBin, value); }
        }
        protected uint _MaxVBin = 1;

        //-------------------------------------------
        /// <summary>
        /// Most recent camera sensor temperature
        /// </summary>
        public double SensorTemperature
        {
            get { return _SensorTemperature; }
            protected set { SetValue<double>(ref _SensorTemperature, value); }
        }
        protected double _SensorTemperature = 0;

        //-------------------------------------------
        /// <summary>
        /// Camera pixel bit depth
        /// </summary>
        public virtual uint PixelDepth
        {
            get { return _PixelDepth; }
            set
            {
                if (SupportedPixelDepths.Contains(value) && SetValue<uint>(ref _PixelDepth, value))
                {
                    SetPixelDepth(value);
                }
            }
        }
        protected uint _PixelDepth = 12;

        //-------------------------------------------
        /// <summary>
        /// Get/Set the camera mode (clock speed, etc)
        /// </summary>
        public int Mode
        {
            get { return _Mode; }
            set
            {
                for (int i = 0; i < SupportedModes.Count; i++)
                {
                    // Make sure the mode exists before assigning it
                    if (value == SupportedModes[i].Value)
                    {
                        if (SetValue<int>(ref _Mode, value))
                        {
                            CameraMode = SupportedModes[i].Name;
                            GetHDREnable(ref IsHDREnable);
                            SetMode(value);

                            //if (!IsHDREnable)
                            //{
                            //    HighGainIndex = LowGainIndex;
                            //}
                        }
                        break;
                    }
                }
            }
        }
        protected int _Mode = 0;

        //-------------------------------------------
        /// <summary>
        /// Horizontal Bin Factor
        /// </summary>
        public uint HBin
        {
            get { return _HBin; }
            set
            {
                if (value < 1)
                {
                    value = 1;
                }
                else if (value > MaxHBin)
                {
                    value = MaxHBin;
                }

                if (SetValue<uint>(ref _HBin, value))
                {
                    if (BinningSquared)
                    {
                        VBin = HBin;
                    }
                    SetBinning();
                }

                if (_HBin > 1)
                {
                    ApplyChannelAveraging = false;
                }
            }
        }
        protected uint _HBin = 1;

        //-------------------------------------------
        /// <summary>
        /// Vertical Bin Factor
        /// </summary>
        public uint VBin
        {
            get { return _VBin; }
            set
            {
                if (value < 1) value = 1;
                else if (value > MaxVBin) value = MaxVBin;

                if (SetValue<uint>(ref _VBin, value))
                {
                    SetBinning();
                }
                if (_VBin > 1)
                {
                    ApplyChannelAveraging = false;
                }
            }
        }
        protected uint _VBin = 1;

        public bool UseSoftwareBinning
        {
            get { return _UseSoftwareBinning; }
            set
            {
                if (SetValue<bool>(ref _UseSoftwareBinning, value))
                {
                    if (value)
                    {
                        MaxHBin = (uint)MaxImageHeight;
                        MaxVBin = (uint)MaxImageWidth;
                    }
                    else
                    {
                        // For the 4040 we now have binning so enable it here
                        if (DeviceType == FPRODEVICETYPE.FPRO_CAM_DEVICE_TYPE_GSENSE4040.GetHashCode())
                        {
                            MaxHBin = 1;
                            MaxVBin = 1;
                            for (int i = 0; i < SupportedBinRatios.Count; i++)
                            {
                                MaxHBin = Math.Max(MaxHBin, SupportedBinRatios[i].HBin);
                                MaxVBin = Math.Max(MaxVBin, SupportedBinRatios[i].VBin);
                            }
                        }
                        else
                        {
                            // For all others only allow 1:1 binning
                            MaxHBin = 1;
                            MaxVBin = 1;
                        }
                        //Set to 1:1 binning
                        HBin = 1;
                        VBin = 1;
                        BinningSquared = true;
                        BinningTypeAverage = true;
                    }
                }
            }
        }
        protected bool _UseSoftwareBinning = false;

        /// <summary>
        /// The Vertical Dir value
        /// </summary>
        public bool VerticalScanDirInverted
        {
            get { return _VerticalScanDirInverted; }
            set
            {
                if (SetValue(ref _VerticalScanDirInverted, value))
                {
                    SetScanDirection(HorizontalScanDirInverted, value);
                }
            }
        }
        private bool _VerticalScanDirInverted = false;

        public uint BlackLevelHigh
        {
            get { return _BlackLevelHigh; }
            set
            {
                if (SetValue<uint>(ref _BlackLevelHigh, value))
                {
                    SetBlackLevel(value, true);
                }
            }
        }
        protected uint _BlackLevelHigh = 15800;

        public uint BlackLevelLow
        {
            get { return _BlackLevelLow; }
            set
            {
                if (SetValue<uint>(ref _BlackLevelLow, value))
                {
                    SetBlackLevel(value, false);
                }
            }
        }
        protected uint _BlackLevelLow = 15800;

        public uint BlackSunHigh
        {
            get { return _BlackSunHigh; }
            set
            {
                if (SetValue<uint>(ref _BlackSunHigh, value))
                {
                    SetBlackSun(value, true);
                }
            }
        }
        protected uint _BlackSunHigh = 28;

        public uint BlackSunLow
        {
            get { return _BlackSunLow; }
            set
            {
                if (SetValue<uint>(ref _BlackSunLow, value))
                {
                    SetBlackSun(value, false);
                }
            }
        }
        protected uint _BlackSunLow = 28;

        public virtual uint BinRatioIndex
        {
            get { return 0; }
            set { }
        }

        /// <summary>
        /// The flag to show if the selected Mode is HDR 
        /// </summary>
        public bool IsHDREnable = true;

        //-------------------------------------------
        /// <summary>
        /// Index of active LDR Gain
        /// </summary>
        public int LowGainIndex
        {
            get { return _LowGainIndex; }
            set
            {
                if (SetValue<int>(ref _LowGainIndex, value))
                {
                    SetLowGainIndex(value);
                    ActiveElectronsPerCountLow = GetGainValueFromIndex(DNETableLG, value).Gain;
                    ActiveManufactureElectronsPerCountLow = GetGainValueFromIndex(DNETableLGFactory, value).Gain;
                    if (!IsHDREnable)
                    {
                        HighGainIndex = LowGainIndex;
                    }
                    if (!RestoringMergeSettings)
                    {
                        UpdateMergeValues();
                    }
                }
            }
        }
        protected int _LowGainIndex = 0;

        public bool MergeValuesChangeActionRegistered
        {
            get { return _MergeValuesChangeActionRegistered; }
            set { SetValue<bool>(ref _MergeValuesChangeActionRegistered, value); }
        }
        protected bool _MergeValuesChangeActionRegistered = false;

        public bool RestoringMergeSettings
        {
            get { return _RestoringMergeSettings; }
            set { SetValue<bool>(ref _RestoringMergeSettings, value); }
        }
        protected bool _RestoringMergeSettings = false;

        //-------------------------------------------
        /// <summary>
        /// Index of active High Gain
        /// </summary>
        public int HighGainIndex
        {
            get { return _HighGainIndex; }
            set
            {
                if (SetValue<int>(ref _HighGainIndex, value))
                {
                    SetHighGainIndex(value);
                    ActiveElectronsPerCountHigh = GetGainValueFromIndex(DNETableHG, value).Gain;
                    ActiveManufactureElectronsPerCountHigh = GetGainValueFromIndex(DNETableHGFactory, value).Gain;
                    ADUsPerElectron = GetGainValueFromIndex(DNETableHG, value).Gain;

                    if (!RestoringMergeSettings)
                    {
                        UpdateMergeValues();
                    }
                }
            }
        }

        protected int _HighGainIndex = 0;

        public uint MaxBlackLevel = 16383;
        public uint MaxBlackSun = 63;

        //-------------------------------------------
        /// <summary>
        /// Camera pixel least significant bit position
        /// </summary>
        public virtual uint PixelLSB
        {
            get { return 0; }
            set { }
        }

        public virtual bool HDRMode
        {
            get { return false; }
            protected set { }
        }

        #endregion Observable Properties

        //**************************************************************
        #region Public Methods

        protected virtual ObservableCollection<BinRatio> GetBinningTable(int pTableSizeBytes)
        {
            ObservableCollection<BinRatio> binRatioList = new ObservableCollection<BinRatio>();
            return binRatioList;
        }

        public string GetModeString(int mode)
        {
            for (int i = 0; i < SupportedModes.Count; i++)
            {
                // Make sure the mode exists before assigning it
                if (mode == SupportedModes[i].Value)
                {
                    return SupportedModes[i].Name;
                }
            }
            return "Unknown(" + mode.ToString() + ")";
        }

        public void SetArrayAreaWH(FliRectangleWH areaWH)
        {
            ArrayAreaWH = areaWH;
        }

        #endregion Public Methods

        //**************************************************************
        #region Protected Methods

        protected virtual void SetBinningIndex(int value) { }
        protected virtual void SetHighGainIndex(int value) { }
        protected virtual void SetLowGainIndex(int value) { }
        protected virtual void SetPixelDepth(uint value) { }
        protected virtual void SetPixelLSB(uint value) { }
        protected virtual void SetMode(int value) { }
        protected virtual void SetBinning() { }
        protected virtual void SetBlackLevel(uint value, bool isHigh) { }
        protected virtual void SetBlackSun(uint value, bool isHigh) { }

        private void UpdateMergeValues()
        {
            // Try the customer defined defaults first
            int index = GetMergeLineOffsetIndex(HighGainIndex, LowGainIndex);
            if (index < 0)
            {
                // if not found try the factory defaults
                index = GetFactoryMergeLineOffsetIndex(HighGainIndex, LowGainIndex);
            }
            if (index < 0)
            {
                // if not found use the manufacturer's defaults
                GainRatio = GainOffsetPairArray[LowGainIndex, HighGainIndex].gainRatio;
                Temperature1 = GainOffsetPairArray[LowGainIndex, HighGainIndex].temperature1;
                LineOffset1 = GainOffsetPairArray[LowGainIndex, HighGainIndex].lineOffset1;
                Temperature2 = GainOffsetPairArray[LowGainIndex, HighGainIndex].temperature2;
                LineOffset2 = GainOffsetPairArray[LowGainIndex, HighGainIndex].lineOffset2;
                Temperature3 = GainOffsetPairArray[LowGainIndex, HighGainIndex].temperature3;
                LineOffset3 = GainOffsetPairArray[LowGainIndex, HighGainIndex].lineOffset3;
            }

            OnMergeValuesChangeAction(new EventArgs());
        }

        #endregion Protected Methods

        //**************************************************************
        #region Public Events

        public event EventHandler<EventArgs> MergeValuesChangeAction;

        protected virtual void OnMergeValuesChangeAction(EventArgs e)
        {
            EventHandler<EventArgs> handler = MergeValuesChangeAction;
            if (handler != null)
            {
                handler(this, e);
            }
        }

        #endregion Public Events

    }
}
