//**************************************************************
// Finger Lakes Instrumentation Camera Interface library
//
// The NativeMethods class contains static methods which use interp services
// to properly invoke and marshal functions exported by libflipro.dll (and libflieng.dll).
//
// These method definitions MUST be consistent with the 
// versions of libflipro.dll (and libflieng.dll) referenced by the application.
// Specifically, the method definitions must match the function prototype
// declarations in libflipro.h
//
// NOTE: These static methods are NOT guaranteed to be MTSafe. 
// They are intended to be invoked by a higher abstraction level layer 
// providing the appropriate critical region protection.
//
// Refer to FLI Software Development Library documentation and libflipro.h
// Doxygen comments for detailed descriptions of each function.

using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Text;
using static FliProCameraLib.FliPro;

namespace FliProCameraLib
{
    //using static FliPro;  Not ready for C#6 yet

    internal static partial class NativeMethods
    {
        //**************************************************************
        #region Private Global Variables

        private const int NUMBER_OF_DAC_CHANNELS = 4;
        private static CAMERACAP CameraCapabilities;
        private static bool CapabilitiesValid = false;

        #endregion Private Global Variables

        //**************************************************************
        #region Library Functions

        //-------------------------------------------
        /// <summary>
        /// Get the current API dll version string
        /// Prior to call, version.Capacity must be >= len
        /// </summary>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetAPIVersion(
                    [Out, MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                    uint uiLength);

        //-------------------------------------------
        /// <summary>
        /// Read the described data from Non Volatile memory
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOffset">Offset into NVM from where the data is to be read</param>
        /// <param name="pData">The data read from the camera</param>
        /// <param name="uiLength">The length of the data to be read</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPRONV_ReadNVStorage(
            int iHandle,
            uint uiOffset,
            [Out] byte[] pData, 
            uint uiLength);

        //-------------------------------------------
        /// <summary>
        /// Write the described data to Non Volatile memory
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOffset">Offset into NVM where the data is to be written</param>
        /// <param name="pData">The data to be written</param>
        /// <param name="uiLength">The length of the data to be written</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPRONV_WriteNVStorage(
            int iHandle,
            uint uiOffset,
            byte[] pData,
            uint uiLength);

        #endregion Library Functions

        //**************************************************************
        #region Camera Enumeration, Open, Close

        //-------------------------------------------
        /// <summary>
        /// Retrieve a list of all connected cameras
        /// </summary>
        /// <param name="DeviceInfo">FPRODEVICEINFO array to receive info on all connected cameras</param>
        /// <param name="NumDevices">On input, max capacity of DeviceInfo[]. On output, actual camera count </param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetCameraList(
                    [Out] FPRODEVICEINFO[] DeviceInfo,
                    ref uint NumDevices);

        //-------------------------------------------
        /// <summary>
        /// Open a connection to the specified camera
        /// </summary>
        /// <param name="DevInfo"></param>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_Open(
                    ref FPRODEVICEINFO DevInfo,
                    out int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Close the connection to a camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_Close(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Get the current camera firmware and fpga version strings
        /// </summary>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCam_GetDeviceVersion(
            int iHandle,
            out FPRODEVICEVERSIONS Version);

        #endregion Device Enumeration, Open, Close

        //**************************************************************
        #region Frame Data Functions

        //-------------------------------------------
        /// <summary>
        /// Abort the current frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureAbort(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Terminates the current frame capture sequence and makes available the data accumulated so far
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureEnd(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Start a frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiFrameCount">Number of frames to capture. 0 = no limit</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureStart(
                    int iHandle,
                    uint uiFrameCount);

        //-------------------------------------------
        /// <summary>
        /// Stop a frame capture sequence
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureStop(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Capture one thumbnail frame image
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_CaptureThumbnail(
                    int iHandle);

        //-------------------------------------------
        /// <summary>
        /// Enable/disable image data in frame. When disabled, only dummy pixels (if enabled) are produced
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnableImageData"></param>
        /// <param name="bEnableDummyPixels"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetImageDataEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Enable/disable image data in frame. When disabled, only dummy pixels (if enabled) are produced
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnableImageData"></param>
        /// <param name="bEnableDummyPixels"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetImageDataEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Get Enable test image generation settings
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <param name="eFormat"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetTestImageEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    out TESTIMAGEMODE eFormat);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy pixel configuration to be appended to row data.
        /// If enabled, eight dummy pixels will are appended to every other row
        /// starting with the second row of data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Enable"> true if dummy pixels will be appended</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyPixelEnable(
                    int iHandle,
                    out bool Enable);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy row configuration to be appended to frame data.
        /// If enabled, the specified number of dummy rows are appended after the last row
        /// of image frame data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pNumRowss">number of rows to append when Enable = true</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyRowPostFrameEnable(
                    int iHandle,
                    out int pNumRows);

        //-------------------------------------------
        /// <summary>
        /// Get the dummy row configuration to be prepended to frame data.
        /// If enabled, the specified number of dummy rows are prepended before the first row
        /// of image frame data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pNumRows">number of rows to prepend when Enable = true</param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetDummyRowPreFrameEnable(
                    int iHandle,
                    out int pNumRows);

        //-------------------------------------------
        /// <summary>
        /// Returns the frame type set by the #FPROFrame_SetFrameType() API.
        /// The default frame type is FPRO_FRAMETYPE_NORMAL.  Typically this is only used for testing
        /// purposes.  Consult your documentation for a description and availability of the different
        /// frame types found on a given camera model.  See also #FPRO_FRAME_TYPE.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pType">The current frame type setting.</param>
        /// <returns>Greater than or equal to 0 on success, less than 0 on failure.</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetFrameType(
                    int iHandle,
                    [MarshalAs(UnmanagedType.I4)]
                    out FPRO_FRAME_TYPE pType);

        //-------------------------------------------
        /// <summary>
        /// Get the current pixel depth and LSB position
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="PixelDepth">Current pixel depth in bits</param>
        /// <param name="PixedlLSB">The Pixel LSB offset</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetPixelConfig(
                    int iHandle,
                    out uint PixelDepth,
                    out uint PixelLSB);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrame(
                    int iHandle,
                    IntPtr pFrameData,
                    ref uint Size,
                    uint uiTimeoutMS);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame after an external trigger has fired
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrameExt(
                    int iHandle,
                    IntPtr pFrameData,
                    ref uint Size);

        //-------------------------------------------
        /// <summary>
        /// Get the next available Video frame
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="FrameData">Buffer to receive frame data</param>
        /// <param name="Size">On input: desired frame size. On return: actual frame size</param>
        /// <param name="uiTimeoutMS">Timeout in milliSec</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_GetVideoFrame(
                    int iHandle,
                    byte[] FrameData,
                    ref uint Size,
                    uint uiTimeoutMS,
                    [MarshalAs(UnmanagedType.I1)]
                    bool Trigger = false);

        //-------------------------------------------
        /// <summary>
        /// Check if a frame is available to transfer
        /// Note: This may not be useable due to hw time constraints.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bAvailable">true if a frame is available</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_IsAvailable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.I1)]
                    out bool bAvailable);

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable dummy pixels at the end of image row data.
        /// When enabled, eight dummy pixels are appended to every other row of image data
        /// starting withthe second row of data.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetDummyPixelEnable(
                    int iHandle,
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Sets Frame Type produced by the camera.  The frame type can be retrieved using the #FPROFrame_GetFrameType()
        /// API. The default frame type is FPRO_FRAMETYPE_NORMAL.  Typically this is only used for testing
        /// purposes.  Consult your documentation for a description and availability of the different
        /// frame types found on a given camera model.  See also #FPRO_FRAME_TYPE.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">The frame type to set.</param>
        /// <returns>Greater than or equal to 0 on success, less than 0 on failure.</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetFrameType(
                    int iHandle,
                    [MarshalAs(UnmanagedType.I4)]
                    FPRO_FRAME_TYPE eType);

        //-------------------------------------------
        /// <summary>
        /// Sets the configuratrion of dummy rows to be appended afte the last row
        /// of frame data
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumRows"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetReferenceRowPostFrameCount(
                    int iHandle,
                    uint uiNumRows);

        //-------------------------------------------
        /// <summary>
        /// Sets the configuratrion of dummy rows to be prepended before the first row
        /// of frame data
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumRows"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetReferenceRowPreFrameCount(
                    int iHandle,
                    uint uiNumRows);

        //-------------------------------------------
        /// <summary>
        /// Enable test image generation
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable"></param>
        /// <param name="eFormat"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTestImageEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    TESTIMAGEMODE eFormat);

        //-------------------------------------------
        /// <summary>
        /// Define the Tracking frame area. During tracking the entire row is captured so we don't need a width parameter.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiStartRow"></param>
        /// <param name="uiEndRow"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTrackingArea(
                    int iHandle,
                    uint uiStartRow,
                    uint uiEndRow);

        //-------------------------------------------
        /// <summary>
        /// Enable/Disable Tracking frame generation
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiNumTrackingFrames"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetTrackingAreaEnable(
                    int iHandle,
                    uint uiNumTrackingFrames);

        //-------------------------------------------
        /// <summary>
        /// Set the current pixel configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPixelDepth">pixel depth in bits</param>
        /// <param name="uiPixelLSB">pixel LSB offset</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetPixelConfig(
                    int iHandle,
                    uint uiPixelDepth,
                    uint uiPixelLSB);

        //-------------------------------------------
        /// <summary>
        /// Define the image size and location
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiColOffset">start column of the image frame</param>
        /// <param name="uiRowOffset">start row of the image frame</param>
        /// <param name="uiWidth">Width of image frame in pixels</param>
        /// <param name="uiHeight">Height of image frame in pixels</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROFrame_SetImageArea(
                    int iHandle,
                    uint uiColOffset,
                    uint uiRowOffset,
                    uint uiWidth,
                    uint uiHeight);

        #endregion Frame Data Functions

        //**************************************************************
        #region Camera Control

        //-------------------------------------------
        /// <summary>
        /// Read the base temperature of the camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Temp">temperature in deg C</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetBaseTemperature(
                    int iHandle,
                    out int Temp);

        //-------------------------------------------
        /// <summary>
        /// Read the current exposure time for the image sensor (in nanoseconds)
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiExposureTime">Exposure Time (in 10uSec increments)</param>
        /// <param name="uiDelayTime">The time in nanoseconds between consecutive exposures</param>
        /// <param name="bImmediate">Flag to not finish flushing the sensor before exposing</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetExposure(
                    int iHandle,
                    out long ExposureTime,
                    out long DelayTime,
                    out bool Immediate);

        //-------------------------------------------
        /// <summary>
        /// Get the current external trigger settings
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">External trigger enabled if true</param>
        /// <param name="eTrigType">The type of external trigger as well as the edge to trigger on</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetExternalTriggerEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    out FliTriggerMode eTrigType);

        //-------------------------------------------
        /// <summary>
        /// Get the current fan enable status
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Fan on if true</param>
        /// <param name="uiPwrPercentage">Fan power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetFanEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Get the Frame interval in nanoseconds
        /// </summary>
        /// <param name="iHandle"></param>
        /// <param name="FrameInterval"></param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetFrameInterval(
                    int iHandle,
                    out long FrameInterval);

        //-------------------------------------------
        /// <summary>
        /// Get the GPS State
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eGPSState">The state of the GPS</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetGPSState(
                    int iHandle,
                    [MarshalAs(UnmanagedType.I4)]
                    out FliGPSState eGPSState);

        //-------------------------------------------
        /// <summary>
        /// Get the current heater configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Heater on if true</param>
        /// <param name="uiPwrPercentage">Heater power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetHeaterConfig(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn,
                    out uint uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Get the current heater configuration
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPwrPercentage">Heater power level (0-100)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetHeaterPower(
                    int iHandle,
                    out uint uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Get the current illumination delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOnDelay">Delay for turning on illumination (in tbd mSec increments)</param>
        /// <param name="uiOffDelay">Delay for turning off illumination (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetIlluminationDelay(
                    int iHandle,
                    out uint uiOnDelay,
                    out uint uiOffDelay);

        //-------------------------------------------
        /// <summary>
        /// Get the current illumination on/off status
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Illumination on if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetIlluminationOn(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Returns the state of the LED on or off setting.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pOn">true: LED on, false: LED off</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetLED(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool pOn);

        //-------------------------------------------
        /// <summary>
        /// Returns the LED Duration setting.
        /// 
        /// Note that each camera model may have different resolution capability on the duration
        /// (i.e. something larger than a micro-second).  The FPROCtrl_SetLEDDuration() converts the
        /// micro-second value passed to the proper resolution for the camera.  This call reverses
        /// the conversion.  As such, the value returned by this call may not match the value exactly
        /// as that passed to FPROCtrl_SetLEDDuration().  On most cameras, the physical resolution for
        /// this duration is 10usecs.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pDurationUsec">Duration in micro seconds after exposure fires.  0xFFFFFFFF = always on.</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetLEDDuration(
                    int iHandle,
                    out int pDurationUsec);

        //-------------------------------------------
        /// <summary>
        /// Reads the internal sensor temperature of the camera.  
        /// Note that if this is called during an exposure and 'Read During Exposure' is not
        /// enabled, the last value successfully read will be returned. 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Temp">temperature in deg C</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetSensorTemperature(
                    int iHandle,
                    out int Temp);

        //-------------------------------------------
        /// <summary>
        /// Read the "sensor temperature during exposure" enable flag
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">Read during exposure enabled if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetSensorTemperatureReadEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bEnable);

        //////////////////////////////////////////
        ///
        /// @brief Gets the current shutter setting.
        ///
        /// By default the camera controls the shutter during exposures.  In order for 
        /// the #FPROCtrl_SetShutterOpen() API to correctly control the shutter,
        /// the #FPROCtrl_SetShutterOverride() API must be called prior with the override 
        /// parameter set to True.
        ///
        ///	@param iHandle - The handle to an open camera device returned from FPROCam_Open()
        /// @param pOpen - true: Shutter open, false: Shutter closed
        ///
        /// @return Greater than or equal to 0 on success, less than 0 on failure.
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetShutterOpen(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool pOpen);

        //////////////////////////////////////////
        ///
        /// @brief Gets the current shutter override setting.
        ///
        /// By default the camera controls the shutter during exposures.  In order for 
        /// the #FPROCtrl_SetShutterOpen() API to correctly control the shutter,
        /// the shutter override must be set to True.
        ///
        ///	@param iHandle - The handle to an open camera device returned from FPROCam_Open()
        /// @param pOverride - true: Allows user control of shutter, false: Camera controls shutter
        ///
        /// @return Greater than or equal to 0 on success, less than 0 on failure.
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetShutterOverride(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool pOverride);

        //-------------------------------------------
        /// <summary>
        /// Read the sensor temperatures
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAmbientTemp">The current Ambient temperature on the camera</param>
        /// <param name="pBaseTemp">The current Base temperature on the camera</param>
        /// <param name="pAmbientTemp">The current temperature of the Cooler unit on the camera</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetTemperatures(
                    int iHandle,
                    ref double pAmbientTemp,
                    ref double pBaseTemp,
                    ref double pCoolerTemp);

        //-------------------------------------------
        /// <summary>
        /// Read the Cooler Duty Cycle
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pDutyCycle">The percentage at which the Cooler is running </param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetCoolerDutyCycle(
                    int iHandle,
                    ref int pDutyCycle);
        
        //-------------------------------------------
        /// <summary>
        /// Get the max time full power is applied to shutter 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiTime">max full power duration (in 10 mSec increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetShutterFullPowerTime(
                    int iHandle,
                    out uint uiTime);

        //-------------------------------------------
        /// <summary>
        /// Get the base temperature setpoint
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="SetPoint">Temperature setpoint (in 0.1 deg C increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_GetTemperatureSetPoint(
                    int iHandle,
                    ref double SetPoint);

        //-------------------------------------------
        /// <summary>
        /// Set exposure time
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiExposureTime">Exposure time in nanoseconds</param>
        /// <param name="uiDelayTime">The time in nanoseconds between consecutive exposures</param>
        /// <param name="bImmediate">Flag to not finish flushing the sensor before exposing</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetExposure(
                    int iHandle,
                    long uiExposureTime,
                    long uiDelayTime,
                    bool bImmediate);

        //-------------------------------------------
        /// <summary>
        /// Set external trigger enable and active edge
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">external trigger enabled if true</param>
        /// <param name="eTrigType">The type of external trigger as well as the edge to trigger on</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetExternalTriggerEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    FliTriggerMode eTrigType);


        //-------------------------------------------
        /// <summary>
        /// Turn fan on/off 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn"></param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetFanEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Set the frame interval in nanoseconds
        /// </summary>
        /// <param name="iHandle"></param>
        /// <param name="FrameInterval"></param>
        /// <returns></returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetFrameInterval(
                    int iHandle,
                    long FrameInterval);

        //-------------------------------------------
        /// <summary>
        /// Turn Heater on/off at specified power level
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn"></param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetHeaterConfig(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn,
                    byte uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Turn Heater on/off at specified power level
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiPwrPercentage"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetHeaterPower(
                    int iHandle,
                    int uiPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Specify the illumination on and off delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOnDelay">Delay for turning on illumination (in tbd mSec increments)</param>
        /// <param name="uiOffDelay">Delay for turning off illumination (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetIlluminationDelay(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U2)]
                    ushort uiOnDelay,
                    [MarshalAs(UnmanagedType.U2)]
                    ushort uiOffDelay);

        //-------------------------------------------
        /// <summary>
        /// Turn Illumination on/off
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">turn illumination on if true, off if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetIlluminationOn(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        //-------------------------------------------
        /// <summary>
        /// Turn LED on/off
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOn">Turn LED on if true, off if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetLED(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOn);

        /// <summary>
        /// Set LED Duration during exposure.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiDuration">Duration in micro seconds after exposure fires.  0xFFFFFFFF = always on.</param>
        /// <returns>Greater than or equal to 0 on success, less than 0 on failure.</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetLEDDuration(
                    int iHandle, 
                    uint uiDuration);

        //-------------------------------------------
        /// <summary>
        /// Set the "sensor temperature during exposure" enable flag
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">Read during exposure enabled if true</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetSensorTemperatureReadEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable);

        //-------------------------------------------
        /// <summary>
        /// Specify the shutter full power time
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiTime">max full power duration (in 10mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterFullPowerTime(
                    int iHandle,
                    byte uiTime);

        //-------------------------------------------
        /// <summary>
        /// Open/Close the shutter
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOpen"Open shutter if true, close if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterOpen(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOpen);


        //-------------------------------------------
        /// <summary>
        /// Overrides the normal shutter control
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bOverride"Overrides shutter control if true, normal if false</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetShutterOverride(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bOverride);

        //-------------------------------------------
        /// <summary>
        /// Set the base temperature setpoint
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="SetPoint">Temperature setpoint (in 0.1 deg C increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCtrl_SetTemperatureSetPoint(
                    int iHandle,
                    double SetPoint);

        #endregion Camera Control

        //**************************************************************
        #region Sensor Functions

        //-------------------------------------------
        /// <summary>
        /// Get the horizontal and vertical binning values
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiVertical">Vertical binning setting</param>
        /// <param name="uiHorizontal">Horizontal binning setting</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBinning(
                    int iHandle,
                    out uint uiVertical,
                    out uint uiHorizontal);

        //-------------------------------------------
        /// <summary>
        /// Get the table containing all the binning value pairs
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pBinTable">Pointer to the binning table</param>
        /// <param name="pTableSizeBytes">Size in bytes of the binning table</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBinningTable(
            int iHandle,
            IntPtr pBinTable, 
            out int pTableSizeBytes);

        //-------------------------------------------
        /// <summary>
        /// Get the current black level adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAdjustValue">The returned value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackLevelAdjust(
                    int iHandle,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the current black level adjust value for the given channel
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eChan">The target channel</param>
        /// <param name="pAdjustValue">The returned value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackLevelAdjustEx(
                    int iHandle,
                    uint eChan,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the current Black Sun adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pAdjustValue">The returned value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackSunAdjust(
                    int iHandle,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the current Black Sun adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eChan">The target channel</param>
        /// <param name="pAdjustValue">The returned value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetBlackSunAdjustEx(
                    int iHandle,
                    uint eChan,
                    out uint pAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Get the collection of camera capabilities
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Cap">Capabilities struct</param>
        /// <param name="CauiCapLength">Length of capabilities struct in bytes</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetCapabilities(
                    int iHandle,
                    out CAMERACAP Cap,
                    ref uint uiCapLength);

        //-------------------------------------------
        /// <summary>
        /// Get the current setting for the specified gain index
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainIndex">returned gain index value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetGainIndex(
                    int iHandle,
                    GainTable eTable,
                    out uint uiGainIndex);

        //-------------------------------------------
        /// <summary>
        /// Get the specified Gain Table
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainValues">Array of returned gain values</param>
        /// <param name="uiNumEntries">Size of GainValues array on entry, number returned on exit</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetGainTable(
                    int iHandle,
                    GainTable eTable,
                    [Out] MFENG_CFGGAIN[] uiGainValues,
                    ref uint uiNumEntries);

        internal static int FPROSensor_GetGainTableEx(
                     int iHandle,
                    GainTable eTable,
                    FliGainValue[] uiGainValues,
                    ref uint uiNumEntries)
        {
            if (uiNumEntries > 0)
            {
                MFENG_CFGGAIN[] MGainValues = new MFENG_CFGGAIN[uiNumEntries];

                FPROSensor_GetGainTable(iHandle, eTable, MGainValues, ref uiNumEntries);

                if (uiGainValues.Length != uiNumEntries)
                {
                    return -1;
                }

                for (int i = 0; i < uiNumEntries; i++)
                {
                    uiGainValues[i].uiValue = MGainValues[i].uiValue;
                    uiGainValues[i].uiDeviceIndex = (int)MGainValues[i].uiDeviceIndex;
                }
                return 0;
            }
            else
            {
                return 0;
            }
        }

        //-------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bIsHDR"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetHDREnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool bIsHDR);
        
        //-------------------------------------------
        /// <summary>
        /// 
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiModeIndex"></param>
        /// <param name="Mode"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetMode(
                    int iHandle,
                    uint uiModeIndex,
                    out SENSEMODE Mode);

        //-------------------------------------------
        /// <summary>
        /// Get the number of modes supported by the camera
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="Count"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetModeCount(
                    int iHandle,
                    out uint Count,
                    out uint CurrentMode);

        //-------------------------------------------
        /// <summary>
        /// Get the camera horizontal and vertical scan direction
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pVInverted"></param>
        /// <param name="pHInverted"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetScanDirection(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool HInverted,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool VInverted);

        //-------------------------------------------
        /// <summary>
        /// Returns the sensor re-training setting.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pEnable">true: The sensor has been set to undergo training.  False - The sensor is not set to undergo training.</param>
        /// <param name="pHInverted"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_GetTrainingEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    out bool pEnable);

        //-------------------------------------------
        /// <summary>
        /// Set the horizontal and vertical binning value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiVertical"></param>
        /// <param name="uiHorizontal"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBinning(
                    int iHandle,
                    uint uiVertical,
                    uint uiHorizontal);

        //-------------------------------------------
        /// <summary>
        /// Set the black level adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiAdjustValue">The new value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackLevelAdjust(
                    int iHandle,
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the black level adjust value on one of the two channels
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eChan">The target channel</param>
        /// <param name="uiAdjustValue">The new value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackLevelAdjustEx(
                    int iHandle,
                    FPROBLACKADJUSTCHAN eChan, 
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the black sun adjust value
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiAdjustValue">The new value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackSunAdjust(
                    int iHandle,
                    uint eChan,
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the black sun adjust value on one of the two channels
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eChan">The target channel</param>
        /// <param name="uiAdjustValue">The new value</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetBlackSunAdjustEx(
                    int iHandle,
                    uint eChan,
                    uint uiAdjustValue);

        //-------------------------------------------
        /// <summary>
        /// Set the current index for the specified gain table
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eTable">Table to return, LDR, HDR, Global</param>
        /// <param name="uiGainIndex">index to set</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetGainIndex(
                    int iHandle,
                    GainTable eTable,
                    uint uiGainIndex);

        //-------------------------------------------
        /// <summary>
        /// Set the camera mode
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiModeIndex"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetMode(
                    int iHandle,
                    uint uiModeIndex);

        //-------------------------------------------
        /// <summary>
        /// Set the camera horizontal and vertical scan direction
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pVInverted"></param>
        /// <param name="pHInverted"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetScanDirection(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool pHInverted,
                    [MarshalAs(UnmanagedType.U1)]
                    bool pVInverted);

        //-------------------------------------------
        /// <summary>
        /// Enables/Disables sensor re-training.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="bEnable">true: Forces sensor to undergo training.  False - stops the sensor training.</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROSensor_SetTrainingEnable(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable);

        #endregion Sensor Functions

        //**************************************************************
        #region Auxiliary I/O Support Functions

        //-------------------------------------------
        /// <summary>
        /// Gets the direction and state for given Auxiliary I/O pin.
        /// For Output pins, the state of the pin will be the value last set with the
        /// #FPROAuxIO_GetPin() call.  For Input pins, the state of the pin reflects the 
        /// state of the physical input signal.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eAuxIO">The pin being retrieved</param>
        /// <param name="pDirection">In or Out</param>
        /// <param name="pState">High or Low</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROAuxIO_GetPin(
                    int iHandle,
                    FPROAUXIO eAuxIO, 
                    out FPROAUXIO_DIR pDirection,
                    out FPROAUXIO_STATE pState);

        //-------------------------------------------
        /// <summary>
        /// Gets Exposure Active Type Signal.
        /// This function gets the Exposure Type Signal for the Exposure Type Auxiliary 
        /// output pin.  Consult your documentation for signal timing details.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="pType">Exposure Active Signal Type</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROAuxIO_GetExposureActiveType(
                    int iHandle,
                    out FPROAUXIO_EXPACTIVETYPE pType);

        //-------------------------------------------
        /// <summary>
        /// Sets the direction and state for given Auxiliary I/O pin.
        /// Note that the state is only applicable to output pins.  It is ignored
        /// for input pins.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eAuxIO">The pin being retrieved</param>
        /// <param name="pDirection">In or Out</param>
        /// <param name="pState">High or Low</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROAuxIO_SetPin(
                    int iHandle,
                    FPROAUXIO eAuxIO,
                    FPROAUXIO_DIR pDirection,
                    FPROAUXIO_STATE pState);

        //-------------------------------------------
        /// <summary>
        /// Exposure Active Type Signal.
        /// This function sets the Exposure Type Signal for the Exposure Type Auxiliary 
        /// output pin.  Consult your documentation for signal timing details.
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="eType">Exposure Active Signal Type</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROAuxIO_SetExposureActiveType(
                    int iHandle,
                    FPROAUXIO_EXPACTIVETYPE eType);

        #endregion Auxiliary I/O Support Functions

        //**************************************************************
        #region Low Level Interface Functions

        //-------------------------------------------
        /// <summary>
        /// Send a Raw USB Message
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="TxData"></param>
        /// <param name="uiTxLength"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_SendRaw(
                    int iHandle,
                    byte[] TxData,
                    uint uiTxLength);

        //-------------------------------------------
        /// <summary>
        /// Send a Raw USB Message and Receive a Response
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="TxData">Byte array containg the data to send</param>
        /// <param name="uiTxLength">The length of the byte array containg the data to send</param>
        /// <param name="RxData">The returned data byte array</param>
        /// <param name="uiRxLength">the length of the returned data byte array</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_SendRecvRaw(
                    int iHandle,
                    byte[] TxData,
                    uint uiTxLength,
                    [Out] byte[] RxData,
                    ref uint uiRxLength);

        //-------------------------------------------
        /// <summary>
        /// Read an FPGA Register
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiReg"></param>
        /// <param name="pValue"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_ReadReg(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    uint uiReg,
                    out uint Value);

        //-------------------------------------------
        /// <summary>
        /// Write a value to an FPGA Register
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiReg"></param>
        /// <param name="uiValue"></param>
        /// <param name="uiMask"></param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPROCmd_WriteReg(
                    int iHandle,
                    [MarshalAs(UnmanagedType.U1)]
                    uint uiReg,
                    uint uiValue,
                    uint uiMask);

        #endregion Low Level Interface Functions

        //**************************************************************
        #region Debug Functions

        //-------------------------------------------
        /// <summary>
        /// Set the library debug level
        /// </summary>
        /// <param name="bEnable"> enable debug logging if true</param>
        /// <param name="level">Maximum debug level to be logged</param>
        /// <returns>negative on error</returns>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FPRODebug_EnableLevel(
                    [MarshalAs(UnmanagedType.U1)]
                    bool bEnable,
                    [MarshalAs(UnmanagedType.I4)]
                    DBGLEVEL level);

        //-------------------------------------------
        /// <summary>
        /// Write a message to the debug log
        /// </summary>
        /// <param name="level">Debug log level</param>
        /// <param name="message">message content</param>
        [DllImport("libflipro.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Ansi)]
        internal static extern void FPRODebug_Write(
                    [MarshalAs(UnmanagedType.I4)]
                    DBGLEVEL level,
                    [MarshalAs(UnmanagedType.LPStr)]
                    string message);

        #endregion Debug Functions

        //**************************************************************
        #region Engineering Functions

        #region Engineering Private Static Extern
        ///
        /// For all of these Private Static External functions, the Managed CS code must call the
        /// managed wrapper function below.  They are of the same name, but with a CS prefix.  The
        /// reason for this is due to the more complex marshalling code required to get the data
        /// back and forth from the dll and managed CS space.
        /// 

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_GetConfig(
                    int iHandle,
                    ref MFENG_CFGCAM cfgCam,
                    IntPtr pProgressCallback);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_GetConfigFromFile(
                    uint eDevType,
                    [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version, 
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_FreeConfig(
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SaveConfigToBin(
                   [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                   ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SaveConfigToCSV(
                    [MarshalAs(UnmanagedType.LPWStr)]
                    StringBuilder version,
                    ref MFENG_CFGCAM cfgCam);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCam_SetConfig(
                     int iHandle,
                     ref MFENG_CFGCAM cfgCam,
                     IntPtr pProgressCallback);


        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetDeviceRegs(
                      int iHandle,
                      uint uiDeviceId,
                      ref IntPtr pRegs,
                      uint uiNumRegs);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetHSMTimings(
                      int iHandle,
                      ref IntPtr pIdleTimings,
                      uint uiNumIdleTimings,
                      ref IntPtr pBinningTimings,
                      uint uiNumBinningTimings);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        private static extern int FENGCfg_SetVSMTimings(
                      int iHandle,
                      ref IntPtr pIdleTimings,
                      uint uiNumIdleTimings,
                      ref IntPtr pLongExposureTimings,
                      uint uiNumLongExposureTimings);

        #endregion Engineering Private Static Extern

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENGCam_SetSerialNumber(
                      int iHandle,
                      [MarshalAs(UnmanagedType.LPWStr)]
                      StringBuilder serialNumber,
                      uint uiNumChars);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetFwVersion(
                      int iHandle,
                      out uint pVersion);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetCoolerDAC(
                      int iHandle,
                      out uint pValue);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_GetCoolerDutyOverride(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U1)]
                      out bool pEnable,
                      out uint pPercent);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG_GetFanSpeed(
                      int iHandle,
                      out uint pPwrPercentage);


        //-------------------------------------------
        /// <summary>
        /// Get the time between setting Shutter Open and the shutter actually opening or closing
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOpenDelay">Shutter open delay (in tbd mSec increments)/param>
        /// <param name="uiCloseDelay">Shutter close delay (in tbd mSec increments)/param>
        /// <returns>negative on error</returns>
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FENG_GetShutterDelay(
                    int iHandle,
                    out uint uiOpenDelay,
                    out uint uiCloseDelay);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_SetCoolerDAC(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pValue);
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG8051_SetCoolerDutyOverride(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U1)]
                      bool pEnable,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pPercent);

        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        public static extern int FENG_SetFanSpeed(
                      int iHandle,
                      [MarshalAs(UnmanagedType.U4)]
                      uint pPwrPercentage);

        //-------------------------------------------
        /// <summary>
        /// Specify the shutter open and close delays
        /// </summary>
        /// <param name="iHandle">Camera device handle</param>
        /// <param name="uiOpenDelay">Shutter open delay (in tbd mSec increments)</param>
        /// <param name="uiCloseDelay">Shutter close delay (in tbd mSec increments)</param>
        /// <returns>negative on error</returns>
        [DllImport("libflieng.dll", CallingConvention = CallingConvention.Cdecl, CharSet = CharSet.Unicode)]
        internal static extern int FENG_SetShutterDelay(
                    int iHandle,
                    uint uiOpenDelay,
                    int uiCloseDelay);


        #region Engineering Marshalling Wrapper Functions
        public static int CSFENGCam_GetConfig(int iHandle, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            if (!CapabilitiesValid)
            {
                CameraCapabilities = new CAMERACAP();
                uint caplength = (uint)Marshal.SizeOf(CameraCapabilities);
                int result = FPROSensor_GetCapabilities(iHandle, out CameraCapabilities, ref caplength);
                CapabilitiesValid = (result >= 0);
            }

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            iResult = FENGCam_GetConfig(iHandle, ref mcfgCam, (IntPtr)0);
            if (iResult >= 0)
            {
                // Marshal the configuration to managed space
                MarshalConfigToManaged(ref mcfgCam, ref cscfgCam);

                // Free the unmanaged space so the caller does not have to
                FENGCam_FreeConfig(ref mcfgCam);
            }


            return (iResult);
        }

        public static int CSFENGCam_GetConfigFromFile(FPRODEVICETYPE eDevType, StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;
            //MFENG_CFGATODSETTING mcfgSetting;

            // needed for sizeof
            //mcfgSetting = new MFENG_CFGATODSETTING();

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;
            mcfgCam.cfgStartup.pCfgVSMIdleTimingRegs = (IntPtr)null; //(IntPtr)Marshal.AllocCoTaskMem(Marshal.SizeOf(mcfgSetting) * 26);
            mcfgCam.cfgStartup.pCfgVSMLongExpTimingRegs = (IntPtr)null; //(IntPtr)Marshal.AllocCoTaskMem(Marshal.SizeOf(mcfgSetting) * 26);
            mcfgCam.cfgStartup.uiNumVSMIdleTimingRegs = 0;
            mcfgCam.cfgStartup.uiNumVSMLongExpTimingRegs = 0;
            iResult = FENGCam_GetConfigFromFile((uint)eDevType.GetHashCode(), strFileName, ref mcfgCam);
            if (iResult >= 0)
            {
                // Marshal the configuration to managed space
                MarshalConfigToManaged(ref mcfgCam, ref cscfgCam);

                // Free the unmanaged space so the caller does not have to
                FENGCam_FreeConfig(ref mcfgCam);
            }

            return (iResult);
        }

        public static int CSFENGCam_SaveConfigToBin(StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SaveConfigToBin(strFileName, ref mcfgCam);

            return (iResult);
        }
        public static int CSFENGCam_SaveConfigToCSV(StringBuilder strFileName, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SaveConfigToCSV(strFileName, ref mcfgCam);

            return (iResult);
        }

        public static int CSFENGCam_SetConfig(int iHandle, ref CSFENG_CFGCAM cscfgCam)
        {
            int iResult;
            MFENG_CFGCAM mcfgCam;

            mcfgCam = new MFENG_CFGCAM();
            mcfgCam.pModes = (IntPtr)0;
            mcfgCam.uiNumModes = 0;

            MarshalConfigToUnManaged(ref cscfgCam, ref mcfgCam);
            iResult = FENGCam_SetConfig(iHandle, ref mcfgCam,(IntPtr)0);

            return (iResult);
        }


        public static int CSFENGCam_SetDeviceRegs(int iHandle,uint uiDeviceId, ref MFENG_CFGREG[] csRegs, uint uiNumRegs)
        {
            int iResult;
            IntPtr pRegs;

            // Marshal the data
            pRegs = MarshalCfgRegsToUnmanaged(uiNumRegs, ref csRegs);

            // call the FENG API ....
            iResult = FENGCfg_SetDeviceRegs(iHandle, uiDeviceId, ref pRegs, uiNumRegs);

            return (iResult);
        }

        public static int CSFENGCam_SetHSMTimings(int iHandle, ref MFENG_CFGREG[] csIdleTimings,uint uiNumIdleTimings,ref MFENG_CFGREG[] csBinningTimings, uint uiNumBinningTimings)
        {
            int iResult;
            IntPtr pIdleTimings;
            IntPtr pBinningTimings;

            // Marshal the data
            pIdleTimings = MarshalCfgRegsToUnmanaged(uiNumIdleTimings, ref csIdleTimings);
            pBinningTimings = MarshalCfgRegsToUnmanaged(uiNumBinningTimings, ref csBinningTimings);

            // call the FENG API ....
            iResult= FENGCfg_SetHSMTimings(iHandle, ref pIdleTimings, uiNumIdleTimings, ref pBinningTimings, uiNumBinningTimings);

            return (iResult);
        }

        public static int CSFENGCam_SetVSMTimings(int iHandle, ref MFENG_CFGREG[] csIdleTimings, uint uiNumIdleTimings, ref MFENG_CFGREG[] csLongExposureTimings, uint uiNumLongExposureTimings)
        {
            int iResult;
            IntPtr pIdleTimings;
            IntPtr pLongExposureTimings;

            // Marshal the data
            pIdleTimings = MarshalCfgRegsToUnmanaged(uiNumIdleTimings, ref csIdleTimings);
            pLongExposureTimings = MarshalCfgRegsToUnmanaged(uiNumLongExposureTimings, ref csLongExposureTimings);

            // call the FENG API ....
            iResult = FENGCfg_SetVSMTimings(iHandle, ref pIdleTimings, uiNumIdleTimings, ref pLongExposureTimings, uiNumLongExposureTimings);

            return (iResult);
        }

        public static int GetCapabilities(int iHandle, out CAMERACAP Cap, ref uint uiCapLength)
        {
            int result = FPROSensor_GetCapabilities(iHandle, out Cap, ref uiCapLength);

            CameraCapabilities = Cap;
            CapabilitiesValid = (result >= 0);
            return result;
        }

        private static void MarshalConfigToManaged(ref MFENG_CFGCAM mcfgCam, ref CSFENG_CFGCAM cscfgCam)
        {
            IntPtr ptrModes;
            MFENG_CFGMODES  mcfgMode;

            // Marshal the modes
            cscfgCam.uiNumModes = mcfgCam.uiNumModes;
            if (cscfgCam.uiNumModes > 0)
            {
                cscfgCam.modes = new CSFENG_CFGMODES[cscfgCam.uiNumModes];
                ptrModes = mcfgCam.pModes;
                for (uint i = 0; (i < cscfgCam.uiNumModes) && (ptrModes != null); ++i)
                {
                    mcfgMode= (MFENG_CFGMODES)Marshal.PtrToStructure(ptrModes, typeof(MFENG_CFGMODES));

                    // now we can marshal to the actual structure
                    MarshalModeToManaged(ref mcfgMode, ref cscfgCam.modes[i]);
                    //cscfgCam.modes[i] = (CSFENG_CFGMODES)Marshal.PtrToStructure(ptrModes, typeof(CSFENG_CFGMODES));
                    ptrModes += Marshal.SizeOf(mcfgMode);
                }
            }

            // Marshal the Startup
            MarshalStartupToManaged(ref mcfgCam.cfgStartup, ref cscfgCam.cfgStartup);

            // Marshal the Device Info
            MarshalDeviceInfoToManaged(ref mcfgCam.cfgCamInfo, ref cscfgCam.cfgCamInfo);

            //cscfgCam.cfgStartup = mcfgCam.cfgStartup;
        }

        //private static int modeIndex = 0;
        private static void MarshalModeToManaged(ref MFENG_CFGMODES fromMode, ref CSFENG_CFGMODES toMode)
        {
            // Do the discrete settings first
            //HDR and LDR flags
            if (fromMode.bHDREnabled == 0)
                toMode.bHDREnabled = false;
            else
                toMode.bHDREnabled = true;
            if (fromMode.bLDCEnabled == 0)
                toMode.bLDCEnabled = false;
            else
                toMode.bLDCEnabled = true;

            // Modeinfo and gains
            toMode.modeInfo = fromMode.modeInfo;
            toMode.iLowGainIndex = fromMode.iLowGainIndex;
            toMode.iHighGainIndex = fromMode.iHighGainIndex;
            toMode.uiNumBlackLevel = fromMode.uiNumBlackLevel;
            MarshalBlackToManaged(fromMode.pBlackLevel, fromMode.uiNumBlackLevel, out toMode.pBlackLevel);
            toMode.uiNumBlackSun = fromMode.uiNumBlackSun;
            MarshalBlackToManaged(fromMode.pBlackSun, fromMode.uiNumBlackSun, out toMode.pBlackSun);

            // now the DAC settings
            toMode.uiNumDACImage1 = fromMode.uiNumDACImage1;
            MarshalDACImagesToManaged(fromMode.dacImage1, fromMode.uiNumDACImage1, out toMode.dacImage1);
            toMode.uiNumDACImage2 = fromMode.uiNumDACImage2;
            MarshalDACImagesToManaged(fromMode.dacImage2, fromMode.uiNumDACImage2, out toMode.dacImage2);

            // now the HSM Timing Settings
            toMode.uiNumHSMIdleTimingRegs = fromMode.uiNumHSMIdleTimingRegs;
            MarshalCfgRegsToManaged(fromMode.pCfgHSMIdleTimingRegs, fromMode.uiNumHSMIdleTimingRegs, out toMode.pCfgHSMIdleTimingRegs);
            toMode.uiNumHSMBinningTimingRegs = fromMode.uiNumHSMBinningTimingRegs;
            MarshalCfgRegsToManaged(fromMode.pCfgHSMBinningTimingRegs, fromMode.uiNumHSMBinningTimingRegs, out toMode.pCfgHSMBinningTimingRegs);

            // now the Other Registers
            toMode.uiNumOtherRegs = fromMode.uiNumOtherRegs;
            MarshalCfgRegsToManaged(fromMode.pModeOtherRegs, fromMode.uiNumOtherRegs, out toMode.pModeOtherRegs);

            // Debug Code
            //if (fromMode.uiNumOtherRegs == 0)
            //{
            //    toMode.uiNumOtherRegs = 2;
            //    toMode.pModeOtherRegs = new MFENG_CFGREG[toMode.uiNumOtherRegs];
            //    for (int i = 0; i < toMode.pModeOtherRegs.Length; i++)
            //    {
            //        toMode.pModeOtherRegs[i].uiDevice = (uint)(0x31 + modeIndex + i);
            //        toMode.pModeOtherRegs[i].uiRegister = (uint)(0x07 + modeIndex + i);
            //        toMode.pModeOtherRegs[i].uiValue = (uint)(0x1 + modeIndex + i);
            //        toMode.pModeOtherRegs[i].uiMask = (uint)(0xF8 + modeIndex + i);
            //    }
            //    modeIndex++;
            //}
            // End Debug Code

            // A to D Regs
            toMode.uiNumAtoD1Regs = fromMode.uiNumAtoD1Regs;
            MarshalCfgRegsToManaged(fromMode.pCfgSensorAtoD1, fromMode.uiNumAtoD1Regs, out toMode.pCfgSensorAtoD1);
            toMode.uiNumAtoD1Regs = fromMode.uiNumAtoD2Regs;
            MarshalCfgRegsToManaged(fromMode.pCfgSensorAtoD2, fromMode.uiNumAtoD2Regs, out toMode.pCfgSensorAtoD2);
        }

        private static void MarshalBlackToManaged(IntPtr pBlack, uint uiNumBlack, out MFENG_CFGBLACK[] cfgBlack)
        {
            MFENG_CFGBLACK csBlack = new MFENG_CFGBLACK();

            cfgBlack = null;

            if (uiNumBlack > 0)
            {
                cfgBlack = new MFENG_CFGBLACK[uiNumBlack];
                for (uint i = 0; (i < uiNumBlack) && (pBlack != null); ++i)
                {
                    csBlack = (MFENG_CFGBLACK)Marshal.PtrToStructure(pBlack, typeof(MFENG_CFGBLACK));
                    cfgBlack[i].eChannel = csBlack.eChannel;
                    cfgBlack[i].uiValue = csBlack.uiValue;
                    pBlack += Marshal.SizeOf(csBlack);
                }
            }
        }

        private static void MarshalCfgRegsToManaged(IntPtr ptrSetting, UInt32 uiNumRegs, out MFENG_CFGREG[] cfgRegs)
        {
            MFENG_CFGREG cscfgSetting;

            cscfgSetting = new MFENG_CFGREG();

            cfgRegs = null;
            if (uiNumRegs > 0)
            {
                cfgRegs = new MFENG_CFGREG[uiNumRegs];
                for (uint i = 0; (i < uiNumRegs) && (ptrSetting != null); ++i)
                {
                    cfgRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        private static void MarshalDACImagesToManaged(IntPtr mfDACImages, UInt32 uiNumDACs, out int[,] dacImages)
        {
            MFENG_CFGIMAGEDAC csDACImage = new MFENG_CFGIMAGEDAC();

            dacImages = null;
            if (uiNumDACs > 0)
            {
                dacImages = new int[CameraCapabilities.LowGain, NUMBER_OF_DAC_CHANNELS];

                // Initialize the values to -1
                for (int x = 0; x < CameraCapabilities.LowGain; x++)
                {
                    for (int y = 0; y < NUMBER_OF_DAC_CHANNELS; y++)
                    {
                        dacImages[x, y] = -1;
                    }
                }

                for (uint i = 0; (i < uiNumDACs) && (mfDACImages != null); ++i)
                {
                    csDACImage = (MFENG_CFGIMAGEDAC)Marshal.PtrToStructure(mfDACImages, typeof(MFENG_CFGIMAGEDAC));
                    int gainIndex = (int)((csDACImage.uiChannel >> 16) & 0xFFFF);
                    int channel = (int)(csDACImage.uiChannel & 0xF);
                    dacImages[gainIndex, channel] = (int)csDACImage.uiValue;
                    mfDACImages += Marshal.SizeOf(csDACImage);
                }

                // Make sure the first gain channels have a valid value
                for (int y = 0; y < NUMBER_OF_DAC_CHANNELS; y++)
                {
                    if (dacImages[0, y] == -1)
                    {
                        dacImages[0, y] = 0;
                    }
                }

                // Now set the the entire array
                for (int x = 1; x < CameraCapabilities.LowGain; x++)
                {
                    for (int y = 0; y < NUMBER_OF_DAC_CHANNELS; y++)
                    {
                        if (dacImages[x, y] == -1)
                        {
                            dacImages[x, y] = dacImages[x - 1, y];
                        }
                    }
                }
            }
        }

        private static IntPtr MarshalBlackToUnmanaged(uint uiNumBlack, MFENG_CFGBLACK[] cfgBlack)
        {
            MFENG_CFGBLACK csBlack = new MFENG_CFGBLACK();
            IntPtr ptrBlack = (IntPtr)0;
            IntPtr ptrBlackStart = (IntPtr)0;

            if (uiNumBlack > 0)
            {
                ptrBlackStart = Marshal.AllocCoTaskMem(Marshal.SizeOf(csBlack) * (int)uiNumBlack);
                if (ptrBlackStart != (IntPtr)0)
                {
                    ptrBlack = ptrBlackStart;
                    for (int i = 0; i < uiNumBlack; i++)
                    {
                        csBlack = new MFENG_CFGBLACK();
                        csBlack.eChannel = cfgBlack[i].eChannel;
                        csBlack.uiValue = cfgBlack[i].uiValue;
                        Marshal.StructureToPtr(csBlack, ptrBlack, false);
                        ptrBlack += Marshal.SizeOf(csBlack);
                    }
                }
            }

            return ptrBlackStart;
        }

        /// <summary>
        /// dacImage2 is needed because all 8 DACs need to be written whenever one field changes they all
        /// have to be written
        /// </summary>
        /// <param name="uiNumDACImage"></param>
        /// <param name="dacImage"></param>
        /// <param name="dacImage2"></param>
        /// <returns></returns>
        private static IntPtr MarshalDACImagesToUnmanaged(ref uint uiNumDACImage, int[,] dacImage, int[,] dacImage2)
        {
            MFENG_CFGIMAGEDAC csDACImage = new MFENG_CFGIMAGEDAC();
            IntPtr ptrDACImage = (IntPtr)0;
            IntPtr ptrDACImageStart = (IntPtr)0;

            // First run through and get a count
            uiNumDACImage = 0;
            for (int x = 0; x < CameraCapabilities.LowGain; x++)
            {
                for (int y = 0; y < NUMBER_OF_DAC_CHANNELS; y++)
                {
                    if ((x == 0) || (dacImage[x, y] != dacImage[x - 1, y]) || (dacImage2[x, y] != dacImage2[x - 1, y]))
                    {
                        // Once we find one that has changed we need to write all 4 entries
                        uiNumDACImage += NUMBER_OF_DAC_CHANNELS;
                        break;
                    }
                }
            }

            if (uiNumDACImage > 0)
            {
                ptrDACImageStart = Marshal.AllocCoTaskMem(Marshal.SizeOf(csDACImage) * (int)uiNumDACImage);
                if (ptrDACImageStart != (IntPtr)0)
                {
                    ptrDACImage = ptrDACImageStart;
                    for (int x = 0; x < CameraCapabilities.LowGain; x++)
                    {
                        for (int y = 0; y < NUMBER_OF_DAC_CHANNELS; y++)
                        {
                            if ((x == 0) || (dacImage[x, y] != dacImage[x - 1, y]) || (dacImage2[x, y] != dacImage2[x - 1, y]))
                            {
                                // Once we find one that has changed we need to write all 4 entries
                                for (int yPrime = 0; yPrime < NUMBER_OF_DAC_CHANNELS; yPrime++)
                                {
                                    csDACImage = new MFENG_CFGIMAGEDAC();
                                    csDACImage.uiChannel = (uint)((x << 16) | yPrime);
                                    csDACImage.uiValue = (uint)dacImage[x, yPrime];
                                    Marshal.StructureToPtr(csDACImage, ptrDACImage, false);
                                    ptrDACImage += Marshal.SizeOf(csDACImage);
                                }
                                break;
                            }
                        }
                    }
                }
            }

            return ptrDACImageStart;
        }

        private static void MarshalDeviceInfoToManaged(ref MFENG_CAMINFO mcfgCamInfo, ref CSFENG_CAMINFO cscfgCamInfo)
        {
            cscfgCamInfo.uiDeviceType = mcfgCamInfo.uiDeviceType;
            cscfgCamInfo.uiSerialNumLength = mcfgCamInfo.uiSerialNumLength;
            cscfgCamInfo.wcSerialNumber = mcfgCamInfo.wcSerialNumber;
        }

        private static void MarshalStartupToManaged(ref MFENG_CAMSTARTUP mcfgStartup, ref CSFENG_CAMSTARTUP cscfgStartup)
        {
            IntPtr ptrSetting;
            MFENG_CFGREG cscfgSetting;

            cscfgSetting = new MFENG_CFGREG();

            // First the discrete settings
            cscfgStartup.dblTemperatureSetPoint = mcfgStartup.dblTemperatureSetPoint;
            cscfgStartup.uiCoolerDAC = mcfgStartup.uiCoolerDAC;
            cscfgStartup.uiFanPercent = mcfgStartup.uiFanPercent;
            cscfgStartup.bFanEnabled = mcfgStartup.bFanEnabled;
            cscfgStartup.uiHeaterPercent = mcfgStartup.uiHeaterPercent;
            cscfgStartup.uiNumVSMIdleTimingRegs = mcfgStartup.uiNumVSMIdleTimingRegs;
            cscfgStartup.uiNumVSMLongExpTimingRegs = mcfgStartup.uiNumVSMLongExpTimingRegs;
            cscfgStartup.uiNumOtherRegs = mcfgStartup.uiNumOtherRegs;

            // Marshal the A to D settings
            if (cscfgStartup.uiNumVSMIdleTimingRegs > 0)
            {
                cscfgStartup.pCfgVSMIdleTimingRegs = new MFENG_CFGREG[cscfgStartup.uiNumVSMIdleTimingRegs];
                ptrSetting = mcfgStartup.pCfgVSMIdleTimingRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumVSMIdleTimingRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pCfgVSMIdleTimingRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
            if (cscfgStartup.uiNumVSMLongExpTimingRegs > 0)
            {
                cscfgStartup.pCfgVSMLongExpTimingRegs = new MFENG_CFGREG[cscfgStartup.uiNumVSMLongExpTimingRegs];
                ptrSetting = mcfgStartup.pCfgVSMLongExpTimingRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumVSMLongExpTimingRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pCfgVSMLongExpTimingRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }

            // 'Other' Registers
            if (cscfgStartup.uiNumOtherRegs > 0)
            {
                cscfgStartup.pOtherRegs = new MFENG_CFGREG[cscfgStartup.uiNumOtherRegs];
                ptrSetting = mcfgStartup.pOtherRegs;
                for (uint i = 0; (i < cscfgStartup.uiNumOtherRegs) && (ptrSetting != (IntPtr)0); ++i)
                {
                    cscfgStartup.pOtherRegs[i] = (MFENG_CFGREG)Marshal.PtrToStructure(ptrSetting, typeof(MFENG_CFGREG));
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        private static void MarshalConfigToUnManaged(ref CSFENG_CFGCAM cscfgCam, ref MFENG_CFGCAM mcfgCam)
        {
            IntPtr ptrModes;
            MFENG_CFGMODES  mcfgMode;

            // just needed for marshalling size
            mcfgMode = new MFENG_CFGMODES();

            // Marshal the modes
            mcfgCam.uiNumModes = cscfgCam.uiNumModes;
            if (mcfgCam.uiNumModes > 0)
            {
                ptrModes = Marshal.AllocCoTaskMem(Marshal.SizeOf(mcfgMode) * (int)mcfgCam.uiNumModes);
                mcfgCam.pModes = ptrModes;
                for (uint i = 0; (i < cscfgCam.uiNumModes) && (ptrModes != null); ++i)
                {
                    // Fill up a mode sructure
                    MarshalModeToUnmanaged(ref cscfgCam.modes[i], ref mcfgMode);

                    // now marshal to our pointer
                    Marshal.StructureToPtr(mcfgMode, ptrModes,false);

                    // fill in the next mode
                    ptrModes += Marshal.SizeOf(mcfgMode);
                }
            }

            // Marshal the Startup
            MarshalStartupToUnManaged(ref cscfgCam.cfgStartup, ref mcfgCam.cfgStartup);

            // Marshal the Device Info
            MarshalDeviceInfoToUnManaged(ref cscfgCam.cfgCamInfo, ref mcfgCam.cfgCamInfo);
        }

        private static void MarshalModeToUnmanaged(ref CSFENG_CFGMODES cscfgMode, ref MFENG_CFGMODES mcfgMode)
        {
            // HDR and LDC Flags
            if (cscfgMode.bHDREnabled)
                mcfgMode.bHDREnabled = 1;
            else
                mcfgMode.bHDREnabled = 0;
            if (cscfgMode.bLDCEnabled)
                mcfgMode.bLDCEnabled = 1;
            else
                mcfgMode.bLDCEnabled = 0;

            // First do all the discrete settings
            mcfgMode.modeInfo = cscfgMode.modeInfo;
            mcfgMode.iLowGainIndex = cscfgMode.iLowGainIndex;
            mcfgMode.iHighGainIndex = cscfgMode.iHighGainIndex;
            mcfgMode.uiNumBlackLevel = cscfgMode.uiNumBlackLevel;
            mcfgMode.pBlackLevel = MarshalBlackToUnmanaged(mcfgMode.uiNumBlackLevel, cscfgMode.pBlackLevel);
            mcfgMode.uiNumBlackSun = cscfgMode.uiNumBlackSun;
            mcfgMode.pBlackSun = MarshalBlackToUnmanaged(mcfgMode.uiNumBlackSun, cscfgMode.pBlackSun);

            // now the DAC Settings
            mcfgMode.dacImage1 = MarshalDACImagesToUnmanaged(ref mcfgMode.uiNumDACImage1, cscfgMode.dacImage1, cscfgMode.dacImage2);
            mcfgMode.dacImage2 = MarshalDACImagesToUnmanaged(ref mcfgMode.uiNumDACImage2, cscfgMode.dacImage2, cscfgMode.dacImage1);

            // now the HSM Timing Settings
            mcfgMode.uiNumHSMIdleTimingRegs = cscfgMode.uiNumHSMIdleTimingRegs;
            mcfgMode.pCfgHSMIdleTimingRegs = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumHSMIdleTimingRegs, ref cscfgMode.pCfgHSMIdleTimingRegs);
            mcfgMode.uiNumHSMBinningTimingRegs = cscfgMode.uiNumHSMBinningTimingRegs;
            mcfgMode.pCfgHSMBinningTimingRegs= MarshalCfgRegsToUnmanaged(mcfgMode.uiNumHSMBinningTimingRegs,ref cscfgMode.pCfgHSMBinningTimingRegs);

            // now the Other Registers
            mcfgMode.uiNumOtherRegs = cscfgMode.uiNumOtherRegs;
            mcfgMode.pModeOtherRegs = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumOtherRegs, ref cscfgMode.pModeOtherRegs);


            // A to D Regs
            mcfgMode.uiNumAtoD1Regs = cscfgMode.uiNumAtoD1Regs;
            mcfgMode.pCfgSensorAtoD1 = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumAtoD1Regs, ref cscfgMode.pCfgSensorAtoD1);
            mcfgMode.uiNumAtoD2Regs = cscfgMode.uiNumAtoD2Regs;
            mcfgMode.pCfgSensorAtoD2 = MarshalCfgRegsToUnmanaged(mcfgMode.uiNumAtoD2Regs, ref cscfgMode.pCfgSensorAtoD2);
        }

        private static IntPtr MarshalCfgRegsToUnmanaged(UInt32 uiNumRegs, ref MFENG_CFGREG[] cfgRegs)
        {
            MFENG_CFGREG cscfgSetting;
            IntPtr ptrSetting;
            IntPtr ptrSettingStart;

            // just needed for marshalling size
            cscfgSetting = new MFENG_CFGREG();

            ptrSetting = (IntPtr)0;
            ptrSettingStart = (IntPtr)0;
            if (uiNumRegs > 0)
            {
                ptrSettingStart = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)uiNumRegs);
                if (ptrSettingStart != (IntPtr)0)
                {
                    ptrSetting = ptrSettingStart;
                    for (uint i = 0; (i < uiNumRegs) && (ptrSetting != (IntPtr)0); ++i)
                    {
                        Marshal.StructureToPtr(cfgRegs[i], ptrSetting, false);
                        ptrSetting += Marshal.SizeOf(cscfgSetting);
                    }
                }
            }

            // return our result
            return (ptrSettingStart);
        }

        private static void MarshalDeviceInfoToUnManaged(ref CSFENG_CAMINFO cscfgCamInfo, ref MFENG_CAMINFO mcfgCamInfo)
        {
            mcfgCamInfo.uiDeviceType = cscfgCamInfo.uiDeviceType;
            mcfgCamInfo.uiSerialNumLength = cscfgCamInfo.uiSerialNumLength;
            mcfgCamInfo.wcSerialNumber = cscfgCamInfo.wcSerialNumber;
        }


        private static void MarshalStartupToUnManaged(ref CSFENG_CAMSTARTUP cscfgStartup, ref MFENG_CAMSTARTUP mcfgStartup)
        {
            IntPtr ptrSetting;
            MFENG_CFGREG cscfgSetting;

            // just needed for marshalling size
            cscfgSetting = new MFENG_CFGREG();

            // First the discrete settings
            mcfgStartup.dblTemperatureSetPoint = cscfgStartup.dblTemperatureSetPoint;
            mcfgStartup.uiCoolerDAC = cscfgStartup.uiCoolerDAC;
            mcfgStartup.uiFanPercent = cscfgStartup.uiFanPercent;
            mcfgStartup.bFanEnabled = cscfgStartup.bFanEnabled;
            mcfgStartup.uiHeaterPercent = cscfgStartup.uiHeaterPercent;
            mcfgStartup.uiNumVSMIdleTimingRegs = cscfgStartup.uiNumVSMIdleTimingRegs;
            mcfgStartup.uiNumVSMLongExpTimingRegs = cscfgStartup.uiNumVSMLongExpTimingRegs;
            mcfgStartup.uiNumOtherRegs = cscfgStartup.uiNumOtherRegs;

            // Marshal the VSM Settings
            if (mcfgStartup.uiNumVSMIdleTimingRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumVSMIdleTimingRegs);
                mcfgStartup.pCfgVSMIdleTimingRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumVSMIdleTimingRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pCfgVSMIdleTimingRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
            if (mcfgStartup.uiNumVSMLongExpTimingRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumVSMLongExpTimingRegs);
                mcfgStartup.pCfgVSMLongExpTimingRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumVSMLongExpTimingRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pCfgVSMLongExpTimingRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }

            // 'Other' Registers
            if (mcfgStartup.uiNumOtherRegs > 0)
            {
                ptrSetting = Marshal.AllocCoTaskMem(Marshal.SizeOf(cscfgSetting) * (int)mcfgStartup.uiNumOtherRegs);
                mcfgStartup.pOtherRegs = ptrSetting;
                for (uint i = 0; (i < mcfgStartup.uiNumOtherRegs) && (ptrSetting != null); ++i)
                {
                    Marshal.StructureToPtr(cscfgStartup.pOtherRegs[i], ptrSetting, false);
                    ptrSetting += Marshal.SizeOf(cscfgSetting);
                }
            }
        }

        #endregion Engineering Marshalling Wrapper Functions

        #endregion Engineering Functions
    }
}
