﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Configuration file management class
//
// Copyright(c) 2014-2016 AppliedLogix, LLC
// Portions of this library are Copyright (c) 2009-2014 VariPath Technologies, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

//--------------------------------------------------------
// Configuration File management class.
// Extends the Ini File Libary
//--------------------------------------------------------
using System;
using System.IO;
using System.Diagnostics;
using System.Linq;
using System.Reflection;

namespace FliUtils.Config
{
    public class Config : IniFile
    {
        //--------------------------------
        string _msg;

        // For error logging:
        public string StatusMsg { get; private set; }

        //--------------------------------
        /// <summary>
        /// Construct an empty config collection 
        /// </summary>
        public Config()
        {
        }

        //--------------------------------
        /// <summary>
        /// Loads one or more config (.ini) files from the application directory.
        /// </summary>
        /// <param name="filename">wildcard expression for files (or progname*.ini if empty)</param>
        /// <returns>false on any error</returns>
        public Boolean LoadProgramConfig(string filename = null)
        {
            if (String.IsNullOrEmpty(filename))
            {
                filename = GetProgramName() + "*.ini";
            }
            return LoadConfig(filename, GetProgramDir());
        }

        //--------------------------------
        /// <summary>
        /// Loads one or more config (.ini) files from the application directory.
        /// </summary>
        /// <param name="filename">wildcard expression for files (or *.ini if empty)</param>
        /// <returns>false on any error</returns>
        public Boolean LoadDataConfig(string filename = null)
        {
            return LoadConfig(filename, GetDataDir());
        }

        //--------------------------------
        /// <summary>
        /// Loads one or more config (.ini) files from the specified
        /// directory (defaults to ProgramData\CompanyName\ProgramName\"
        /// </summary>
        /// <param name="filename">wildcard expression for files (or *.ini if empty)</param>
        /// <param name="dir">directory (or default if empty)</param>
        /// <returns>false if erroneous config file(s) found</returns>
        public Boolean LoadConfig(string filename = null, string dir = null)
        {
            StatusMsg = "";
            int goodFiles = 0;
            int badFiles = 0;

            if(String.IsNullOrEmpty(dir))
            {
                dir = GetDataDir();
            }

            if (!CreateDir(dir))
            {
                StatusMsg = String.Format("LoadConfig : Unable to open {0}", dir);
                return false;
            }


            if (String.IsNullOrEmpty(filename))
            {
                filename = "*.ini";
            }

            StatusMsg = String.Format("LoadConfig : {0}\\{1}", dir, filename);

            string[] files = Directory.GetFiles(dir, filename);

            if(files.Length > 0)
            {
                foreach (string f in files)
                {
                    uint nSections = NumSections;
                    uint nKeys = NumKeys;
                    if(LoadFile(f, true))
                    {
                        goodFiles++;
                    }
                    else
                    {
                        _msg = String.Format("    error processing file {0}", f);
                        StatusMsg += _msg;
                        badFiles++;
                    }
                }
                _msg = String.Format("  valid config files : {0}, invalid : {1}", goodFiles, badFiles);
                StatusMsg += _msg;
            }
            else
            {
                // Not an error
                StatusMsg = String.Format("  No config file was found");
            }
            return (badFiles == 0);
        }

        //--------------------------------
        /// <summary>
        /// Save current configuration data to a file
        /// **** Not implemented
        /// </summary>
        /// <param name="filename"></param>
        /// <param name="dir"></param>
        /// <returns>false on any error</returns>
        public Boolean SaveConfig(string filename = "", string dir = "")
        {
            return false;
        }

        /// <summary>
        /// Get the default configuration file name
        /// </summary>
        /// <returns> the file name</returns>
        public string GetDefaultFileName()
        {
            string name = Assembly.GetEntryAssembly().GetName().Name;
            if(String.IsNullOrEmpty(name))
            {
                name = "configuration";
            }

            return name + ".ini";
        }

        //--------------------------------
        /// <summary>
        /// Create a directory if it does not exist
        /// </summary>
        /// <param name="dirPath"></param>
        /// <returns>true if created or already exists</returns>
        public Boolean CreateDir(string dirPath)
        {
            Boolean ok = false;
            if (!String.IsNullOrEmpty(dirPath))
            {
                try
                {
                    if (!Directory.Exists(dirPath))
                    {
                        Directory.CreateDirectory(dirPath);
                    }
                    ok = true;
                }
                catch
                {
                    Trace.WriteLine("Config.CreateDir: Unable to create directory '" + dirPath + "'");
                }
            }
            else
            {
                Trace.WriteLine("Config.CreateDir: Empty or null directory path");
            }

            return ok;
        }

        //--------------------------------
        /// <summary>
        /// Returns the path of the directory containing the current exectable file
        /// </summary>
        /// <returns></returns>
        public string GetProgramDir()
        {
            string dir =  AppDomain.CurrentDomain.BaseDirectory;
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }
            Trace.WriteLine("Program directory: '" + dir + "'");
            return dir;
        }

        //--------------------------------
        /// <summary>
        /// Get the program data directory path for "all users"
        /// </summary>
        /// <returns>directory path string</returns>
        public static string GetDataDir()
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }
            dir += GetShortCompanyName() + "\\" + GetProgramName() + "\\";
            Trace.WriteLine("Program Data directory: '" + dir + "'");
            return (dir);
        }

        //--------------------------------
        /// <summary>
        /// Get the desktop directory for the current user
        /// </summary>
        /// <returns>directory path string</returns>
        public static string GetDesktopDir()
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }
            Trace.WriteLine("User Desktop directory: '" + dir + "'");
            return (dir);
        }

        //--------------------------------
        /// <summary>
        /// Get the home directory for the current user
        /// </summary>
        /// <returns>directory path string</returns>
        public static string GetHomeDir()
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.UserProfile);
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }
            Trace.WriteLine("User Home directory: '" + dir + "'");
            return (dir);
        }

        //--------------------------------
        /// <summary>
        /// Get the "My Documents" directory for the current user
        /// </summary>
        /// <returns>directory path string</returns>
        public static string GetDocumentsDir()
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }
            Trace.WriteLine("User Documents directory: '" + dir + "'");
            return (dir);
        }

        //--------------------------------
        /// <summary>
        /// Get the program folder in the "My Documents" directory for the current user
        /// </summary>
        /// <returns>directory path string</returns>
        public static string GetProgramDocumentsDir()
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            if (!dir.EndsWith("\\"))
            {
                dir += "\\";
            }

            dir += GetShortCompanyName() + "\\" + GetProgramName() + "\\";
            Directory.CreateDirectory(dir);

            Trace.WriteLine("Program documents directory: '" + dir + "'");
            return (dir);
        }

        //--------------------------------
        /// <summary>
        /// Get the current program name
        /// </summary>
        /// <returns>program name</returns>
        public static string GetProgramName()
        {
           string progName = Assembly.GetEntryAssembly().GetName().Name;
           Trace.WriteLine("Program Name: '" + progName + "'");
           return (progName);
        }

        //--------------------------------
        /// <summary>
        /// Get the program version string
        /// </summary>
        /// <returns></returns>
        public static string GetProgramVersion()
        {
            var versionInfo = FileVersionInfo.GetVersionInfo(Assembly.GetEntryAssembly().Location);
            Trace.WriteLine("Program Version: " + versionInfo.FileVersion);
            return versionInfo.FileVersion;
        }

        //--------------------------------
        /// <summary>
        /// Get the Company name
        /// </summary>
        /// <returns></returns>
        public static string GetFullCompanyName()
        {
            var versionInfo = FileVersionInfo.GetVersionInfo(Assembly.GetEntryAssembly().Location);
            Trace.WriteLine("Full Company Name: " + versionInfo.CompanyName);
            return versionInfo.CompanyName;
        }

        //--------------------------------
        /// <summary>
        /// Get the first word of the company name
        /// </summary>
        /// <returns></returns>
        public static string GetShortCompanyName()
        {
            String fullName = GetFullCompanyName();

            string shortName = new string(fullName.TakeWhile(c => !Char.IsPunctuation(c)).ToArray());
            if (String.IsNullOrEmpty(shortName))
            {
                shortName = "company";
            }
            Trace.WriteLine("Short Company Name: " + shortName);
            return shortName;
        }

        //--------------------------------
        /// <summary>
        /// Get the program description string
        /// </summary>
        /// <returns></returns>
        public static string GetProgramDescription()
        {
            Assembly assembly = Assembly.GetEntryAssembly();
            var descriptionAttribute = assembly
                     .GetCustomAttributes(typeof(AssemblyDescriptionAttribute), false)
                     .OfType<AssemblyDescriptionAttribute>()
                     .FirstOrDefault();

            if (descriptionAttribute == null)
            {
                Trace.WriteLine("No assembly description found");
                return "";
            }
            else
            {
                Trace.WriteLine("Assembly description: " + descriptionAttribute.Description);
                return descriptionAttribute.Description;
            }
        }
    }
}

//--------------------------------------------------------
// End of file
//--------------------------------------------------------
