﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Value converters for use in XAML Bindings
//
// Copyright(c) 2016-2017 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************
using System;
using System.Windows;
using System.Windows.Markup;
using System.Windows.Data;

//********************************************************************************
// Value converters for use in XAML Bindings
// Usage:
// Declare namespace:
//          xmlns:cvt="clr-namespace:MemToolApp.Converters"
// In Binding, specify Converter class name without "Extension". e.g.:":
//           IsEnabled="{Binding xxxx, Converter={cvt:Invert}}"
//          Visibility="{Binding xxxx, Converter={cvt:TrueVisibleFalseCollapsed}}"
//
//********************************************************************************

namespace FliUtils.Converters
{
    //-------------------------------------------
    /// <summary>
    /// Convert bool to its complement
    /// </summary>
    public class InvertExtension : ValueConverterBase
    {
        private static InvertExtension _converter;
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            if (_converter == null)
            {
                _converter = new InvertExtension();
            }
            return _converter;
        }

        public override object Convert(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            object rtn = Binding.DoNothing;
            var val = value as bool?;
            if (val == true)
            {
                rtn = false;
            }
            else if (val == false)
            {
                rtn = true;
            }
            return rtn;
        }

        public override object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            bool rtn = false;

            if (value != null && ((value as bool?) == false))
            {
                rtn = true;
            }
            return rtn;
        }
    }

    //-------------------------------------------
    /// <summary>
    /// Convert bool to visibility, true=Visible, false=Collapsed
    /// </summary>
    public class TrueVisibleFalseCollapsedExtension : ValueConverterBase
    {
        private static TrueVisibleFalseCollapsedExtension _converter;
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            if (_converter == null)
            {
                _converter = new TrueVisibleFalseCollapsedExtension();
            }
            return _converter;
        }

        public override object Convert(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            object rtn = Binding.DoNothing;
            var val = value as bool?;
            if (val == true)
            {
                rtn = Visibility.Visible;
            }
            else if (val == false)
            {
                rtn = Visibility.Collapsed;
            }
            return rtn;
        }

        public override object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            bool rtn = false;

            if (value != null && ((value as Visibility?) == Visibility.Visible))
            {
                rtn = true;
            }
            return rtn;
        }
    }

    //-------------------------------------------
    /// <summary>
    /// Convert bool to visibility, true=Visible, false=Collapsed
    /// </summary>
    public class FalseVisibleTrueCollapsedExtension : ValueConverterBase
    {
        private static FalseVisibleTrueCollapsedExtension _converter;
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            if (_converter == null)
            {
                _converter = new FalseVisibleTrueCollapsedExtension();
            }
            return _converter;
        }

        public override object Convert(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            object rtn = Binding.DoNothing;
            var val = value as bool?;
            if (val == true)
            {
                rtn = Visibility.Collapsed;
            }
            else if (val == false)
            {
                rtn = Visibility.Visible;
            }
            return rtn;
        }

        public override object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            bool rtn = false;

            if (value != null && ((value as Visibility?) == Visibility.Collapsed))
            {
                rtn = true;
            }
            return rtn;
        }
    }

    //-------------------------------------------
    /// <summary>
    /// Convert bool to visibility, true=Visible, false=Hidden
    /// </summary>
    public class TrueVisibleFalseHiddenExtension : ValueConverterBase
    {
        private static TrueVisibleFalseHiddenExtension _converter;
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            if (_converter == null)
            {
                _converter = new TrueVisibleFalseHiddenExtension();
            }
            return _converter;
        }

        public override object Convert(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            object rtn = Binding.DoNothing;
            var val = value as bool?;
            if (val == true)
            {
                rtn = Visibility.Visible;
            }
            else if (val == false)
            {
                rtn = Visibility.Hidden;
            }
            return rtn;
        }

        public override object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            bool rtn = false;
            
            if (value != null && ((value as Visibility?) == Visibility.Visible))
            {
                rtn = true;
            }
            return rtn;
        }
    }

    //-------------------------------------------
    /// <summary>
    /// Convert bool to visibility, true=Visible, false=Hidden
    /// </summary>
    public class FalseVisibleTrueHiddenExtension : ValueConverterBase
    {
        private static FalseVisibleTrueHiddenExtension _converter;
        public override object ProvideValue(IServiceProvider serviceProvider)
        {
            if (_converter == null)
            {
                _converter = new FalseVisibleTrueHiddenExtension();
            }
            return _converter;
        }

        public override object Convert(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            object rtn = Binding.DoNothing;
            var val = value as bool?;
            if (val == true)
            {
                rtn = Visibility.Hidden;
            }
            else if (val == false)
            {
                rtn = Visibility.Visible;
            }
            return rtn;
        }

        public override object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
        {
            bool rtn = false;

            if (value != null && ((value as Visibility?) != Visibility.Visible))
            {
                rtn = true;
            }
            return rtn;
        }
    }
}
