﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Producer/Consumer Background job 
//
// Copyright(c) 2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

namespace FliUtils.Jobs
{
    //**************************************************************
    #region Constructor

    /// <summary>
    /// Produce-Consumer queue abstraction supporting multiple Producers, and a single Consumer Thread
    /// that will call the supplied Action to consume each item.  Both Soft Stop (finishes consuming
    /// queued items) and Hard Stop (all unprocessed items are discarded) are supported.
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class ConsumerJob<T>
    {
        /// <summary>
        /// c'tor
        /// </summary>
        /// <param name="consumeOne">Action method to consume the next queued item</param>
        public ConsumerJob(Action<T> consumeOne, string name = "", ThreadPriority priority = ThreadPriority.Normal)
        {
            _shutdown = true;
            _syncObject = new object();
            _queue = new Queue<T>();
            _consumeOne = consumeOne;
            _name = name;
            _priority = priority;
        }

        #endregion Constructor

        /// <summary>
        /// Returns an instantaneous snapshot of the number of items in the queue
        /// </summary>
        public int QueueCount { get { return _queue.Count; } }


        //**************************************************************
        #region Private Properties

        private volatile bool _shutdown;
        private Object _syncObject;
        private Queue<T> _queue;
        private Action<T> _consumeOne;
        private Thread _thread;
        private ThreadPriority _priority;
        private string _name;

        #endregion Private Properties

        //**************************************************************
        #region Public Properties

        public int ThreadId { get { return (_thread == null) ? 0 : _thread.ManagedThreadId; } }
        public string ThreadName { get { return (_thread == null) ? "null" : _thread.Name; } }

        #endregion Public Properties

        //**************************************************************
        #region Public Methods

        //-------------------------------------------
        /// <summary>
        /// Discard any unprocessed items in the queue.
        /// </summary>
        public void Clear()
        {
            lock (_syncObject)
            {
                _queue.Clear();
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Start the consumer Thread running, and allow items to be queued.
        /// </summary>
        public void Start()
        {
            lock (_syncObject)
            {
                if (_shutdown)
                {
                    _shutdown = false;
                    _thread = new Thread(new ThreadStart(ConsumerThread));
                    _thread.IsBackground = true;
                    if (!String.IsNullOrEmpty(_name))
                    {
                        _thread.Name = _name;
                    }
                    _thread.Priority = _priority;
                    _thread.Start();
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Soft stop the consumer thread, allowing it to finish processing all items in the queue.
        /// </summary>
        public void Stop()
        {
            lock (_syncObject)
            {
                if (!_shutdown)
                {
                    _shutdown = true;

                    // Pulse the sync object if the consumer is waiting on an empty queue
                    if (!_queue.Any())
                    {
                        Monitor.Pulse(_syncObject);
                    }
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Hard stop the consumer thread, discarding unprocessed items in the queue.
        /// Thread Interrupt is used rather than Abort.
        /// </summary>
        public void HardStop()
        {
            lock (_syncObject)
            {
                if (!_shutdown)
                {
                    _shutdown = true;
                    _queue.Clear();
                    _thread.Interrupt();
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Enqueue a single item if room.
        /// </summary>
        /// <param name="item"></param>
        public void Enqueue(T item, int maxcount = int.MaxValue)
        {
            lock (_syncObject)
            {
                if (!_shutdown && (_queue.Count < maxcount))
                {
                    _queue.Enqueue(item);
                    Monitor.Pulse(_syncObject);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Enqueue mutiple items.
        /// </summary>
        /// <param name="items"></param>
        public void Enqueue(IEnumerable<T> items, int maxcount = int.MaxValue)
        {
            lock (_syncObject)
            {
                if (!_shutdown && (_queue.Count < (maxcount - items.Count())))
                {
                    foreach (T item in items)
                    {
                        _queue.Enqueue(item);
                    }
                    Monitor.Pulse(_syncObject);
                }
            }
        }

        #endregion Public Methods

        //**************************************************************
        #region Private Methods

        //-------------------------------------------
        /// <summary>
        /// Thread body
        /// </summary>
        private void ConsumerThread()
        {
            try
            {
                while (true)
                {
                    bool doWork = false;
                    T item = default(T);

                    lock (_syncObject)
                    {
                        if (_queue.Any())
                        {
                            // Dequeue a work item
                            item = _queue.Dequeue();
                            doWork = true;
                        }
                        else if (_shutdown)
                        {
                            // Soft stop, simply return once queue is empty
                            return;
                        }
                        else
                        {
                            // Wait for something to do, or interruption
                            Monitor.Wait(_syncObject);
                        }
                    }

                    if (doWork)
                    {
                        // Consume the work item
                        try
                        {
                            _consumeOne(item);
                        }
                        catch (ThreadInterruptedException)
                        {
                            throw;
                        }
                        catch (Exception)
                        {
                            throw;
                        }
                    }
                }
            }
            catch (ThreadInterruptedException)
            {
                if (_shutdown)
                {
                    // Normal exit path for hard stops
                }
                else
                {
                    throw;
                }
            }
            catch (Exception)
            {
                throw;
            }
        }

        #endregion Private Methods
    }
}
