﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// Simple debug logging class
//
// Copyright(c) 2014-2016 AppliedLogix, LLC
// Portions of this library are Copyright (c) 2009-2014 VariPath Technologies, Inc.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using System.Text;

namespace FliUtils.Logger
{
    /// <summary>
    /// Logger is a lightweight multi-thread-safe (but NOT multi-process-safe)
    /// debug logging system capable of managing multiple logfiles simultaneously
    ///
    /// A singleton instance of the Logger class is maintained for
    /// each logfile path (log directory/basefile name). 
    /// 
    /// Any code in the process can obtain a reference to a specific logger
    /// instance by passing the desired directory and basename to the 
    /// static GetInstance method.
    /// 
    /// The maximum log file size, number of retained log files and logging level
    /// can be specified per instance.
    /// 
    /// </summary>
    public class Logger 
    {
        //**************************************************************
        // Constants

        // Convenience Log levels
        public const int LOGALL         = 99;
        public const int LOGDEBUG       = 8;
        public const int LOGTRACE       = 7;
        public const int LOGVERBOSE     = 6;
        public const int LOGINFO        = 5;
        public const int LOGWARN        = 4;
        public const int LOGERROR       = 3;
        public const int LOGEXCEPTION   = 2;
        public const int LOGNOTE        = 1;
        public const int LOGNONE        = 0;

        public const int MAX_THRESHOLD  = LOGALL;
        public const int MIN_THRESHOLD  = LOGNONE;
        public const int DEF_THRESHOLD  = LOGVERBOSE;
        public const int DEF_CONSOLE_THRSHOLD = LOGINFO;

        // File size and count
        private const int DEF_MAX_FILE_COUNT = 10;          // default max number of log files to preserve
        private const int DEF_MAX_FILE_SIZE = 1000000;      // default max size of each file
        private const Int32 LINES_PER_SIZE_CHECK = 50;      // how often (in lines) to check for max size

        // Misc
        private const string DEF_COMPANY="AppliedLogix";    // Default company name if not specified in assembly
        private const string LOG_FILE_EXTENSION = ".log";   // log file extension
        public const string TIME_FORMAT = "MMdd_HH:mm:ss.fff";    // log timestamp format


        //**************************************************************
        // Constructor

        //-------------------------------------------
        /// <summary>
        /// Private constructor. Must use GetInstance to
        /// obtain reference to a logger instance
        /// </summary>
        /// <param name="baseName">log file base name, defaults to app name</param>
        /// <param name="logdir">log file dir, defaults to .</param>
        protected Logger(string basename, string logdir)
        {
            _logLock = new Object();

            BaseName = basename;
            LogDir = logdir;
            BasePath = LogDir + BaseName;

            Threshold = DEF_THRESHOLD;
            ConsoleThreshold = DEF_CONSOLE_THRSHOLD;
            MaxFiles = DEF_MAX_FILE_COUNT;
            MaxFileSize = DEF_MAX_FILE_SIZE;
            EnableLog = false;
        }

        //-------------------------------------------
        /// <summary>
        /// Static constructor - initialize static properties
        /// </summary>
        static Logger()
        {
            _createLock = new Object();
            _logs = new Dictionary<string, Logger>();
            EnableConsole = Debugger.IsAttached;
            ConsoleThreshold = DEF_CONSOLE_THRSHOLD;
        }

        //**************************************************************
        // Private Static Properties

        private static Dictionary<string, Logger> _logs;
        private static object _createLock;

        //**************************************************************
        //  Public Static properties
        public static bool EnableConsole { get; set; }
        public static int ConsoleThreshold {get; set;}

        //**************************************************************
        // Private Properties
        private object _logLock;
        private StreamWriter LogFile { get; set; }
        private int NLines {get; set;}

        //**************************************************************
        // Public Properties

        public string LogDir {get; protected set;}
        public string BaseName {get; protected set;}
        public string BasePath {get; protected set;}
        public string LogFileName {get; protected set;}
        public int Threshold {get; set;}
        public int MaxFiles {get; set;}
        public long MaxFileSize {get; set;}
        public bool EnableLog {get; set;}

        //**************************************************************
        // Public methods

        //-------------------------------------------
        /// <summary>
        /// Public Factory method
        /// Rather than calling the constructor directly,
        /// any code using the logger must call the
        /// GetInstance() factory method.
        /// e.g. var log = Logger.GetInstance(..)
        ///
        /// This ensures that a single instance of the
        /// Logger class exists for each Logfile basepath
        /// i.e. LogDir/basename* 
        /// 
        /// </summary>
        /// <param name="basename">log file base name</param>
        /// <param name="logdir">log file directory path, defaults to .</param>
        /// <returns>reference to the requested logger instance</returns>
        public static Logger GetInstance(string basename = null, string logdir = null)
        {

            if (String.IsNullOrEmpty(basename))
            {
                basename = GetAppName();
            }

            if (String.IsNullOrEmpty(logdir))
            {
                logdir = ".\\";
            }
            else if(!logdir.EndsWith("\\"))
            {
               logdir += "\\";
            }

            string basePath = logdir + basename;
            Logger instance = null;

            lock (_createLock)
            {
                if (_logs.ContainsKey(basePath))
                {
                    // An instance of this logfile pattern already exists,
                    // so just return a reference to it.
                    instance = _logs[basePath];
                }
                else
                {
                    // Otherwise, create and initialize a new instance
                    instance = new Logger(basename, logdir);
                    _logs.Add(basePath, instance);

                    // Make sure the directory exists
                    try
                    {
                        if (!Directory.Exists(logdir))
                        {
                            Directory.CreateDirectory(logdir);
                        }

                        instance.OpenFile();
                    }
                    catch (Exception ex)
                    {
                        LogConsole(LOGERROR, ": Unable to create log directory '" + logdir + "': " + ex.Message);
                    }
                }
            }
            return instance;
        }

        //**************************************************************
        // Public Logging methods

        //-------------------------------------------
        /// <summary>
        /// Log message at the specified log level
        /// </summary>
        /// <param name="level">log level</param>
        /// <param name="line">log message</param>
        public void Log(Int32 level, string line)
        {
            Log(level, line, true);
        }

        //-------------------------------------------
        /// <summary>
        /// Trace level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogTrace(string line)
        {
            Log(LOGTRACE, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Debug level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogDebug(string line)
        {
            Log(LOGDEBUG, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Verbose level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogVerbose(string line)
        {
            Log(LOGVERBOSE, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Information level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogInfo(string line)
        {
            Log(LOGINFO, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Warning level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogWarn(string line)
        {
            Log(LOGWARN, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Error level log
        /// </summary>
        /// <param name="line">log message</param>
        public void LogError(string line)
        {
            Log(LOGERROR, line);
        }

        //-------------------------------------------
        /// <summary>
        /// Note level log -- No timestamp header
        /// </summary>
        /// <param name="line">log message</param>
        public void LogNote(string line)
        {
            Log(LOGNOTE, line, false);
        }

        //-------------------------------------------
        /// <summary>
        /// Exception log message
        /// </summary>
        /// <param name="ex">Exception object</param>
        /// <param name="line">log message</param>
        /// <param name="stackTrace">true to log stacktrace</param>
        public void LogEx(Exception ex, string line = "Exception caught:", Boolean stackTrace = true)
        {
            LogEx(LOGEXCEPTION, ex, line, stackTrace);
        }

        //-------------------------------------------
        /// <summary>
        /// Base log version - rarely used externally, but no reason to make private
        /// </summary>
        /// <param name="level">log level</param>
        /// <param name="line">log message</param>
        /// <param name="addHeader">True to add header before line</param>
        public void Log(Int32 level, string line, Boolean addHeader = true)
        {
            if ((level <= Threshold) && (EnableLog || EnableConsole))
            {
                string header = String.Empty;
                if (addHeader)
                {
                    header = GetHeader(level);
                }

                lock (_logLock)
                {
                    WriteFile(header + line);
                }
            }
        }

        //-------------------------------------------
        /// <summary>
        /// base exception log method. If exception is null, this defaults to a standard error message
        /// </summary>
        /// <param name="level">log level</param>
        /// <param name="ex">Exceptioin object</param>
        /// <param name="line">log message</param>
        /// <param name="stackTrace">True to display stacktrace</param>
        public void LogEx(Int32 level, Exception ex, string line, Boolean stackTrace)
        {
            StringBuilder msg = new StringBuilder();
            if (!String.IsNullOrEmpty(line))
            {
                msg.Append(line);
            }
            if( ex != null)
            {
                if (!String.IsNullOrEmpty(ex.Message))
                {
                    msg.Append(Environment.NewLine + "   " + ex.Message);
                }
                if (stackTrace)
                {
                    msg.Append(Environment.NewLine + ex.StackTrace);
                }
            }
            Log(level, msg.ToString());
        }


        //**************************************************************
        // Console Logging methods

        //-------------------------------------------
        /// <summary>
        /// Log to console when all else fails
        /// </summary>
        /// <param name="level">log level</param>
        /// <param name="line"> log message</param>
        public static void LogConsole(Int32 level, string line)
        {
            if (level <= ConsoleThreshold)
            {
                string msg = String.Empty;

                if (level <= LOGERROR)
                {
                    msg += "**> " + line;
                }
                else
                {
                    msg += "--> " + line;
                }

                try
                {
                    if (Debugger.IsAttached)
                    {
                        Debug.WriteLine(msg);
                    }
                    else
                    {
                        Console.WriteLine(msg);
                    }
                }
                catch
                {
                    // Not a lot we can do here!
                }
            }
        }

        //**************************************************************
        // Private Utility methods

        //-------------------------------------------
        // Unformatted LogFile Write
        private void WriteLine(Int32 level, string line)
        {
            if ((level <= Threshold) && (EnableLog || EnableConsole))
            {
                WriteFile(line);
            }
        }

        //-------------------------------------------
        // Open a logfile
        private bool OpenFile()
        {
            try
            {
                // Close the existing logfile for this instance, if it exists
                CloseFile();

                if (!String.IsNullOrEmpty(LogDir) && !Directory.Exists(LogDir))
                {
                    Directory.CreateDirectory(LogDir);
                }

                CleanUpLogFiles(-1);    // Delete 1 extra file if needed to make room for new one.
                LogFileName = BasePath + "_" + DateTime.Now.ToString("yyyyMMdd_HHmmss") + LOG_FILE_EXTENSION;
                LogFile = new StreamWriter(LogFileName);
                LogFile.AutoFlush = true;
                NLines = 0;

                // Write the initial message to the log
                this.Log(LOGNOTE, DateTime.Now.ToString("dddd, MMMM dd, yyyy"));
                this.Log(LOGNOTE, "Log file created: " + this.LogFileName);
                if (Debugger.IsAttached)
                {
                    this.Log(LOGNOTE, "Running under the debugger");
                }
                this.WriteLine(LOGNOTE, "");
            }
            catch( Exception ex )
            {
                if (String.IsNullOrEmpty(LogFileName))
                {
                    LogFileName = "unknown";
                }
                LogConsole(LOGERROR, LogFileName + ": Unable to open log file: " + ex.Message);
            }

            return (LogFile != null);
        }

        //-------------------------------------------
        // Close the logfile
        private void CloseFile()
        {
            if (LogFile != null)
            {
                try
                {
                    Log(LOGNOTE,"Log file closed");
                    LogFile.Close();
                    LogFile.Dispose();
                    LogFile = null;
                    NLines = 0;
                }
                catch( Exception ex)
                {
                    if (String.IsNullOrEmpty(LogFileName))
                    {
                        LogFileName = "unknown";
                    }
                    LogConsole(LOGWARN, LogFileName + ": Exception trying to Close log file: " + ex.Message);
                }
            }
        }

        //-------------------------------------------
        // Write line to file
        private void WriteFile(string line)
        {
            Boolean fileError = (LogFile == null);
            if (!fileError && EnableLog)
            {
                try
                {
                    if (++NLines > LINES_PER_SIZE_CHECK)
                    {
                        NLines = 0;
                        LimitFileSize();
                    }
                    LogFile.WriteLine(line);
                }
                catch( Exception ex )
                { 
                    fileError = true;
                    if (String.IsNullOrEmpty(LogFileName))
                    {
                        LogFileName = "unknown";
                    }
                    LogConsole(LOGERROR, LogFileName + ": Exception trying to write log file: " + ex.Message);
                }
            }

            if (fileError || EnableConsole)
            {
                try
                {
                    if( Debugger.IsAttached)
                    {
                        Debug.WriteLine(line);
                    }
                    else
                    {
                        Console.WriteLine(line);
                    }
                }
                catch
                {
                    // Not a lot we can do here!
                }
            }
        }

        //-------------------------------------------
        // delete oldest logfile(s) if max logfiles has  been exceeded
        private void CleanUpLogFiles(Int32 correction )
        {
            if (Directory.Exists(LogDir))
            {
                // Keep last n log files
                string[] logFiles = Directory.GetFiles(LogDir, BaseName + "_*" + LOG_FILE_EXTENSION);
                Array.Sort(logFiles);
                Array.Reverse(logFiles);
                for (Int32 i = MaxFiles+correction; i < logFiles.Length; i++)
                {
                    try
                    {
                        File.Delete(logFiles[i]);
                    }
                    catch( Exception ex )
                    {
                        LogError(logFiles[i] + " : Unable to delete: " + ex.Message);
                    }
                }
            }
        }

        //-------------------------------------------
        // Start a new logfile if max file size has been exceeded
        private void LimitFileSize()
        {
            if (LogFile != null)
            {
                try
                {
                    LogFile.Flush();
                    NLines = 0;

                    // Check size of file.  If greater than Max size, close file and open new one
                    if (LogFile.BaseStream.Length > MaxFileSize)
                    {
                        this.LogNote("Log file has reached maximum size");
                        CloseFile();
                        OpenFile();
                    }
                }
                catch (Exception ex)
                {
                    LogError("Exception while starting new log file: " + ex.Message);
                }
            }
        }

        //-------------------------------------------
        // Return a string containing a log entry header
        // with the current timestamp
        private static string GetHeader(Int32 level)
        {
            string timeStr = DateTime.Now.ToString(TIME_FORMAT);
            string threadId = System.Threading.Thread.CurrentThread.ManagedThreadId.ToString().PadLeft(2);
            string threadName = System.Threading.Thread.CurrentThread.Name.PadRight(9);
            if (String.IsNullOrEmpty(threadName))
            {
                threadName = "NoName   ";
            }
            return (timeStr + " (" + threadId + "," + threadName + ") " + GetLevelString(level) + ": ");
        }

        //-------------------------------------------
        // Return a string representation of a log level
        public static string GetLevelString(Int32 level)
        {
            switch (level)
            {
                case LOGALL:
                    return "ALL    ";
                case LOGDEBUG:
                    return "DEBUG  ";
                case LOGTRACE:
                    return "TRACE  ";
                case LOGVERBOSE:
                    return "VERBOSE";
                case LOGINFO:
                    return "INFO   ";
                case LOGWARN:
                    return "WARNING";
                case LOGERROR:
                    return "**ERROR";
                case LOGEXCEPTION:
                    return "*EXCEPTION";
                case LOGNOTE:
                    return "NOTE   ";
                default:
                    return "LVL_" + level.ToString() + " ";
            }
        }

        //-------------------------------------------
        /// <summary>
        /// Get the default log directory path
        /// (e.g. Program Data\company\subdir\logs)
        /// </summary>
        /// <param name="subdir">optional subdirectory</param>
        /// <returns>full path to log directory</returns>
        private static string GetDefaultLogDir(string subdir = null)
        {
            string dir = Environment.GetFolderPath(Environment.SpecialFolder.CommonApplicationData);
            dir += GetAppName();
            if(!String.IsNullOrEmpty(subdir))
            {
                dir += @"\" + subdir;
            }
            return dir + @"\logs\";
        }

        //-------------------------------------------
        /// <summary>
        /// Return the base name of the current application
        /// </summary>
        /// <returns></returns>
        private static string GetAppName()
        {
            string name = Assembly.GetEntryAssembly().GetName().Name;
            if (String.IsNullOrEmpty(name))
            {
                name = System.Diagnostics.Process.GetCurrentProcess().ProcessName;
            }

            return name;
        }

        //-------------------------------------------
        // Get the company name from the application
        private static string GetCompanyName()
        {
            string company = "";
            var appAssembly = Assembly.GetEntryAssembly();
            foreach (Attribute attr in Attribute.GetCustomAttributes(appAssembly))
            {
                if (attr.GetType() == typeof(AssemblyCompanyAttribute))
                {
                    company = ((AssemblyCompanyAttribute)attr).Company;
                    break;
                }
            }

            if (String.IsNullOrEmpty(company))
            {
                company = DEF_COMPANY;
            }
            return company;
        }

    }
}

//--------------------------------------------------------
// End of file
//--------------------------------------------------------
