﻿//**************************************************************
// AppliedLogix DotNet Utilities Library
// USB Device Connection/Disconnection Event and Information Class
//
// Copyright(c) 2016 AppliedLogix, LLC
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//**************************************************************
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Management;
using System.Text;
using System.Threading.Tasks;

namespace FliUtils.Usb
{
    /// <summary>
    /// USB Device Information Class
    /// </summary>
    public class USBDeviceInfo
    {
        //**************************************************************
        #region Constructor
        
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="path"> Full device path from WMI __InstanceOperationEvent.Dependent property </param>
        /// <param name="connect">true if device is being connected, false if being disconnected</param>
        public USBDeviceInfo(string path = "", bool connect = false, Collection<USBDeviceID> ids = null)
        {
            //DeviceID/HardwareID Parsing definitions
            const string IDTAG  = @"=";         // Start of Device ID
            const string VIDTAG = @"\VID_";     // Vendor ID
            const int VIDLEN    = 4;    
            const string PIDTAG = @"&PID_";     // Product ID
            const int PIDLEN    = 4;
            const string REVTAG = @"&REV_";     // Revision
            const int REVLEN    = 4;
            const string MITAG  = @"&MI_";      // Multi-Interface number
            const int MILEN     = 2;
            const string IIDTAG = @"\";         // Instance ID (or Serial number)

            // Get basic device ID fields
            DeviceID = getField(path, IDTAG, -1);
            Char[] quotes = { '"' };
            DeviceID = DeviceID.Trim(quotes);
            VID = getField(DeviceID, VIDTAG, VIDLEN);
            PID = getField(DeviceID, PIDTAG, PIDLEN);
            MI = getField(DeviceID, MITAG, MILEN);

            // See if we are interested in this one
            Valid = (ids == null) || (ids.Count == 0);
            if(!Valid)
            { 
                // at least one constraint has been defined - check for match
                foreach(var id in ids)
                {
                    if ((String.IsNullOrEmpty(id.VID) || (VID.ToLower() == id.VID.ToLower())) &&
                        (String.IsNullOrEmpty(id.PID) || (PID.ToLower() == id.PID.ToLower())) &&
                        (String.IsNullOrEmpty(id.MI)  || (MI.ToLower()  == id.MI.ToLower() )))
                    {
                        Key = id.KEY;
                        Valid = true;
                        break;
                    }
                }
            }

            if(Valid)
            {
                // We have a match - get the remainder of the parameters
                int index = DeviceID.LastIndexOf(IIDTAG);
                if ((index > 0) && (index < DeviceID.Length))
                {
                    IID = DeviceID.Substring(index + 1);
                }

                // Get additional device information when connected
                if (connect)
                {
                    string query = "SELECT * FROM Win32_PnPEntity WHERE DeviceID LIKE " + "'" + DeviceID + "'";
                    //string query = "SELECT * FROM Win32_PnPEntity WHERE DeviceID LIKE " +  DeviceID ;
                    ManagementObjectSearcher searcher = new ManagementObjectSearcher("root\\CIMV2", query);
                    ManagementObjectCollection colItems = searcher.Get();

                    foreach (ManagementObject queryObj in colItems)
                    {
                        try
                        {
                            ClassGuid = queryObj["ClassGuid"].ToString();
                            Name = queryObj["Name"].ToString();
                            Manufacturer = queryObj["Manufacturer"].ToString();
                            Status = queryObj["Status"].ToString();
                            PNPClass = queryObj["PNPClass"].ToString();
                            HardwareID = queryObj["HardwareID"].ToString();
                            REV = getField(HardwareID, REVTAG, REVLEN);

                            if (PNPClass == "Ports")
                            {
                                int end = Name.Length - 1;
                                if ((end > 5) && Name[end] == ')')
                                {
                                    int start = Name.LastIndexOf("(");
                                    if (start >= 0)
                                    {
                                        Port = Name.Substring(start + 1, (end - start - 1));
                                    }
                                }
                            }
                        }
                        catch
                        {
                        }
                    }
                }
            }
        }

        #endregion Constructor
        //**************************************************************
        #region Properties

        // USB Device Information Properties
        public bool   Valid { get; private set; }
        public string Key { get; private set; }
        public string Path { get; private set; }
        public string PNPClass { get; private set; }
        public string ClassGuid { get; private set; }
        public string DeviceID { get; private set; }
        public string HardwareID { get; private set; }
        public string VID { get; private set; }
        public string PID { get; private set; }
        public string MI { get; private set; }
        public string REV { get; private set; }
        public string IID { get; private set; }
        public string Manufacturer { get; private set; }
        public string Name { get; private set; }
        public string Status { get; private set; }
        public string Port { get; private set; }

        #endregion Properties

        //**************************************************************
        #region Public Methods

        /// <summary>
        /// Default string representation of USB Device Info
        /// </summary>
        public override string ToString()
        {
            string desc = "";

            if(!String.IsNullOrEmpty(Key))
            {
                desc += Key + ": ";
            }

            desc += "VID=" + VID + ", PID=" + PID;

            if (!String.IsNullOrEmpty(MI))
            {
                desc += ", MI=" + MI;
            }

            if (!String.IsNullOrEmpty(REV))
            {
                desc += ", REV=" + REV;
            }

            desc += ", IID=" + IID;

            if (!String.IsNullOrEmpty(Port))
            {
                desc += ", Port=" + Port;
            }

            if (!String.IsNullOrEmpty(Name))
            {
                desc += ", \"" + Name + "\"";
            }

            return desc;
        }

        #endregion Public Methods
        
        //**************************************************************
        #region Private Methods

        private string getField(string src, string tag, int length)
        {
            string field = "";
            int index = src.IndexOf(tag);
            if (index > 0)
            {
                try
                {
                    if (length < 0)
                    {
                        field = src.Substring(index + tag.Length);
                    }
                    else
                    {
                        field = src.Substring(index + tag.Length, length);
                    }
                }
                catch
                {
                }
            }
            return field;
        }

        #endregion Private Methods
   }
}

